// GUI_fluid_sim.cpp
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <chrono>
#include <cstdlib> // For exit()

#include <GL/glew.h>
#include <GLFW/glfw3.h>   // Use GLFW instead of GLUT

#include "GL_utilities.h"

#include "imgui.h"
#include "imgui_impl_glfw.h"     // Use the GLFW backend
#include "imgui_impl_opengl3.h"  // Use the OpenGL3 backend

#include "include/ImGuiManager.h"
#include "include/FluidConfig.h"
#include "include/PCISPH_sim.h" // Header for CUDA function declarations
#include "include/spawn_particles.h"
#include "include/Camera.h"
#include "settings/constants.h"

#define MAIN
#include "VectorUtils4.h"

FluidConfig fluidConfig;

// Number of particles
const int NUM_PARTICLES = 50000;
const float GRID_SPACING = 0.25f;
int NUM_BOUNDARY_PARTICLES = 0;

const float pointRadius = 20.0f;
const float pointScale = 10.0f;

auto previousTime = std::chrono::high_resolution_clock::now();
float deltaTime = 0.016f;

GLuint shaderProgram;
GLuint positionAttribLoc;
GLuint vbo;
GLuint vbo_boundary;

GLuint modelMatrixLoc;
GLuint viewMatrixLoc;
GLuint projectionMatrixLoc;
GLuint pointRadiusLoc;
GLuint pointScaleLoc;
GLuint uColorLoc;

GLuint fbo;
GLuint fboTexture;
GLuint depthRBO; // Depth renderbuffer if needed
int fbWidth = 800;  // Adjust as needed, or dynamically update on window resize
int fbHeight = 600;

Camera* cameraPtr = nullptr;

void printFrameStats() {
    static float timeAccumulator = 0.0f;
    static int frameCount = 0;

    timeAccumulator += deltaTime;
    frameCount++;

    if (timeAccumulator >= 1.0f) {
        float averageDeltaTime = timeAccumulator / frameCount;
        float fps = frameCount / timeAccumulator;
        printf("FPS: %.2f, Average DeltaTime: %.1f ms\n", fps, averageDeltaTime*1000);
        timeAccumulator = 0.0f;
        frameCount = 0;
    }
}

void renderBorder() {
    glUniform3f(uColorLoc, 0.29f, 1.0f, 0.57f);
    glBindBuffer(GL_ARRAY_BUFFER, vbo_boundary);
    glEnableVertexAttribArray(positionAttribLoc);
    glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, GL_FALSE, sizeof(float3), (void*)0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glDrawArrays(GL_POINTS, 0, NUM_BOUNDARY_PARTICLES);
    glDisableVertexAttribArray(positionAttribLoc);
}

void renderFluid() {
    glUniform3f(uColorLoc, 0.29f, 0.573f, 1.0f);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);

    glEnableVertexAttribArray(positionAttribLoc);
    glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, GL_FALSE, sizeof(float3), (void*)0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glDrawArrays(GL_POINTS, 0, NUM_PARTICLES);
    glDisableVertexAttribArray(positionAttribLoc);
}

void display() {
    // Bind FBO
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    glViewport(0, 0, fbWidth, fbHeight); // Set viewport to FBO size

    // Set clear color before clearing
    glClearColor(0.9f, 0.9f, 0.9f, 1.0f); // Adjust as needed
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(shaderProgram);
    glDepthMask(GL_TRUE);

    auto currentTime = std::chrono::high_resolution_clock::now();
    deltaTime = std::chrono::duration<float>(currentTime - previousTime).count();
    previousTime = currentTime;

    // Clamp deltaTime to prevent large jumps
    if (deltaTime > 0.1f) deltaTime = 0.1f;
    if (deltaTime < 0.0001f) deltaTime = 0.0001f;

    // Update particles multiple times for stability
    for (int i = 0; i < 4; ++i) { // Adjust iterations as needed
        updateParticles(deltaTime);
    }

    // Update shader uniforms with the new camera matrices
    mat4 projectionMatrix = cameraPtr->getProjectionMatrix();
    glUniformMatrix4fv(projectionMatrixLoc, 1, GL_TRUE, projectionMatrix.m);

    mat4 viewMatrix = cameraPtr->getViewMatrix();
    glUniformMatrix4fv(viewMatrixLoc, 1, GL_TRUE, viewMatrix.m);

    if (cameraPtr->shouldDisplayBorder()) {
        renderBorder();
    }

    renderFluid();

    // Unbind FBO to return to default framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    // Restore viewport to window size
    int winWidth, winHeight;
    glfwGetFramebufferSize(glfwGetCurrentContext(), &winWidth, &winHeight);
    glViewport(0, 0, winWidth, winHeight);

    //printFrameStats();
}

void initializeBoundaryVBO(const std::vector<float3>& h_boundary_positions) {
    glGenBuffers(1, &vbo_boundary);
    glBindBuffer(GL_ARRAY_BUFFER, vbo_boundary);
    glBufferData(GL_ARRAY_BUFFER, h_boundary_positions.size() * sizeof(float3), h_boundary_positions.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    registerBoundaryVBOWithCUDA(vbo_boundary);
}

void initBoundary() {
    std::vector<float3> boundary_positions;
    std::vector<float3> boundary_velocities;
    int boundaryLayers = 3;
    float spacing = h_H/1.5;

    createBoundaryParticles(boundary_positions, boundary_velocities, boundaryLayers, spacing);
    NUM_BOUNDARY_PARTICLES = (int)boundary_positions.size();
    printf("Num Boundary particles: %i\n", NUM_BOUNDARY_PARTICLES);

    initializeBoundryDeviceArrays(boundary_positions.data(), boundary_velocities.data(), NUM_BOUNDARY_PARTICLES);
    initializeBoundaryVBO(boundary_positions);
    boundaryNBSearch(NUM_BOUNDARY_PARTICLES);
}

void initParticleSystem() {
    std::vector<float3> h_positions(NUM_PARTICLES);
    std::vector<float3> h_velocities(NUM_PARTICLES);
    
    float girdSpacing = h_H/1.5;
    initParticles(h_positions, h_velocities, NUM_PARTICLES, girdSpacing);
    initializeFluidDeviceArrays(h_positions.data(), h_velocities.data(), NUM_PARTICLES);

    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, h_positions.size() * sizeof(float3), h_positions.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    registerVBOWithCUDA(vbo);
}

void initShaderUniforms() {
    positionAttribLoc = glGetAttribLocation(shaderProgram, "aPosition");
    if (positionAttribLoc == -1) {
        fprintf(stderr, "Could not find attribute 'aPosition'\n");
    } else {
        printf("Attribute 'aPosition' location: %d\n", positionAttribLoc);
    }

    modelMatrixLoc = glGetUniformLocation(shaderProgram, "modelMatrix");
    viewMatrixLoc = glGetUniformLocation(shaderProgram, "viewMatrix");
    projectionMatrixLoc = glGetUniformLocation(shaderProgram, "projectionMatrix");
    pointRadiusLoc = glGetUniformLocation(shaderProgram, "pointRadius");
    pointScaleLoc = glGetUniformLocation(shaderProgram, "pointScale");
    uColorLoc = glGetUniformLocation(shaderProgram, "uColor");

    if (pointRadiusLoc != -1) glUniform1f(pointRadiusLoc, pointRadius);
    else fprintf(stderr, "Failed to find 'pointRadius' uniform location!\n");
    
    if (pointScaleLoc != -1) glUniform1f(pointScaleLoc, pointScale);
    else fprintf(stderr, "Failed to find 'pointScale' uniform location!\n");

    mat4 modelMatrix = IdentityMatrix();
    glUniformMatrix4fv(modelMatrixLoc, 1, GL_TRUE, modelMatrix.m);
}

// Function to initialize the FBO
void initFBO(int width, int height, GLuint &fbo, GLuint &fboTexture, GLuint &depthRBO) {
    glGenFramebuffers(1, &fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);

    // Create texture to render into
    glGenTextures(1, &fboTexture);
    glBindTexture(GL_TEXTURE_2D, fboTexture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    // Optional: set texture wrapping
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);  
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fboTexture, 0);

    // Create a depth renderbuffer
    glGenRenderbuffers(1, &depthRBO);
    glBindRenderbuffer(GL_RENDERBUFFER, depthRBO);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, width, height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthRBO);

    // Check framebuffer completeness
    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE) {
        fprintf(stderr, "FBO not complete: 0x%X\n", status);
        // Handle error appropriately (e.g., exit or throw exception)
    }

    glBindFramebuffer(GL_FRAMEBUFFER, 0); // Unbind FBO
}

// Function to resize the FBO
void resizeFBO(int newWidth, int newHeight, GLuint &fbo, GLuint &fboTexture, GLuint &depthRBO) {
    if (newWidth <= 0 || newHeight <= 0)
        return;

    // Bind FBO
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);

    // Resize texture
    glBindTexture(GL_TEXTURE_2D, fboTexture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, newWidth, newHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

    // Resize renderbuffer
    glBindRenderbuffer(GL_RENDERBUFFER, depthRBO);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, newWidth, newHeight);

    // Check framebuffer completeness
    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE) {
        fprintf(stderr, "FBO not complete after resize: 0x%X\n", status);
        // Handle error appropriately
    }

    glBindFramebuffer(GL_FRAMEBUFFER, 0); // Unbind FBO
}

void init() {
    initializeConstants();

    glewExperimental = GL_TRUE;
    GLenum err = glewInit();
    if (err != GLEW_OK) {
        fprintf(stderr, "Error initializing GLEW: %s\n", glewGetErrorString(err));
        exit(1);
    }

    // Create and bind a VAO
    GLuint vao;
    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);

    shaderProgram = loadShaders("../shaders/particle.vert", "../shaders/particle.frag");
    glUseProgram(shaderProgram);
    initShaderUniforms();

    cameraPtr = new Camera(
        vec3(5.0f, 15.0f, 20.0f),
        vec3(0.0f, 4.0f, 0.0f),
        vec3(0.0f, 1.0f, 0.0f),
        45.0f,
        800.0f / 600.0f,
        0.1f,
        500.0f
    );

    initCamera(*cameraPtr);
    initParticleSystem();
    initBoundary();

    glEnable(GL_PROGRAM_POINT_SIZE);
    //glEnable(GL_POINT_SPRITE);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);

    initFBO(fbWidth, fbHeight, fbo, fboTexture, depthRBO);
}

void cleanup() {
    unregisterVBOFromCUDA();
    unregisterBoundaryVBOFromCUDA();
    cleanupDeviceArrays();
    glDeleteBuffers(1, &vbo);
    glDeleteBuffers(1, &vbo_boundary);
}

void GLFWKeyCallback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    // Retrieve camera from window user pointer
    Camera* cam = static_cast<Camera*>(glfwGetWindowUserPointer(window));
    if (!cam) return;

    // GLFW keys differ from ASCII keys. For simplicity:
    if (action == GLFW_PRESS || action == GLFW_REPEAT) {
        unsigned char ascii_key;
        // Map GLFW key to ASCII (for WASD keys, this is straightforward)
        switch(key) {
            case GLFW_KEY_W: ascii_key = 'w'; break;
            case GLFW_KEY_S: ascii_key = 's'; break;
            case GLFW_KEY_A: ascii_key = 'a'; break;
            case GLFW_KEY_D: ascii_key = 'd'; break;
            case GLFW_KEY_J: ascii_key = 'j'; break;
            case GLFW_KEY_L: ascii_key = 'l'; break;
            case GLFW_KEY_I: ascii_key = 'i'; break;
            case GLFW_KEY_K: ascii_key = 'k'; break;
            case GLFW_KEY_SPACE: ascii_key = ' '; break;
            case GLFW_KEY_C: ascii_key = 'c'; break;
            case GLFW_KEY_V: ascii_key = 'v'; break;
            case GLFW_KEY_ESCAPE: ascii_key = 0x1b; break;
            default: return; // key not handled
        }

        cam->processKey(ascii_key);
    }
}

void GLFWFramebufferSizeCallback(GLFWwindow* window, int width, int height)
{
    Camera* cam = static_cast<Camera*>(glfwGetWindowUserPointer(window));
    if (!cam) return;

    cam->reshape(width, height);
}


// Just KEEP IMGUI setup in this file. Move simulation/aplication to anothere file.
int main(int argc, char** argv) {
    // Initialize GLFW
    if (!glfwInit()) {
        fprintf(stderr, "Failed to initialize GLFW\n");
        return -1;
    }

    // Set OpenGL version (3.3 Core)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // Create GLFW window
    GLFWwindow* window = glfwCreateWindow(1280, 720, "Simulation with ImGui Docking & FBO", NULL, NULL);
    if (!window) {
        fprintf(stderr, "Failed to create GLFW window\n");
        glfwTerminate();
        return -1;
    }

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1); // Enable V-sync

    // Set initial viewport
    int initialWidth = 1280, initialHeight = 720;
    glViewport(0, 0, initialWidth, initialHeight);

    // Initialize OpenGL, CUDA, shaders, camera, particles, etc.
    init(); // Your existing initialization code

    // Initialize FBO with initial window size (1280x720)
    initFBO(initialWidth, initialHeight, fbo, fboTexture, depthRBO);

    // Instantiate ImGuiManager
    ImGuiManager imguiManager(window, fluidConfig);

    // Set GLFW callbacks
    glfwSetWindowUserPointer(window, cameraPtr);
    glfwSetKeyCallback(window, GLFWKeyCallback);
    glfwSetFramebufferSizeCallback(window, GLFWFramebufferSizeCallback);

    // Main loop
    while (!glfwWindowShouldClose(window)) {
        // Poll events
        glfwPollEvents();

        // Start ImGui frame
        imguiManager.newFrame();

        // Obtain ImGuiIO reference
        ImGuiIO& io = imguiManager.getIO();

        // Create a full-screen dockspace
        ImGui::SetNextWindowPos(ImVec2(0, 0));
        ImGui::SetNextWindowSize(io.DisplaySize);
        ImGuiWindowFlags dockspace_flags = ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoCollapse |
                                           ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove |
                                           ImGuiWindowFlags_NoBringToFrontOnFocus | ImGuiWindowFlags_NoNavFocus;
        
        ImGui::Begin("MainDockSpaceWindow", NULL, dockspace_flags);
        ImGuiID dockspace_id = ImGui::GetID("MainDockSpace");
        ImGui::DockSpace(dockspace_id, ImVec2(0.0f, 0.0f), ImGuiDockNodeFlags_None);
        ImGui::End();

        // Controls window
        imguiManager.createControlsWindow();

        // Frame Stats window
        imguiManager.createFrameStatsWindow();

        // FBO Selection window
        imguiManager.createFBOSelectionWindow();

        // Simulation window
        ImGui::Begin("Simulation");

        // Get available size in Simulation window
        ImVec2 availableSize = ImGui::GetContentRegionAvail();
        int desiredWidth = static_cast<int>(availableSize.x);
        int desiredHeight = static_cast<int>(availableSize.y);

        // Resize FBO if the Simulation window size has changed
        if (desiredWidth > 0 && desiredHeight > 0 && (desiredWidth != fbWidth || desiredHeight != fbHeight)) {
            resizeFBO(desiredWidth, desiredHeight, fbo, fboTexture, depthRBO);
            fbWidth = desiredWidth;
            fbHeight = desiredHeight;

            // Update camera's aspect ratio
            float newAspect = static_cast<float>(fbWidth) / static_cast<float>(fbHeight);
            cameraPtr->setAspectRatio(newAspect);
        }

        // Render to FBO
        display(); // Your existing display() function renders to the current FBO

        // Display the FBO texture, filling the entire window
        ImGui::Image((ImTextureID)(uintptr_t)fboTexture, availableSize, ImVec2(0, 1), ImVec2(1, 0)); // Flip Y-axis
        ImGui::End();

        // Render ImGui
        imguiManager.render();

        // Swap buffers
        glfwSwapBuffers(window);
    }

    // Cleanup ImGui and OpenGL resources is handled by ImGuiManager destructor

    // Cleanup your own resources
    cleanup();
    glDeleteFramebuffers(1, &fbo);
    glDeleteTextures(1, &fboTexture);
    glDeleteRenderbuffers(1, &depthRBO);

    // Terminate GLFW
    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}