// kernels_cuda_pcisph.cu

#include "include/PCISPH_sim.h"
#include "constants.h"
#include "include/float3_helper_math.h"
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <device_launch_parameters.h>
#include <cmath>

// Include Thrust for sorting and reductions
#include <thrust/device_ptr.h>
#include <thrust/sort.h>
#include <thrust/sequence.h>
#include <thrust/reduce.h>

// CUDA Graphics Resource (for OpenGL-CUDA interoperability)
cudaGraphicsResource_t cuda_vbo_resource;
cudaGraphicsResource_t cuda_vbo_boundary_resource;

// Constants
__constant__ float PI;
__constant__ float H;
__constant__ float H2;
__constant__ float POLY6_CONST;
__constant__ float SPIKY_GRAD_CONST;
__constant__ float VISC_LAP_CONST;

__constant__ float PARTICLE_MASS;
__constant__ float REST_DENSITY;
__constant__ float VISCOSITY;

// Boundary constants
__constant__ float BOUND_X_MIN;
__constant__ float BOUND_X_MAX;
__constant__ float BOUND_Y_MIN;
__constant__ float BOUND_Y_MAX;
__constant__ float BOUND_Z_MIN;
__constant__ float BOUND_Z_MAX;

// Device arrays
float3* d_positions;         // Positions mapped from OpenGL VBO
float3* d_velocities;

unsigned int* d_cellIndices;
unsigned int* d_particleIndices;

float3* d_positions_sorted;
float3* d_velocities_sorted;

unsigned int* d_cellStart;
unsigned int* d_cellEnd;

// Arrays for densities, pressures, accelerations
float* d_densities;
float* d_density_errors;
float* d_pressure_increments;
float3* d_pressure_forces;
float3* d_accelerations;

// Additional arrays for PCISPH
float3* d_velocities_pred;
float3* d_positions_pred;
float* d_sum_gradW2; // For computing A

// Boundary particles
float3* d_boundary_positions;
float3* d_boundary_velocities;

unsigned int* d_boundaryCellStart;
unsigned int* d_boundaryCellEnd;

unsigned int* d_boundaryCellIndices;
unsigned int* d_boundaryParticleIndices;

// =================== PCISPH KERNELS =========================

// Smoothing kernels
__device__ float poly6Kernel(float r2)
{
    if (r2 >= 0.0f && r2 <= H2) {
        float term = H2 - r2;
        return POLY6_CONST * term * term * term;
    } else {
        return 0.0f;
    }
}

__device__ float spikyKernelGrad(float r)
{
    if (r > 0.0f && r <= H) {
        float term = H - r;
        return SPIKY_GRAD_CONST * term * term;
    } else {
        return 0.0f;
    }
}

__device__ float viscosityKernelLap(float r)
{
    if (r >= 0.0f && r <= H) {
        return VISC_LAP_CONST * (H - r);
    } else {
        return 0.0f;
    }
}

// Compute cell indices
__global__ void computeCellIndices(
    float3* positions,
    unsigned int* cellIndices,
    int numParticles,
    float3 gridMin,
    float3 gridCellSize,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos = positions[idx];

        int ix = (int)((pos.x - gridMin.x) / gridCellSize.x);
        int iy = (int)((pos.y - gridMin.y) / gridCellSize.y);
        int iz = (int)((pos.z - gridMin.z) / gridCellSize.z);

        // Clamp indices to grid bounds
        ix = max(0, min(ix, gridSizeX - 1));
        iy = max(0, min(iy, gridSizeY - 1));
        iz = max(0, min(iz, gridSizeZ - 1));

        unsigned int cellIndex = ix + iy * gridSizeX + iz * gridSizeX * gridSizeY;

        cellIndices[idx] = cellIndex;
    }
}

// Reordering kernel
__global__ void reorderData(
    unsigned int* particleIndices,
    float3* positions,
    float3* velocities,
    float3* positions_sorted,
    float3* velocities_sorted,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int sortedIdx = particleIndices[idx];

        positions_sorted[idx] = positions[sortedIdx];
        velocities_sorted[idx] = velocities[sortedIdx];
    }
}

// Find cell start and end indices
__global__ void findCellStartEnd(
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int cellIndex = cellIndices[idx];

        if (idx == 0) {
            cellStart[cellIndex] = 0;
        } else {
            unsigned int prevCellIndex = cellIndices[idx - 1];
            if (prevCellIndex != cellIndex) {
                cellStart[cellIndex] = idx;
                cellEnd[prevCellIndex] = idx;
            }
        }

        if (idx == numParticles - 1) {
            cellEnd[cellIndex] = numParticles;
        }
    }
}

// Compute density errors (used in PCISPH)
__global__ void computeDensityError(
    float3* positions,
    float* densities,
    float* density_errors,
    float* sum_gradW2,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos_i = positions[idx];
        float density = 0.0f;
        float sumGradW2_local = 0.0f;

        unsigned int cellIndex = cellIndices[idx];

        // Compute cell coordinates
        int cellX = cellIndex % gridSizeX;
        int cellY = (cellIndex / gridSizeX) % gridSizeY;
        int cellZ = cellIndex / (gridSizeX * gridSizeY);

        // Iterate over neighboring cells
        for (int dz = -1; dz <= 1; ++dz) {
            int nz = cellZ + dz;
            if (nz < 0 || nz >= gridSizeZ) continue;
            for (int dy = -1; dy <= 1; ++dy) {
                int ny = cellY + dy;
                if (ny < 0 || ny >= gridSizeY) continue;
                for (int dx = -1; dx <= 1; ++dx) {
                    int nx = cellX + dx;
                    if (nx < 0 || nx >= gridSizeX) continue;

                    unsigned int neighborCellIndex = nx + ny * gridSizeX + nz * gridSizeX * gridSizeY;

                    unsigned int startIdx = cellStart[neighborCellIndex];
                    unsigned int endIdx = cellEnd[neighborCellIndex];

                    if (startIdx != 0xFFFFFFFF) {
                        for (unsigned int j = startIdx; j < endIdx; ++j) {
                            float3 pos_j = positions[j];

                            float3 r = pos_i - pos_j;
                            float r2 = dot(r, r);

                            if (r2 < H2 && r2 > 0.0001f) {
                                float dist = sqrtf(r2);

                                float W = poly6Kernel(r2);
                                density += PARTICLE_MASS * W;

                                if (sum_gradW2 != nullptr) {
                                    // Compute gradW
                                    float gradW_coeff = spikyKernelGrad(dist);
                                    float3 gradW = gradW_coeff * (r / dist);

                                    sumGradW2_local += dot(gradW, gradW);
                                }
                            }
                        }
                    }
                }
            }
        }

        densities[idx] = density;
        density_errors[idx] = density - REST_DENSITY;

        if (sum_gradW2 != nullptr) {
            sum_gradW2[idx] = sumGradW2_local;
        }
    }
}

// Update pressures based on density errors
__global__ void updatePressures(
    float* pressure_increments,
    float* density_errors,
    float A,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float deltaP = -density_errors[idx] / A;
        pressure_increments[idx] += deltaP;
    }
}

// Compute pressure forces
__global__ void computePressureForces(
    float3* positions,
    float* pressure_increments,
    float3* pressure_forces,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos_i = positions[idx];
        float deltaP_i = pressure_increments[idx];

        float3 fPressure = make_float3(0.0f, 0.0f, 0.0f);

        unsigned int cellIndex = cellIndices[idx];

        // Compute cell coordinates
        int cellX = cellIndex % gridSizeX;
        int cellY = (cellIndex / gridSizeX) % gridSizeY;
        int cellZ = cellIndex / (gridSizeX * gridSizeY);

        // Iterate over neighboring cells
        for (int dz = -1; dz <= 1; ++dz) {
            int nz = cellZ + dz;
            if (nz < 0 || nz >= gridSizeZ) continue;
            for (int dy = -1; dy <= 1; ++dy) {
                int ny = cellY + dy;
                if (ny < 0 || ny >= gridSizeY) continue;
                for (int dx = -1; dx <= 1; ++dx) {
                    int nx = cellX + dx;
                    if (nx < 0 || nx >= gridSizeX) continue;

                    unsigned int neighborCellIndex = nx + ny * gridSizeX + nz * gridSizeX * gridSizeY;

                    unsigned int startIdx = cellStart[neighborCellIndex];
                    unsigned int endIdx = cellEnd[neighborCellIndex];

                    if (startIdx != 0xFFFFFFFF) {
                        for (unsigned int j = startIdx; j < endIdx; ++j) {
                            if (j == idx) continue;

                            float3 pos_j = positions[j];
                            float deltaP_j = pressure_increments[j];

                            float3 r = pos_i - pos_j;

                            float distSq = dot(r, r);

                            if (distSq < H2 && distSq > 0.0001f) {
                                float dist = sqrtf(distSq);

                                // Compute gradient of kernel
                                float gradW = spikyKernelGrad(dist);
                                float3 gradWij = gradW * (r / dist);

                                // Symmetric pressure force
                                float factor = - (deltaP_i + deltaP_j) / (REST_DENSITY * REST_DENSITY);
                                fPressure += factor * gradWij;
                            }
                        }
                    }
                }
            }
        }

        // Multiply by constants
        fPressure *= PARTICLE_MASS * PARTICLE_MASS;

        pressure_forces[idx] = fPressure;
    }
}

// Compute viscosity and external forces (excluding pressure forces)
__global__ void computeViscosityAndExternalForces(
    float3* positions,
    float3* velocities,
    float3* accelerations,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos_i = positions[idx];
        float3 vel_i = velocities[idx];

        float3 fViscosity = make_float3(0.0f, 0.0f, 0.0f);
        float3 fGravity = make_float3(0.0f, -20.0f, 0.0f) * PARTICLE_MASS;

        unsigned int cellIndex = cellIndices[idx];

        // Compute cell coordinates
        int cellX = cellIndex % gridSizeX;
        int cellY = (cellIndex / gridSizeX) % gridSizeY;
        int cellZ = cellIndex / (gridSizeX * gridSizeY);

        // Iterate over neighboring cells
        for (int dz = -1; dz <= 1; ++dz) {
            int nz = cellZ + dz;
            if (nz < 0 || nz >= gridSizeZ) continue;
            for (int dy = -1; dy <= 1; ++dy) {
                int ny = cellY + dy;
                if (ny < 0 || ny >= gridSizeY) continue;
                for (int dx = -1; dx <= 1; ++dx) {
                    int nx = cellX + dx;
                    if (nx < 0 || nx >= gridSizeX) continue;

                    unsigned int neighborCellIndex = nx + ny * gridSizeX + nz * gridSizeX * gridSizeY;

                    unsigned int startIdx = cellStart[neighborCellIndex];
                    unsigned int endIdx = cellEnd[neighborCellIndex];

                    if (startIdx != 0xFFFFFFFF) {
                        for (unsigned int j = startIdx; j < endIdx; ++j) {
                            if (j == idx) continue;

                            float3 pos_j = positions[j];
                            float3 vel_j = velocities[j];
                            float3 r = pos_i - pos_j;

                            float distSq = dot(r, r);

                            if (distSq < H2 && distSq > 0.0001f) {
                                float dist = sqrtf(distSq);

                                // Viscosity force
                                float3 velDiff = vel_j - vel_i;
                                float viscLap = viscosityKernelLap(dist);
                                fViscosity += VISCOSITY * velDiff * viscLap * PARTICLE_MASS / REST_DENSITY;
                            }
                        }
                    }
                }
            }
        }

        // Total acceleration (excluding pressure force)
        accelerations[idx] = (fViscosity + fGravity) / PARTICLE_MASS;
    }
}

// Enforce boundary conditions, very simple should bee improved.
__global__ void enforceBoundary(
    float3* positions,
    float3* velocities,
    int numParticles,
    float damping
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // X-axis
        if (positions[idx].x < BOUND_X_MIN) {
            positions[idx].x = BOUND_X_MIN;
            velocities[idx].x *= -damping;
        }
        if (positions[idx].x > BOUND_X_MAX) {
            positions[idx].x = BOUND_X_MAX;
            velocities[idx].x *= -damping;
        }

        // Y-axis
        if (positions[idx].y < BOUND_Y_MIN) {
            positions[idx].y = BOUND_Y_MIN;
            velocities[idx].y *= -damping;
        }
        if (positions[idx].y > BOUND_Y_MAX) {
            positions[idx].y = BOUND_Y_MAX;
            velocities[idx].y *= -damping;
        }

        // Z-axis
        if (positions[idx].z < BOUND_Z_MIN) {
            positions[idx].z = BOUND_Z_MIN;
            velocities[idx].z *= -damping;
        }
        if (positions[idx].z > BOUND_Z_MAX) {
            positions[idx].z = BOUND_Z_MAX;
            velocities[idx].z *= -damping;
        }
    }
}

// Kernel to predict velocities and positions
__global__ void predictVelocitiesPositions(
    float3* velocities,
    float3* velocities_pred,
    float3* positions,
    float3* positions_pred,
    float3* accelerations,
    float timeStep,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < numParticles) {
        // Compute predicted velocity
        velocities_pred[idx] = velocities[idx] + accelerations[idx] * timeStep;
        // Compute predicted position
        positions_pred[idx] = positions[idx] + velocities_pred[idx] * timeStep;
    }
}

// Update velocities and positions based on pressure forces in 
// pressure correcting loop. 
__global__ void updateVelocitiesPositions(
    float3* velocities_pred,
    float3* positions_pred,
    float3* pressure_forces,
    float timeStep,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // Update predicted velocity
        velocities_pred[idx] += (pressure_forces[idx] / PARTICLE_MASS) * timeStep;

        // Update predicted position
        positions_pred[idx] += (pressure_forces[idx] / PARTICLE_MASS) * (timeStep * timeStep);
    }
}

void neigbourHoodSearch(
        int nParticles, 
        float3* d_pos,
        float3* d_vel,
        float3* d_posSorted,
        float3* d_velSorted,
        unsigned int* d_cellIndc,
        unsigned int* d_parIndc,
        unsigned int* d_cEnd,
        unsigned int* d_cStart,
        const int bpg,
        const int tpb,
        int &gridSizeX,
        int &gridSizeY,
        int &gridSizeZ) 
{

    // Compute grid size based on smoothing radius
    float3 gridMin = make_float3(h_BOUND_X_MIN, h_BOUND_Y_MIN, h_BOUND_Z_MIN);
    float3 gridMax = make_float3(h_BOUND_X_MAX, h_BOUND_Y_MAX, h_BOUND_Z_MAX);
    float3 gridCellSize = make_float3(h_H, h_H, h_H);

    gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;
    
    // Compute cell indices for current positions
    computeCellIndices<<<bpg, tpb>>>(
        d_pos, d_cellIndc, nParticles, gridMin, gridCellSize, gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Wrap device pointers with thrust iterators
    thrust::device_ptr<unsigned int> dev_cellIndices_ptr(d_cellIndc);
    thrust::device_ptr<unsigned int> dev_particleIndices_ptr(d_parIndc);

    // Initialize particle indices
    thrust::sequence(dev_particleIndices_ptr, dev_particleIndices_ptr + nParticles);

    // Sort particle indices based on cell indices
    thrust::sort_by_key(dev_cellIndices_ptr, dev_cellIndices_ptr + nParticles, dev_particleIndices_ptr);

    // Reorder positions and velocities
    reorderData<<<bpg, tpb>>>(
        d_parIndc,
        d_pos,
        d_vel,
        d_posSorted,
        d_velSorted,
        nParticles);
    cudaDeviceSynchronize();

    // Copy sorted data back to original arrays
    cudaMemcpy(d_pos, d_posSorted, nParticles * sizeof(float3), cudaMemcpyDeviceToDevice);
    cudaMemcpy(d_vel, d_velSorted, nParticles * sizeof(float3), cudaMemcpyDeviceToDevice);

    // Build cell start and end indices
    cudaMemset(d_cStart, 0xFF, maxCells * sizeof(unsigned int));
    cudaMemset(d_cEnd, 0xFF, maxCells * sizeof(unsigned int));

    findCellStartEnd<<<bpg, tpb>>>(
        d_cellIndc,
        d_cStart,
        d_cEnd,
        nParticles);
    cudaDeviceSynchronize();
}

void updateParticles(float deltaTime)
{
    // Map the VBO resource to get a device pointer
    size_t num_bytes;
    cudaGraphicsMapResources(1, &cuda_vbo_resource, 0);
    cudaGraphicsResourceGetMappedPointer((void**)&d_positions, &num_bytes, cuda_vbo_resource);

    int threadsPerBlock = 256;
    int blocksPerGrid = (NUM_PARTICLES + threadsPerBlock - 1) / threadsPerBlock;


    // Time steep should be rather small but we can run thee updatParticles
    // function Multiple times dureeing our display func. 
    // e.g timeStep =0.004 for 4 times = 0.016 ms per framee -> 60 fps
    float timeStep = 0.004f; 


    // -------------------- Neighbor Search --------------------
    // We only do nb-search onnce peer step since we assume particles are unlikeely to
    // Change positions enough to motivat doing nb-search when itterating in pressure
    // Correcting loop
    int gridSizeX, gridSizeY, gridSizeZ;
    
    neigbourHoodSearch(
            NUM_PARTICLES, 
            d_positions,
            d_velocities,
            d_positions_sorted,
            d_velocities_sorted,
            d_cellIndices,
            d_particleIndices,
            d_cellEnd,
            d_cellStart,
            blocksPerGrid,
            threadsPerBlock,
            gridSizeX,
            gridSizeY,
            gridSizeZ); 

    // -------------------- Prediction Step --------------------

    // Compute viscosity and external forces (excluding pressure)
    computeViscosityAndExternalForces<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        d_accelerations,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Predict velocities and positions
    predictVelocitiesPositions<<<blocksPerGrid, threadsPerBlock>>>(
        d_velocities,
        d_velocities_pred,
        d_positions,
        d_positions_pred,
        d_accelerations,
        timeStep,
        NUM_PARTICLES);
    cudaDeviceSynchronize();

    // Initialize pressure increments to zero
    cudaMemset(d_pressure_increments, 0, NUM_PARTICLES * sizeof(float));


    // Compute densities and density errors, and per-particle sumGradW2
    computeDensityError<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions_pred,
        d_densities,
        d_density_errors,
        d_sum_gradW2,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Reduce sum_gradW2 over all particles
    thrust::device_ptr<float> dev_sumGradW2_ptr(d_sum_gradW2);
    float totalSumGradW2 = thrust::reduce(dev_sumGradW2_ptr, dev_sumGradW2_ptr + NUM_PARTICLES, 0.0f, thrust::plus<float>());

    // Compute A
    float m = h_PARTICLE_MASS;
    float rho0 = h_REST_DENSITY;
    float A = 2.0f * timeStep * timeStep * m * m / (rho0 * rho0) * totalSumGradW2;

    // ------------ Iterative Pressure Correction Loop --------------------

    const int maxIterations = 5;

    for (int iter = 0; iter < maxIterations; ++iter) {
        // Compute densities and density errors
        computeDensityError<<<blocksPerGrid, threadsPerBlock>>>(
            d_positions_pred,
            d_densities,
            d_density_errors,
            nullptr, // No need to compute sum_gradW2 again
            d_cellIndices,
            d_cellStart,
            d_cellEnd,
            NUM_PARTICLES,
            gridSizeX, gridSizeY, gridSizeZ);
        cudaDeviceSynchronize();

        // Update pressure 
        updatePressures<<<blocksPerGrid, threadsPerBlock>>>(
            d_pressure_increments,
            d_density_errors,
            A,
            NUM_PARTICLES);
        cudaDeviceSynchronize();

        // Compute pressure forces using pressure increments
        computePressureForces<<<blocksPerGrid, threadsPerBlock>>>(
            d_positions_pred,
            d_pressure_increments,
            d_pressure_forces,
            d_cellIndices,
            d_cellStart,
            d_cellEnd,
            NUM_PARTICLES,
            gridSizeX, gridSizeY, gridSizeZ);
        cudaDeviceSynchronize();

        // Update predicted velocities and positions
        updateVelocitiesPositions<<<blocksPerGrid, threadsPerBlock>>>(
            d_velocities_pred,
            d_positions_pred,
            d_pressure_forces,
            timeStep,
            NUM_PARTICLES);
        cudaDeviceSynchronize();

        // Could impleement check to break loop if dens_err < threshold
    }

    // -------------------- Finalization --------------------

    // Copy predicted velocities and positions to actual velocities and positions
    cudaMemcpy(d_velocities, d_velocities_pred, NUM_PARTICLES * sizeof(float3), cudaMemcpyDeviceToDevice);
    cudaMemcpy(d_positions, d_positions_pred, NUM_PARTICLES * sizeof(float3), cudaMemcpyDeviceToDevice);

    // Enforce boundary conditions
    float damping = 0.75f;
    enforceBoundary<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        NUM_PARTICLES,
        damping);
    cudaDeviceSynchronize();

    // Unmap the VBO resource
    cudaGraphicsUnmapResources(1, &cuda_vbo_resource, 0);
}

void initializeConstants()
{
    // Host constants
    const float h_H2 = h_H * h_H;
    const float h_POLY6_CONST = 315.0f / (64.0f * h_PI * powf(h_H, 9));
    const float h_SPIKY_GRAD_CONST = -45.0f / (h_PI * powf(h_H, 6));
    const float h_VISC_LAP_CONST = 45.0f / (h_PI * powf(h_H, 6));

    // Copy constants to device
    cudaMemcpyToSymbol(PI, &h_PI, sizeof(float));
    cudaMemcpyToSymbol(H, &h_H, sizeof(float));
    cudaMemcpyToSymbol(H2, &h_H2, sizeof(float));
    cudaMemcpyToSymbol(POLY6_CONST, &h_POLY6_CONST, sizeof(float));
    cudaMemcpyToSymbol(SPIKY_GRAD_CONST, &h_SPIKY_GRAD_CONST, sizeof(float));
    cudaMemcpyToSymbol(VISC_LAP_CONST, &h_VISC_LAP_CONST, sizeof(float));

    cudaMemcpyToSymbol(PARTICLE_MASS, &h_PARTICLE_MASS, sizeof(float));
    cudaMemcpyToSymbol(REST_DENSITY, &h_REST_DENSITY, sizeof(float));
    cudaMemcpyToSymbol(VISCOSITY, &h_VISCOSITY, sizeof(float));

    // Copy boundary constants to device
    cudaMemcpyToSymbol(BOUND_X_MIN, &h_BOUND_X_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_X_MAX, &h_BOUND_X_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MIN, &h_BOUND_Y_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MAX, &h_BOUND_Y_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MIN, &h_BOUND_Z_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MAX, &h_BOUND_Z_MAX, sizeof(float));
}

// Register VBO with CUDA
void registerVBOWithCUDA(unsigned int vbo)
{
    cudaGraphicsGLRegisterBuffer(&cuda_vbo_resource, vbo, cudaGraphicsMapFlagsNone);
}

void registerBoundaryVBOWithCUDA(unsigned int vbo)
{
    cudaGraphicsGLRegisterBuffer(&cuda_vbo_boundary_resource, vbo, cudaGraphicsMapFlagsNone);
}

// Unregister VBO from CUDA
void unregisterVBOFromCUDA()
{
    cudaGraphicsUnregisterResource(cuda_vbo_resource);
}

void unregisterBoundaryVBOFromCUDA()
{
    cudaGraphicsUnregisterResource(cuda_vbo_boundary_resource);
}

// Initialize device arrays
void initializeFluidDeviceArrays(
    const float3* h_positions,
    const float3* h_velocities,
    int numParticles
)
{
    // Allocate device memory for velocities
    cudaMalloc(&d_velocities, numParticles * sizeof(float3));
    cudaMemcpy(d_velocities, h_velocities, numParticles * sizeof(float3), cudaMemcpyHostToDevice);

    // Allocate device memory for predicted velocities and positions
    cudaMalloc(&d_velocities_pred, numParticles * sizeof(float3));
    cudaMalloc(&d_positions_pred, numParticles * sizeof(float3));

    // Allocate device memory for cell indices and particle indices
    cudaMalloc(&d_cellIndices, numParticles * sizeof(unsigned int));
    cudaMalloc(&d_particleIndices, numParticles * sizeof(unsigned int));

    // Allocate device memory for sorted positions and velocities
    cudaMalloc(&d_positions_sorted, numParticles * sizeof(float3));
    cudaMalloc(&d_velocities_sorted, numParticles * sizeof(float3));

    // Allocate cell start and end arrays
    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    cudaMalloc(&d_cellStart, maxCells * sizeof(unsigned int));
    cudaMalloc(&d_cellEnd, maxCells * sizeof(unsigned int));

    // Allocate device memory for densities, pressure increments, and accelerations
    cudaMalloc(&d_densities, numParticles * sizeof(float));
    cudaMalloc(&d_density_errors, numParticles * sizeof(float));
    cudaMalloc(&d_pressure_increments, numParticles * sizeof(float));
    cudaMalloc(&d_pressure_forces, numParticles * sizeof(float3));
    cudaMalloc(&d_accelerations, numParticles * sizeof(float3));

    // Allocate device memory for sum_gradW2
    cudaMalloc(&d_sum_gradW2, numParticles * sizeof(float));

    // Initialize positions_pred to initial positions
    cudaMemcpy(d_positions_pred, h_positions, numParticles * sizeof(float3), cudaMemcpyHostToDevice);
}

void initializeBoundryDeviceArrays(
    const float3* h_boundaryPositions,
    const float3* h_boundaryVelocities,
    int numBoundaryParticles
)
{
    // Allocate cell start and end arrays
    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    // cudaMalloc(&d_boundary_positions, numBoundaryParticles * sizeof(float3));
    cudaMalloc(&d_boundary_velocities, numBoundaryParticles * sizeof(float3));

    //cudaMemcpy(d_boundary_positions, h_boundaryPositions, numBoundaryParticles * sizeof(float3), cudaMemcpyHostToDevice);
    cudaMemcpy(d_boundary_velocities, h_boundaryVelocities, numBoundaryParticles * sizeof(float3), cudaMemcpyHostToDevice);

    cudaMalloc(&d_boundaryCellIndices, numBoundaryParticles * sizeof(unsigned int));
    cudaMalloc(&d_boundaryParticleIndices, numBoundaryParticles * sizeof(unsigned int));
    
    cudaMalloc(&d_boundaryCellStart, maxCells * sizeof(unsigned int));
    cudaMalloc(&d_boundaryCellEnd, maxCells * sizeof(unsigned int));
}


// NeighbourHoodSearch for boundary particles 1 time at start since they dont move
void boundaryNBSearch(int numBoundaryParticles) {
    
    // Map the VBO resource to get a device pointer
    size_t num_bytes;
    cudaGraphicsMapResources(1, &cuda_vbo_boundary_resource, 0);
    cudaGraphicsResourceGetMappedPointer((void**)&d_boundary_positions, &num_bytes, cuda_vbo_boundary_resource);
    
    int gsX, gsY, gsZ;

    int threadsPerBlock = 256;
    int blocksPerGrid = (numBoundaryParticles + threadsPerBlock - 1) / threadsPerBlock;
    
    float3* d_boundary_positions_sorted;
    float3* d_boundary_velocities_sorted;
    

    cudaMalloc(&d_boundary_positions_sorted, numBoundaryParticles * sizeof(float3));
    cudaMalloc(&d_boundary_velocities_sorted, numBoundaryParticles * sizeof(float3));
    
    neigbourHoodSearch(
            numBoundaryParticles, 
            d_boundary_positions,
            d_boundary_velocities,
            d_boundary_positions_sorted,
            d_boundary_velocities_sorted,
            d_boundaryCellIndices,
            d_boundaryParticleIndices,
            d_boundaryCellEnd,
            d_boundaryCellStart,
            blocksPerGrid,
            threadsPerBlock,
            gsX,
            gsY,
            gsZ); 
    
    // Unmap the VBO resource
    cudaGraphicsUnmapResources(1, &cuda_vbo_boundary_resource, 0);
    
    cudaFree(d_boundary_positions_sorted);
    cudaFree(d_boundary_velocities_sorted);
    
    // Move some of above init code for boundary to seeparate function
}

// Cleanup device arrays
void cleanupDeviceArrays()
{
    cudaFree(d_velocities);
    cudaFree(d_velocities_pred);
    cudaFree(d_positions_pred);

    cudaFree(d_cellIndices);
    cudaFree(d_particleIndices);

    cudaFree(d_positions_sorted);
    cudaFree(d_velocities_sorted);

    cudaFree(d_cellStart);
    cudaFree(d_cellEnd);

    cudaFree(d_densities);
    cudaFree(d_density_errors);
    cudaFree(d_pressure_increments);
    cudaFree(d_pressure_forces);
    cudaFree(d_accelerations);

    cudaFree(d_boundary_positions);
    cudaFree(d_boundary_velocities);

    cudaFree(d_boundaryCellIndices);
    cudaFree(d_boundaryParticleIndices);
    cudaFree(d_boundaryCellEnd);
    cudaFree(d_boundaryCellStart);

    cudaFree(d_sum_gradW2);
}

