// kernels_cuda.cu
#include "kernels_cuda.h"
#include "constants.h"
#include "include/float3_helper_math.h"
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <device_launch_parameters.h>
#include <cmath>

// Include Thrust for sorting
#include <thrust/device_ptr.h>
#include <thrust/sort.h>
#include <thrust/sequence.h>

// CUDA Graphics Resource (for OpenGL-CUDA interoperability)
cudaGraphicsResource_t cuda_vbo_resource;

// Constants
__constant__ float PI;
__constant__ float H;
__constant__ float H2;
__constant__ float POLY6_CONST;
__constant__ float SPIKY_GRAD_CONST;
__constant__ float VISC_LAP_CONST;

__constant__ float PARTICLE_MASS = 1.0f;
__constant__ float REST_DENSITY = 60.0f;
__constant__ float GAS_CONSTANT;
__constant__ float VISCOSITY;

// Boundary constants
__constant__ float BOUND_X_MIN;
__constant__ float BOUND_X_MAX;
__constant__ float BOUND_Y_MIN;
__constant__ float BOUND_Y_MAX;
__constant__ float BOUND_Z_MIN;
__constant__ float BOUND_Z_MAX;

// Device arrays
float3* d_positions; // Positions mapped from OpenGL VBO
float3* d_velocities;

unsigned int* d_cellIndices;
unsigned int* d_particleIndices;

float3* d_positions_sorted;
float3* d_velocities_sorted;

unsigned int* d_cellStart;
unsigned int* d_cellEnd;

// Arrays for densities, pressures, accelerations
float* d_densities;
float* d_pressures;
float3* d_accelerations;

// Smoothing kernels
__device__ float poly6Kernel(float r2)
{
    if (r2 >= 0.0f && r2 <= H2) {
        float term = H2 - r2;
        return POLY6_CONST * term * term * term;
    } else {
        return 0.0f;
    }
}

__device__ float spikyKernelGrad(float r)
{
    if (r > 0.0f && r <= H) {
        float term = H - r;
        return SPIKY_GRAD_CONST * term * term;
    } else {
        return 0.0f;
    }
}

__device__ float viscosityKernelLap(float r)
{
    if (r >= 0.0f && r <= H) {
        return VISC_LAP_CONST * (H - r);
    } else {
        return 0.0f;
    }
}

// Compute cell indices
__global__ void computeCellIndices(
    float3* positions,
    unsigned int* cellIndices,
    int numParticles,
    float3 gridMin,
    float3 gridCellSize,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos = positions[idx];

        int ix = (int)((pos.x - gridMin.x) / gridCellSize.x);
        int iy = (int)((pos.y - gridMin.y) / gridCellSize.y);
        int iz = (int)((pos.z - gridMin.z) / gridCellSize.z);

        // Clamp indices to grid bounds
        ix = max(0, min(ix, gridSizeX - 1));
        iy = max(0, min(iy, gridSizeY - 1));
        iz = max(0, min(iz, gridSizeZ - 1));

        unsigned int cellIndex = ix + iy * gridSizeX + iz * gridSizeX * gridSizeY;

        cellIndices[idx] = cellIndex;
    }
}

// Reordering kernel
__global__ void reorderData(
    unsigned int* particleIndices,
    float3* positions,
    float3* velocities,
    float3* positions_sorted,
    float3* velocities_sorted,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int sortedIdx = particleIndices[idx];

        positions_sorted[idx] = positions[sortedIdx];
        velocities_sorted[idx] = velocities[sortedIdx];
    }
}

// Find cell start and end indices
// can bli reeplaceed med atomicAdd för varje ceell index??
__global__ void findCellStartEnd(
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int cellIndex = cellIndices[idx];

        if (idx == 0) {
            cellStart[cellIndex] = 0;
        } else {
            unsigned int prevCellIndex = cellIndices[idx - 1];
            if (prevCellIndex != cellIndex) {
                cellStart[cellIndex] = idx;
                cellEnd[prevCellIndex] = idx;
            }
        }

        if (idx == numParticles - 1) {
            cellEnd[cellIndex] = numParticles;
        }
    }
}

// Compute density and pressure
__global__ void computeDensityPressure(
    float3* positions,
    float* densities,
    float* pressures,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos_i = positions[idx];
        float density = 0.0f;

        unsigned int cellIndex = cellIndices[idx];

        // Compute cell coordinates
        int cellX = cellIndex % gridSizeX;
        int cellY = (cellIndex / gridSizeX) % gridSizeY;
        int cellZ = cellIndex / (gridSizeX * gridSizeY);

        // Iterate over neighboring cells
        for (int dz = -1; dz <= 1; ++dz) {
            int nz = cellZ + dz;
            if (nz < 0 || nz >= gridSizeZ) continue;
            for (int dy = -1; dy <= 1; ++dy) {
                int ny = cellY + dy;
                if (ny < 0 || ny >= gridSizeY) continue;
                for (int dx = -1; dx <= 1; ++dx) {
                    int nx = cellX + dx;
                    if (nx < 0 || nx >= gridSizeX) continue;

                    unsigned int neighborCellIndex = nx + ny * gridSizeX + nz * gridSizeX * gridSizeY;

                    unsigned int startIdx = cellStart[neighborCellIndex];
                    unsigned int endIdx = cellEnd[neighborCellIndex];

                    if (startIdx != 0xFFFFFFFF) {
                        for (unsigned int j = startIdx; j < endIdx; ++j) {
                            float3 pos_j = positions[j];

                            float3 r = pos_j - pos_i;
                            float r2 = r.x * r.x + r.y * r.y + r.z * r.z;

                            density += PARTICLE_MASS * poly6Kernel(r2);
                        }
                    }
                }
            }
        }

        densities[idx] = density;
        //pressures[idx] = GAS_CONSTANT * fmaxf((density - REST_DENSITY), 0.0);
        float dens = density/REST_DENSITY;
        pressures[idx] = GAS_CONSTANT * ((dens*dens*dens*dens*dens*dens*dens)-1.0);
        //pressures[idx] = GAS_CONSTANT * min((density - REST_DENSITY), 0.0);
        //pressures[idx] = GAS_CONSTANT * (density - REST_DENSITY);
    }
}

// Compute forces
__global__ void computeForces(
    float3* positions,
    float3* velocities,
    float* densities,
    float* pressures,
    float3* accelerations,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos_i = positions[idx];
        float3 vel_i = velocities[idx];
        float density_i = densities[idx];
        float pressure_i = pressures[idx];

        float3 fPressure = make_float3(0.0f, 0.0f, 0.0f);
        float3 fViscosity = make_float3(0.0f, 0.0f, 0.0f);
        float3 fGravity = make_float3(0.0f, -4.0f , 0.0f);

        unsigned int cellIndex = cellIndices[idx];

        // Compute cell coordinates
        int cellX = cellIndex % gridSizeX;
        int cellY = (cellIndex / gridSizeX) % gridSizeY;
        int cellZ = cellIndex / (gridSizeX * gridSizeY);

        // Iterate over neighboring cells
        for (int dz = -1; dz <= 1; ++dz) {
            int nz = cellZ + dz;
            if (nz < 0 || nz >= gridSizeZ) continue;
            for (int dy = -1; dy <= 1; ++dy) {
                int ny = cellY + dy;
                if (ny < 0 || ny >= gridSizeY) continue;
                for (int dx = -1; dx <= 1; ++dx) {
                    int nx = cellX + dx;
                    if (nx < 0 || nx >= gridSizeX) continue;

                    unsigned int neighborCellIndex = nx + ny * gridSizeX + nz * gridSizeX * gridSizeY;

                    unsigned int startIdx = cellStart[neighborCellIndex];
                    unsigned int endIdx = cellEnd[neighborCellIndex];

                    if (startIdx != 0xFFFFFFFF) {
                        for (unsigned int j = startIdx; j < endIdx; ++j) {
                            if (j == idx) continue;

                            float3 pos_j = positions[j];
                            float3 r = (pos_i - pos_j);
                            
                            float distSq = r.x * r.x + r.y * r.y + r.z * r.z;

                            if (distSq < H2 && distSq > 0.0001f) {
                                float dist = sqrtf(distSq);

                                // Pressure force
                                float spikyGrad = spikyKernelGrad(dist);

                                // Cache pressures[j]
                                float density_j = densities[j];
                                float pressure_j = pressures[j];

                                // avreged pressure scaleed by inverse j density
                                /*
                                float pressureTerm = - PARTICLE_MASS * 0.5 * (pressure_i + pressure_j) / density_j * spikyGrad;
                                fPressure += pressureTerm*(r/dist); // r/dist gives direction
                                */

                                // Symmetric pressure -- can cahnge to inverse dennsity in density/pressure calc to avoid devide by density here
                                // Add particle_mass to pressure term according to equations !!! Added down below in aceleration calcualtion
                                float pressureTerm = - (pressure_i / (density_i * density_i) + pressure_j / (density_j * density_j)) * spikyGrad;
                                float3 pressureForce = pressureTerm * (r / dist); // Directional force
                                fPressure += pressureForce;

                                // Viscosity force
                                float3 velDiff = velocities[j] - vel_i;
                                float viscLap = viscosityKernelLap(dist);
                                fViscosity  += VISCOSITY * velDiff * viscLap / density_j * PARTICLE_MASS;
                            }
                        }
                    }
                }
            }
        }

        // Total acceleration
        // According to paper v_i(t+dt) = v_(i+1) + (dt/m_i)*Fp_i
        // While external and viscocisty force is calculated before But mass and nnot deeensity is used to deevidee force
        // Deviding by mass, looks a lot better
        accelerations[idx] = (fPressure * PARTICLE_MASS + fViscosity + fGravity) / PARTICLE_MASS;
        //accelerations[idx] = (fPressure * PARTICLE_MASS + fViscosity + fGravity) / density_i;
    }
}

// Integration
__global__ void integrate(
    float3* positions,
    float3* velocities,
    float3* accelerations,
    float timeStep,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // Update velocity
        float3 vel_i = velocities[idx] + (accelerations[idx] * timeStep);
        velocities[idx] = vel_i;

        // Update position
        positions[idx] += vel_i * timeStep;
    }
}

// Enforce boundary conditions
__global__ void enforceBoundary(
    float3* positions,
    float3* velocities,
    int numParticles,
    float damping
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // X-axis
        if (positions[idx].x < BOUND_X_MIN) {
            positions[idx].x = BOUND_X_MIN;
            velocities[idx].x *= -damping;
        }
        if (positions[idx].x > BOUND_X_MAX) {
            positions[idx].x = BOUND_X_MAX; // Should account for particle radius!!
            velocities[idx].x *= -damping;
        }

        // Y-axis
        if (positions[idx].y < BOUND_Y_MIN) {
            positions[idx].y = BOUND_Y_MIN;
            velocities[idx].y *= -damping;
        }
        if (positions[idx].y > BOUND_Y_MAX) {
            positions[idx].y = BOUND_Y_MAX;
            velocities[idx].y *= -damping;
        }

        // Z-axis
        if (positions[idx].z < BOUND_Z_MIN) {
            positions[idx].z = BOUND_Z_MIN;
            velocities[idx].z *= -damping;
        }
        if (positions[idx].z > BOUND_Z_MAX) {
            positions[idx].z = BOUND_Z_MAX;
            velocities[idx].z *= -damping;
        }
    }
}

void updateParticles(float deltaTime)
{
    // Map the VBO resource to get a device pointer
    size_t num_bytes;
    cudaGraphicsMapResources(1, &cuda_vbo_resource, 0);
    cudaGraphicsResourceGetMappedPointer((void**)&d_positions, &num_bytes, cuda_vbo_resource);

    int threadsPerBlock = 256;
    int blocksPerGrid = (NUM_PARTICLES + threadsPerBlock - 1) / threadsPerBlock;

    // Compute grid size based on smoothing radius
    float3 gridMin = make_float3(h_BOUND_X_MIN, h_BOUND_Y_MIN, h_BOUND_Z_MIN);
    float3 gridMax = make_float3(h_BOUND_X_MAX, h_BOUND_Y_MAX, h_BOUND_Z_MAX);
    float3 gridCellSize = make_float3(h_H, h_H, h_H);

    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    // Allocate or resize cellStart and cellEnd arrays if necessary
    // (You may need to adjust your initialization code to handle dynamic sizes)

    // Compute cell indices
    computeCellIndices<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions, d_cellIndices, NUM_PARTICLES, gridMin, gridCellSize, gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Wrap device pointers with thrust iterators
    thrust::device_ptr<unsigned int> dev_cellIndices_ptr(d_cellIndices);
    thrust::device_ptr<unsigned int> dev_particleIndices_ptr(d_particleIndices);

    // Initialize particle indices
    thrust::sequence(dev_particleIndices_ptr, dev_particleIndices_ptr + NUM_PARTICLES);

    // Sort particle indices based on cell indices
    thrust::sort_by_key(dev_cellIndices_ptr, dev_cellIndices_ptr + NUM_PARTICLES, dev_particleIndices_ptr);

    // Reorder positions and velocities
    reorderData<<<blocksPerGrid, threadsPerBlock>>>(
        d_particleIndices,
        d_positions,
        d_velocities,
        d_positions_sorted,
        d_velocities_sorted,
        NUM_PARTICLES);
    cudaDeviceSynchronize();

    // Swap sorted data with original data
    float3* temp_positions = d_positions;
    d_positions = d_positions_sorted;
    d_positions_sorted = temp_positions;

    float3* temp_velocities = d_velocities;
    d_velocities = d_velocities_sorted;
    d_velocities_sorted = temp_velocities;

    // Build cell start and end indices
    // Initialize cellStart and cellEnd
    cudaMemset(d_cellStart, 0xFF, maxCells * sizeof(unsigned int));
    cudaMemset(d_cellEnd, 0xFF, maxCells * sizeof(unsigned int));

    findCellStartEnd<<<blocksPerGrid, threadsPerBlock>>>(
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES);
    cudaDeviceSynchronize();

    // Compute density and pressure
    computeDensityPressure<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_densities,
        d_pressures,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Compute forces
    computeForces<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        d_densities,
        d_pressures,
        d_accelerations,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Integrate
    float timeStep = 0.001f;
    integrate<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        d_accelerations,
        timeStep,
        //deltaTime,
        NUM_PARTICLES);
    cudaDeviceSynchronize();

    // Enforce boundary conditions
    float damping = 0.7f;
    enforceBoundary<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        NUM_PARTICLES,
        damping);
    cudaDeviceSynchronize();

    // Unmap the VBO resource
    cudaGraphicsUnmapResources(1, &cuda_vbo_resource, 0);
}

void initializeConstants()
{
    // Host constants
    const float h_H2 = h_H * h_H;
    const float h_POLY6_CONST = 315.0f / (64.0f * h_PI * powf(h_H, 9));
    const float h_SPIKY_GRAD_CONST = -45.0f / (h_PI * powf(h_H, 6));
    const float h_VISC_LAP_CONST = 45.0f / (h_PI * powf(h_H, 6));

    // Copy constants to device
    cudaMemcpyToSymbol(PI, &h_PI, sizeof(float));
    cudaMemcpyToSymbol(H, &h_H, sizeof(float));
    cudaMemcpyToSymbol(H2, &h_H2, sizeof(float));
    cudaMemcpyToSymbol(POLY6_CONST, &h_POLY6_CONST, sizeof(float));
    cudaMemcpyToSymbol(SPIKY_GRAD_CONST, &h_SPIKY_GRAD_CONST, sizeof(float));
    cudaMemcpyToSymbol(VISC_LAP_CONST, &h_VISC_LAP_CONST, sizeof(float));

    cudaMemcpyToSymbol(GAS_CONSTANT, &h_GAS_CONSTANT, sizeof(float));
    cudaMemcpyToSymbol(VISCOSITY, &h_VISCOSITY, sizeof(float));

    // Copy boundary constants to device
    cudaMemcpyToSymbol(BOUND_X_MIN, &h_BOUND_X_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_X_MAX, &h_BOUND_X_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MIN, &h_BOUND_Y_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MAX, &h_BOUND_Y_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MIN, &h_BOUND_Z_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MAX, &h_BOUND_Z_MAX, sizeof(float));
}

// Register VBO with CUDA
void registerVBOWithCUDA(unsigned int vbo)
{
    cudaGraphicsGLRegisterBuffer(&cuda_vbo_resource, vbo, cudaGraphicsMapFlagsNone);
}

// Unregister VBO from CUDA
void unregisterVBOFromCUDA()
{
    cudaGraphicsUnregisterResource(cuda_vbo_resource);
}

// Initialize device arrays
void initializeDeviceArrays(
    const float3* h_positions,
    const float3* h_velocities,
    int numParticles
)
{
    // Allocate device memory for velocities
    cudaMalloc(&d_velocities, numParticles * sizeof(float3));
    cudaMemcpy(d_velocities, h_velocities, numParticles * sizeof(float3), cudaMemcpyHostToDevice);

    // Allocate device memory for cell indices and particle indices
    cudaMalloc(&d_cellIndices, numParticles * sizeof(unsigned int));
    cudaMalloc(&d_particleIndices, numParticles * sizeof(unsigned int));

    // Allocate device memory for sorted positions and velocities
    cudaMalloc(&d_positions_sorted, numParticles * sizeof(float3));
    cudaMalloc(&d_velocities_sorted, numParticles * sizeof(float3));

    // Allocate cell start and end arrays
    // Note: You need to ensure that 'maxCells' is available here or adjust the code to allocate after computing 'maxCells'
    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    cudaMalloc(&d_cellStart, maxCells * sizeof(unsigned int));
    cudaMalloc(&d_cellEnd, maxCells * sizeof(unsigned int));

    // Allocate device memory for densities, pressures, and accelerations
    cudaMalloc(&d_densities, numParticles * sizeof(float));
    cudaMalloc(&d_pressures, numParticles * sizeof(float));
    cudaMalloc(&d_accelerations, numParticles * sizeof(float3));
}

// Cleanup device arrays
void cleanupDeviceArrays()
{
    cudaFree(d_velocities);

    cudaFree(d_cellIndices);
    cudaFree(d_particleIndices);

    cudaFree(d_positions_sorted);
    cudaFree(d_velocities_sorted);

    cudaFree(d_cellStart);
    cudaFree(d_cellEnd);

    cudaFree(d_densities);
    cudaFree(d_pressures);
    cudaFree(d_accelerations);
}