// main.cpp: This example uses FreeGLUT for OpenGL and CUDA interoperation.
// It creates a simple particle simulation and uses CUDA to update particle positions with gravity.
// main.cpp
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath> // For std::pow
#include <chrono>
#include <cstdlib> // For exit()

//#include <GL/glew.h>
//#include <GL/freeglut.h>
#include "MicroGlut.h"
#include "GL_utilities.h"

#include "include/PCISPH_sim.h" // Header for CUDA function declarations
#include "spawn_particles.h"
#include "include/Camera.h"
#include "constants.h"

#define MAIN
#include "VectorUtils4.h"

// Number of particles
const int NUM_PARTICLES = 50000; // Should match a cubic grid for simplicity
const float GRID_SPACING = 0.25f;  // Spacing between particles in the grid
int NUM_BOUNDARY_PARTICLES = 0;

const float pointRadius = 20.0f;  // Point size in world space
const float pointScale = 10.0f;

auto previousTime = std::chrono::high_resolution_clock::now();
float deltaTime = 0.016f;


// Shader and VBO handles
GLuint shaderProgram;
GLuint positionAttribLoc;
// Vertex Buffer Object (VBO) for particle positions
GLuint vbo;
GLuint vbo_boundary;

// Uniform locations
GLuint modelMatrixLoc;
GLuint viewMatrixLoc;
GLuint projectionMatrixLoc;
GLuint pointRadiusLoc;
GLuint pointScaleLoc;
GLuint uColorLoc;

// CAMEERA VARIABLES
Camera* cameraPtr = nullptr;

void printFrameStats() {
    // FPS and deltaTime accumulation
    static float timeAccumulator = 0.0f;
    static int frameCount = 0;

    timeAccumulator += deltaTime;
    frameCount++;

    // Print FPS and average deltaTime every second
    if (timeAccumulator >= 1.0f)
    {
        float averageDeltaTime = timeAccumulator / frameCount;
        float fps = frameCount / timeAccumulator;

        printf("FPS: %.2f, Average DeltaTime: %.1f ms\n", fps, averageDeltaTime*1000);

        // Reset accumulators
        timeAccumulator = 0.0f;
        frameCount = 0;
    }
}

void renderBorder()
{
    // === Render Boundary Particles ===
    // Set color for boundary particles (e.g., red)
    glUniform3f(uColorLoc, 0.29f, 1.0f, 0.57f); // Red color

    // Bind boundary VBO and set up vertex attribute
    glBindBuffer(GL_ARRAY_BUFFER, vbo_boundary);
    glEnableVertexAttribArray(positionAttribLoc);
    glVertexAttribPointer(
        positionAttribLoc,  // Attribute location
        3,                  // Number of components (x, y, z)
        GL_FLOAT,           // Data type
        GL_FALSE,           // Normalized
        sizeof(float3),     // Stride (size of one position)
        (void*)0            // Offset to the position data
    );

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    // Draw boundary particles as points
    glDrawArrays(GL_POINTS, 0, NUM_BOUNDARY_PARTICLES);

    // Disable vertex attribute
    glDisableVertexAttribArray(positionAttribLoc);
}


void renderFluid()
{
    // === Render Fluid Particles ===
    // Set color for fluid particles (e.g., blue)
    glUniform3f(uColorLoc, 0.29, 0.573, 1.0); // Blue color

    // Bind fluid VBO and set up vertex attribute
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glEnableVertexAttribArray(positionAttribLoc);
    glVertexAttribPointer(
        positionAttribLoc,  // Attribute location
        3,                  // Number of components (x, y, z)
        GL_FLOAT,           // Data type
        GL_FALSE,           // Normalized
        sizeof(float3),     // Stride (size of one position)
        (void*)0            // Offset to the position data
    );

    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Draw fluid particles as points
    glDrawArrays(GL_POINTS, 0, NUM_PARTICLES);

    // Disable vertex attribute
    glDisableVertexAttribArray(positionAttribLoc);
}


// Display callback for OpenGL
void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable(GL_BLEND);
    glDepthMask(GL_TRUE);
    auto currentTime = std::chrono::high_resolution_clock::now();
    deltaTime = std::chrono::duration<float>(currentTime - previousTime).count();
    previousTime = currentTime;

     // Clamp deltaTime to prevent instability
    if (deltaTime > 0.1f) // Max 100 ms
        deltaTime = 0.1f;
    if (deltaTime < 0.0001f) // Min 0.1 ms
        deltaTime = 0.0001f;

    // Update particle positions using CUDA
    for(int i = 0; i < 4; ++i) {
        updateParticles(deltaTime);
    }

    // Use the shader program
    glUseProgram(shaderProgram);

    // Update the projection and view matrices based on the camera
    mat4 projectionMatrix = cameraPtr->getProjectionMatrix();
    glUniformMatrix4fv(projectionMatrixLoc, 1, GL_TRUE, projectionMatrix.m);

    mat4 viewMatrix = cameraPtr->getViewMatrix();
    glUniformMatrix4fv(viewMatrixLoc, 1, GL_TRUE, viewMatrix.m);

    // Render boundary particles if enabled
    if(cameraPtr->shouldDisplayBorder()) {
        renderBorder();
    }

    renderFluid();
    
    // Swap buffers
    glutSwapBuffers();

    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);

    // Print frame statistics
    printFrameStats();
}


// Function to initialize and register boundary VBO
void initializeBoundaryVBO(const std::vector<float3>& h_boundary_positions) {
    // Create VBO for boundary particles
    glGenBuffers(1, &vbo_boundary);
    glBindBuffer(GL_ARRAY_BUFFER, vbo_boundary);
    glBufferData(GL_ARRAY_BUFFER, h_boundary_positions.size() * sizeof(float3), h_boundary_positions.data(), GL_DYNAMIC_DRAW); // Static since boundaries don't move
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Register boundary VBO with CUDA
    registerBoundaryVBOWithCUDA(vbo_boundary);
}

void initBoundary()
{
    std::vector<float3> boundary_positions;
    std::vector<float3> boundary_velocities;
    int boundaryLayers = 3;
    float spacing = h_H/1.5;

    createBoundaryParticles(boundary_positions, boundary_velocities, boundaryLayers, spacing);

    NUM_BOUNDARY_PARTICLES = (int)boundary_positions.size();
    printf("Num Bounndnary particles: %i \n", NUM_BOUNDARY_PARTICLES);
    // Initialize device arrays
    initializeBoundryDeviceArrays(
        boundary_positions.data(),
        boundary_velocities.data(),
        NUM_BOUNDARY_PARTICLES
    );

    initializeBoundaryVBO(boundary_positions);

    boundaryNBSearch(NUM_BOUNDARY_PARTICLES);
}

void initParticleSystem()
{
    // Particle data (initialized to 0)
    std::vector<float3> h_positions(NUM_PARTICLES);
    std::vector<float3> h_velocities(NUM_PARTICLES);
    
    float girdSpacing = h_H/1.5;
    // Initialize particle data
    initParticles(
        h_positions,
        h_velocities,
        NUM_PARTICLES,
        girdSpacing
    );
    
    initializeFluidDeviceArrays(
        h_positions.data(),
        h_velocities.data(),
        NUM_PARTICLES
    );

    // Initialize VBO for particle positions
    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    // Upload initial positions to the VBO
    glBufferData(GL_ARRAY_BUFFER, h_positions.size() * sizeof(float3), h_positions.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Register VBO with CUDA
    registerVBOWithCUDA(vbo);
}

void initShaderUniforms()
{
    // Query the attribute location for position
    positionAttribLoc = glGetAttribLocation(shaderProgram, "aPosition");
    if (positionAttribLoc == -1) {
        fprintf(stderr, "Could not find attribute 'aPosition'\n");
    } else {
        printf("Attribute 'aPosition' location: %d\n", positionAttribLoc);
    }

    // Camera uniform locs
    modelMatrixLoc = glGetUniformLocation(shaderProgram, "modelMatrix");
    viewMatrixLoc = glGetUniformLocation(shaderProgram, "viewMatrix");
    projectionMatrixLoc = glGetUniformLocation(shaderProgram, "projectionMatrix");
    
    // Other uniform locs
    pointRadiusLoc = glGetUniformLocation(shaderProgram, "pointRadius");
    pointScaleLoc = glGetUniformLocation(shaderProgram, "pointScale");
    uColorLoc = glGetUniformLocation(shaderProgram, "uColor");


    // Set point size uniforms
    if (pointRadiusLoc != -1) {
        glUniform1f(pointRadiusLoc, pointRadius);
    } else {
        fprintf(stderr, "Failed to find 'pointRadius' uniform location!\n");
    }

    if (pointScaleLoc != -1) {
        glUniform1f(pointScaleLoc, pointScale);
    } else {
        fprintf(stderr, "Failed to find 'pointScale' uniform location!\n");
    }

    // Initialize model matrix (to Identity matrix)
    mat4 modelMatrix = IdentityMatrix();
    glUniformMatrix4fv(modelMatrixLoc, 1, GL_TRUE, modelMatrix.m);
}

// Initialize OpenGL and CUDA
void init()
{
    // Initialize constants in CUDA
    initializeConstants();

    // Set glewExperimental to enable modern OpenGL features
    //glewExperimental = GL_TRUE;

    // Initialize GLEW
    GLenum err = glewInit();
    if (err != GLEW_OK) {
        fprintf(stderr, "Error initializing GLEW: %s\n", glewGetErrorString(err));
        exit(1);
    }

    // Load and compile shaders
    shaderProgram = loadShaders("../../shaders/particle.vert", "../../shaders/particle.frag");
    glUseProgram(shaderProgram);

    initShaderUniforms();

    // Create Camera
    cameraPtr = new Camera(
        vec3(5.0f, 15.0f, 20.0f),  // Position
        vec3(0.0f, 4.0f, 0.0f),    // Target
        vec3(0.0f, 1.0f, 0.0f),    // Up vector
        45.0f,                      // FOV
        800.0f / 600.0f,            // Aspect ratio
        0.1f,                       // zNear
        500.0f                      // zFar
    );

    initCamera(*cameraPtr);
    
    initParticleSystem();
    initBoundary();

    // Enable point size for rendering particles
    glEnable(GL_PROGRAM_POINT_SIZE);
    glEnable(GL_POINT_SPRITE);

    // Enable depth testing
    glEnable(GL_DEPTH_TEST);
}

// Cleanup resources
void cleanup()
{
    unregisterVBOFromCUDA();
    unregisterBoundaryVBOFromCUDA();
    cleanupDeviceArrays();
    glDeleteBuffers(1, &vbo);
    glDeleteBuffers(1, &vbo_boundary);
}


int main(int argc, char** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB /*| GLUT_DEPTH*/);
    glutInitWindowSize(800, 600);
    //glutInitContextVersion(3, 2);
    glutCreateWindow("CUDA-OpenGL Particle Simulation");

    glClearColor(0.9f, 0.9f, 0.9f, 1.0f);

    // Initialize OpenGL and CUDA

    glutDisplayFunc(display);
    glutIdleFunc(display);

    // Cleanup on exit
    atexit(cleanup);
    dumpInfo(); 

    init();

    glutMainLoop();
    return 0;
}
