// main.cpp: This example uses FreeGLUT for OpenGL and CUDA interoperation.
// It creates a simple particle simulation and uses CUDA to update particle positions with gravity.
// main.cpp
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath> // For std::pow
#include <chrono>
#include <cstdlib> // For exit()

#include <GL/glew.h>
#include <GL/freeglut.h>

#include "GL_utilities.h"
#include "PCISPH_sim.h" // Header for CUDA function declarations

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp> // For transformation functions
#include <glm/gtc/type_ptr.hpp>

#define MAIN
#include "GL_utilities.h"

// Number of particles
const int NUM_PARTICLES =25600; // Should match a cubic grid for simplicity
const float GRID_SPACING = 0.25f;  // Spacing between particles in the grid

const float pointRadius = 16.0f;  // Point size in world space
const float pointScale = 10.0f;

auto previousTime = std::chrono::high_resolution_clock::now();
float deltaTime = 0.016f;

// Particle data (initialized to 0)
std::vector<float3> h_positions(NUM_PARTICLES);
std::vector<float3> h_velocities(NUM_PARTICLES);
// Acceleration, density, and pressure will be recalculated each time; no need to store them on host

// Shader and VBO handles
GLuint shaderProgram;
GLuint positionAttribLoc;
// Vertex Buffer Object (VBO) for particle positions
GLuint vbo;

// Uniform locations
GLuint modelMatrixLoc;
GLuint viewMatrixLoc;
GLuint projectionMatrixLoc;
GLuint pointRadiusLoc;
GLuint pointScaleLoc;
GLuint sphereRadiusLoc;
GLuint lightDirLoc;

void printMatrix(const glm::mat4& mat, const char* name)
{
    printf("%s:\n", name);
    const float* m = (const float*)glm::value_ptr(mat);
    for (int i = 0; i < 4; ++i)
    {
        printf("%f %f %f %f\n", m[i], m[i + 4], m[i + 8], m[i + 12]);
    }
}

void printFrameStats() {
    // FPS and deltaTime accumulation
    static float timeAccumulator = 0.0f;
    static int frameCount = 0;

    timeAccumulator += deltaTime;
    frameCount++;

    // Print FPS and average deltaTime every second
    if (timeAccumulator >= 1.0f)
    {
        float averageDeltaTime = timeAccumulator / frameCount;
        float fps = frameCount / timeAccumulator;

        printf("FPS: %.2f, Average DeltaTime: %.1f ms\n", fps, averageDeltaTime*1000);

        // Reset accumulators
        timeAccumulator = 0.0f;
        frameCount = 0;
    }
}

// Display callback for OpenGL
void display()
{

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
   
    auto currentTime = std::chrono::high_resolution_clock::now();
    deltaTime = std::chrono::duration<float>(currentTime - previousTime).count();
    previousTime = currentTime;

     // Clamp deltaTime to prevent instability
    if (deltaTime > 0.1f) // Max 100 ms
        deltaTime = 0.1f;
    if (deltaTime < 0.0001f) // Min 0.1 ms
        deltaTime = 0.0001f;

    // Update particle positions using CUDA
    updateParticles(deltaTime);

    // Use the shader program
    glUseProgram(shaderProgram);

    // Bind VBO and set up vertex attribute for position
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glEnableVertexAttribArray(positionAttribLoc);
    glVertexAttribPointer(
        positionAttribLoc,  // Attribute location
        3,                  // Number of components (x, y, z)
        GL_FLOAT,           // Data type
        GL_FALSE,           // Normalized
        sizeof(float3),     // Stride (size of one position)
        (void*)0            // Offset to the position data
    );

    // Draw particles as points
    glDrawArrays(GL_POINTS, 0, NUM_PARTICLES);

    // Disable vertex attribute
    glDisableVertexAttribArray(positionAttribLoc);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glutSwapBuffers();

    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);

    printFrameStats();

}

void initParticles()
{
    // Seed the random number generator
    srand(static_cast<unsigned int>(time(0)));

    // Calculate particles per axis
    int particlesPerAxis = static_cast<int>(ceil(pow(NUM_PARTICLES, 1.0f / 3.0f)));

    int index = 0;
    for (int i = 0; i < particlesPerAxis && index < NUM_PARTICLES; ++i) {
        for (int j = 0; j < particlesPerAxis && index < NUM_PARTICLES; ++j) {
            for (int k = 0; k < particlesPerAxis && index < NUM_PARTICLES; ++k) {
                // Center particles around the origin
                float x = (i - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING;
                float y = (j - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING + 2.0f;
                float z = (k - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING;

                // Add random perturbation within [-0.1*dx, 0.1*dx]
                float perturbation = 0.1f * GRID_SPACING;
                float rand_x = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                float rand_y = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                float rand_z = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;

                h_positions[index] = make_float3(
                    x + rand_x,
                    y + rand_y,
                    z + rand_z
                );

                // Initialize velocities
                h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f);

                ++index;
            }
        }
    }

    // If NUM_PARTICLES is not a perfect cube, initialize remaining particles
    for (; index < NUM_PARTICLES; ++index) {
        h_positions[index] = make_float3(0.0f, 10.0f, 0.0f); // Default position
        h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f); // Default velocity
    }
}

// Initialize OpenGL and CUDA
void initialize()
{
    // Initialize constants in CUDA
    initializeConstants();

    // Set glewExperimental to enable modern OpenGL features
    glewExperimental = GL_TRUE;

    // Initialize GLEW
    GLenum err = glewInit();
    if (err != GLEW_OK) {
        fprintf(stderr, "Error initializing GLEW: %s\n", glewGetErrorString(err));
        exit(1);
    }

    // Load and compile shaders
    shaderProgram = loadShaders("../../shaders/depth.vert", "../../shaders/depth.frag");
    glUseProgram(shaderProgram);

    // Query the attribute location for position
    positionAttribLoc = glGetAttribLocation(shaderProgram, "aPosition");
    if (positionAttribLoc == -1) {
        fprintf(stderr, "Could not find attribute 'aPosition'\n");
    } else {
        printf("Attribute 'aPosition' location: %d\n", positionAttribLoc);
    }

    // Camera setup
    glm::vec3 eyePosition(-22.0f, 8.0f, 0.0f);
    glm::vec3 targetPosition(12.0f, 4.0f, 0.0f);
    glm::vec3 upVector(0.0f, 1.0f, 0.0f);

    // Create View Matrix using glm::lookAt
    glm::mat4 viewMatrix = glm::lookAt(eyePosition, targetPosition, upVector);

    // Create Projection Matrix using glm::perspective
    float fov = 45.0f; // Field of view in degrees
    float aspectRatio = 800.0f / 600.0f; // Window width / height
    float zNear = 0.1f;
    float zFar = 500.0f;
    glm::mat4 projectionMatrix = glm::perspective(glm::radians(fov), aspectRatio, zNear, zFar);

    // Model Matrix (Identity matrix)
    glm::mat4 modelMatrix = glm::mat4(1.0f);

    // Query uniform locations
    modelMatrixLoc = glGetUniformLocation(shaderProgram, "modelMatrix");
    viewMatrixLoc = glGetUniformLocation(shaderProgram, "viewMatrix");
    projectionMatrixLoc = glGetUniformLocation(shaderProgram, "projectionMatrix");
    pointRadiusLoc = glGetUniformLocation(shaderProgram, "pointRadius");
    pointScaleLoc = glGetUniformLocation(shaderProgram, "pointScale");
    sphereRadiusLoc = glGetUniformLocation(shaderProgram, "sphereRadius");
    lightDirLoc = glGetUniformLocation(shaderProgram, "lightDir");
    
    // Print matrices for debugging
    printMatrix(modelMatrix, "Model Matrix");
    printMatrix(viewMatrix, "View Matrix");
    printMatrix(projectionMatrix, "Projection Matrix");

    // Set uniform matrices
    glUniformMatrix4fv(modelMatrixLoc, 1, GL_FALSE, glm::value_ptr(modelMatrix));
    glUniformMatrix4fv(viewMatrixLoc, 1, GL_FALSE, glm::value_ptr(viewMatrix));
    glUniformMatrix4fv(projectionMatrixLoc, 1, GL_FALSE, glm::value_ptr(projectionMatrix));

    // Set point size uniforms
    if (pointRadiusLoc != -1) {
        glUniform1f(pointRadiusLoc, pointRadius);
    } else {
        fprintf(stderr, "Failed to find 'pointRadius' uniform location!\n");
    }

    if (pointScaleLoc != -1) {
        glUniform1f(pointScaleLoc, pointScale);
    } else {
        fprintf(stderr, "Failed to find 'pointScale' uniform location!\n");
    }

    if (sphereRadiusLoc != -1) {
        glUniform1f(sphereRadiusLoc, pointRadius); // Use the same value as pointRadius
    } else {
        fprintf(stderr, "Failed to find 'sphereRadius' uniform location!\n");
    }

    // Define light direction in world space
    glm::vec3 lightDirWorldSpace(-5.0f, 5.0f, 0.0f);

    // Transform light direction to eye space
    glm::vec3 lightDirEyeSpace = glm::vec3(glm::normalize(viewMatrix * glm::vec4(lightDirWorldSpace, 0.0f)));

    // Set the uniform
    if (lightDirLoc != -1) {
        glUniform3fv(lightDirLoc, 1, glm::value_ptr(lightDirEyeSpace));
    } else {
        fprintf(stderr, "Failed to find 'lightDir' uniform location!\n");
    }


    // Initialize particle data
    initParticles();

    // Initialize VBO for particle positions
    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    // Upload initial positions to the VBO
    glBufferData(GL_ARRAY_BUFFER, h_positions.size() * sizeof(float3), h_positions.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Register VBO with CUDA
    registerVBOWithCUDA(vbo);

    // Initialize device arrays
    initializeDeviceArrays(
        h_positions.data(),
        h_velocities.data(),
        NUM_PARTICLES
    );

    // Enable point size for rendering particles
    glEnable(GL_PROGRAM_POINT_SIZE);
    glEnable(GL_POINT_SPRITE);

    // Enable depth testing
    glEnable(GL_DEPTH_TEST);
}

// Cleanup resources
void cleanup()
{
    unregisterVBOFromCUDA();
    cleanupDeviceArrays();
    glDeleteBuffers(1, &vbo);
}

int main(int argc, char** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB /*| GLUT_DEPTH*/);
    glutInitWindowSize(800, 600);
    glutCreateWindow("CUDA-OpenGL Particle Simulation");
    glClearColor(0.9f, 0.9f, 0.9f, 1.0f);

    // Initialize OpenGL and CUDA
    initialize();

    // Register display callback
    glutDisplayFunc(display);
    glutIdleFunc(display);

    // Cleanup on exit
    atexit(cleanup);

    // Start GLUT main loop
    glutMainLoop();

    return 0;
}
