// spawn_particles.cpp
#include "spawn_particles.h"
#include "constants.h"

#include <cmath>          // For std::pow
#include <ctime>
#include <vector>
#include <cstdlib>

// Initialize fluid particles
void initParticles( 
    std::vector<float3> &h_positions,
    std::vector<float3> &h_velocities,
    const int NUM_PARTICLES,
    const float GRID_SPACING)
{
    // Seed the random number generator
    srand(static_cast<unsigned int>(time(0)));

    // Calculate particles per axis
    int particlesPerAxis = static_cast<int>(ceil(pow(NUM_PARTICLES, 1.0f / 3.0f)));

    int index = 0;
    for (int i = 0; i < particlesPerAxis && index < NUM_PARTICLES; ++i) {
        for (int j = 0; j < particlesPerAxis && index < NUM_PARTICLES; ++j) {
            for (int k = 0; k < particlesPerAxis && index < NUM_PARTICLES; ++k) {
                // Center particles around the origin
                float x = (i - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING;
                float y = (j - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING + 2.0f;
                float z = (k - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING;

                // Add random perturbation within [-0.1*dx, 0.1*dx]
                float perturbation = 0.1f * GRID_SPACING;
                float rand_x = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                float rand_y = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                float rand_z = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;

                h_positions[index] = make_float3(
                    x + rand_x,
                    y + rand_y,
                    z + rand_z
                );

                // Initialize velocities
                h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f);

                ++index;
            }
        }
    }

    // If NUM_PARTICLES is not a perfect cube, initialize remaining particles
    for (; index < NUM_PARTICLES; ++index) {
        h_positions[index] = make_float3(0.0f, 10.0f, 0.0f); // Default position
        h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f); // Default velocity
    }
}

// Create boundary particles
void createBoundaryParticles(
    std::vector<float3> &boundary_positions, 
    std::vector<float3> &boundary_velocities,
    int boundaryLayers, 
    float spacing) 
{
    // Compute the number of particles along each dimension for a given spacing
    int nx = static_cast<int>((h_BOUND_X_MAX - h_BOUND_X_MIN)/spacing) + 1;
    int ny = static_cast<int>((h_BOUND_Y_MAX - h_BOUND_Y_MIN)/spacing) + 1;
    int nz = static_cast<int>((h_BOUND_Z_MAX - h_BOUND_Z_MIN)/spacing) + 1;

    // Initialize velocities to zero for boundary
    float3 zeroVel = make_float3(0.0f, 0.0f, 0.0f);

    // Y-min boundary (e.g., "floor")
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float y = h_BOUND_Y_MIN - (layer+1)*spacing; // placing layers below Ymin
        for (int ix = 0; ix < nx; ix++) {
            for (int iz = 0; iz < nz; iz++) {
                float x = h_BOUND_X_MIN + ix*spacing;
                float z = h_BOUND_Z_MIN + iz*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }

    /*
    // Y-max boundary (ceiling)
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float y = h_BOUND_Y_MAX + (layer+1)*spacing;
        for (int ix = 0; ix < nx; ix++) {
            for (int iz = 0; iz < nz; iz++) {
                float x = h_BOUND_X_MIN + ix*spacing;
                float z = h_BOUND_Z_MIN + iz*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }
    */

    // X-min boundary (vertical wall on the left)
    int ny_cells = ny; // from earlier computation
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float x = h_BOUND_X_MIN - (layer+1)*spacing;
        for (int iy = 0; iy < ny_cells; iy++) {
            for (int iz = 0; iz < nz; iz++) {
                float y = h_BOUND_Y_MIN + iy*spacing;
                float z = h_BOUND_Z_MIN + iz*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }

    // X-max boundary (vertical wall on the right)
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float x = h_BOUND_X_MAX + (layer+1)*spacing;
        for (int iy = 0; iy < ny_cells; iy++) {
            for (int iz = 0; iz < nz; iz++) {
                float y = h_BOUND_Y_MIN + iy*spacing;
                float z = h_BOUND_Z_MIN + iz*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }

    // Z-min boundary (front wall)
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float z = h_BOUND_Z_MIN - (layer+1)*spacing;
        for (int ix = 0; ix < nx; ix++) {
            for (int iy = 0; iy < ny_cells; iy++) {
                float x = h_BOUND_X_MIN + ix*spacing;
                float y = h_BOUND_Y_MIN + iy*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }

    // Z-max boundary (back wall)
    for (int layer = 0; layer < boundaryLayers; layer++) {
        float z = h_BOUND_Z_MAX + (layer+1)*spacing;
        for (int ix = 0; ix < nx; ix++) {
            for (int iy = 0; iy < ny_cells; iy++) {
                float x = h_BOUND_X_MIN + ix*spacing;
                float y = h_BOUND_Y_MIN + iy*spacing;
                boundary_positions.push_back(make_float3(x, y, z));
                boundary_velocities.push_back(zeroVel);
            }
        }
    }
}
