#version 330 core

in vec2 TexCoords;
out vec4 FragColor;

uniform sampler2D uDepthTex;
uniform float sigma_r;
uniform float depthThreshold;
uniform vec2 passDirection;  // (1.0/width, 0) or (0, 1.0/height)

// Bilateral parameters
uniform float zNear;
uniform float zFar;
uniform float cameraFovY;
uniform int   screenHeight;

// World-space smoothing
uniform float worldRadiusMeters;
uniform float maxPixelRadius; // clamp so we don't do huge loops

uniform float sigmaSpatialFactor; //  if we still want a gaussWeights fallback

void main()
{
    // 1) Read the center depth [0..1], convert to eye-dist
    float centerDepthN = texture(uDepthTex, TexCoords).r;
    if (centerDepthN <= 0.0 || centerDepthN >= 1.0) {
        // out of range, maybe skip
        discard;
    }
    float eyeDist = mix(zNear, zFar, centerDepthN);

    // 2) Convert to screen-space pixel radius
    float planeHeight = 2.0 * eyeDist * tan(cameraFovY * 0.5);
    float pxPerMeter  = float(screenHeight) / planeHeight;
    float pixelRadius = worldRadiusMeters * pxPerMeter;
    pixelRadius = min(pixelRadius, maxPixelRadius);

    int iRadius = int(ceil(pixelRadius));

    float centerDepth = centerDepthN; // if our depth is actually linear, that's fine
    float sum = 0.0;
    float wSum = 0.0;

    // Loop from -maxPixelKernel..+maxPixelKernel
    int maxPixelKernel = 64; // or something bigger, but 64 seems good enough
    for (int i = -maxPixelKernel; i <= maxPixelKernel; i++) 
    {
        if (abs(i) > iRadius) {
            continue;
        }

        // offset
        vec2 offset = passDirection * float(i);
        vec2 sampleUV = TexCoords + offset;

        float neighborDepthN = texture(uDepthTex, sampleUV).r; 
        if (neighborDepthN <= 0.0 || neighborDepthN >= 1.0) {
            continue;
        }
        float d = abs(neighborDepthN - centerDepth);

        // range check
        if (d > depthThreshold) {
            continue;
        }

        // Spatial weight (some 1D Gaussian). 
        // Compute on-the-fly: 
        // e.g. float spatial = exp(-(i*i)/(2.0*sigmaSpatialFactor*sigmaSpatialFactor));
        // could be replaced by passinng in a small look-up array
        float wSpatial = exp(- (float(i)*float(i)) / (2.0 * sigmaSpatialFactor * sigmaSpatialFactor));

        // Range weight
        float wRange = exp(- (d*d) / (2.0 * sigma_r * sigma_r));

        float w = wSpatial * wRange;
        sum += neighborDepthN * w;
        wSum += w;
    }

    float blurredDepth = sum / max(wSum, 1e-5);
    FragColor = vec4(blurredDepth, 0, 0, 1);
}
