#version 330 core

in vec2 TexCoords;
out vec4 FragColor;

uniform sampler2D uDepthTex;

// We need these to convert linear [0..1] to actual distance
uniform float zNear;
uniform float zFar;

// The inverse of the same projection matrix used when rendering depth
uniform mat4 invProjectionMatrix;

// We can discard if depth is bigger than some clamp
uniform float maxDepth;

// Pixel-size step for neighbor sampling
uniform vec2 uTexelSize;

vec3 reconstructEyePos(vec2 uv, float depthLin)
{
    float dEye = depthLin * (zFar - zNear) + zNear;

    float xNDC = uv.x * 2.0 - 1.0;
    float yNDC = uv.y * 2.0 - 1.0;

    vec4 clipPos  = vec4(xNDC, yNDC, -1.0, 1.0);
    vec4 eyePos4  = invProjectionMatrix * clipPos;
    // perspective divide
    vec3 dir = eyePos4.xyz / eyePos4.w;
    dir = normalize(dir);

    return dir * dEye;
}

void main()
{
    float depthLin = texture(uDepthTex, TexCoords).r;
    if (depthLin < 0.0 || depthLin > 1.0 || depthLin > maxDepth) {
        discard;
    }

    vec3 centerEye = reconstructEyePos(TexCoords, depthLin);

    vec2 uvLeft  = TexCoords + vec2(-uTexelSize.x, 0.0);
    vec2 uvRight = TexCoords + vec2( uTexelSize.x, 0.0);

    float depthL = texture(uDepthTex, uvLeft).r;
    float depthR = texture(uDepthTex, uvRight).r;

    if (depthL < 0.0 || depthL > 1.0 || depthL > maxDepth) discard;
    if (depthR < 0.0 || depthR > 1.0 || depthR > maxDepth) discard;

    vec3 leftEye  = reconstructEyePos(uvLeft,  depthL);
    vec3 rightEye = reconstructEyePos(uvRight, depthR);

    vec3 ddx  = rightEye - centerEye;
    vec3 ddx2 = centerEye - leftEye;

    if (abs(ddx2.z) < abs(ddx.z)) {
        ddx = ddx2;
    }

    vec2 uvUp = TexCoords + vec2(0.0,  uTexelSize.y);
    vec2 uvDn = TexCoords + vec2(0.0, -uTexelSize.y);

    float depthU = texture(uDepthTex, uvUp).r;
    float depthD = texture(uDepthTex, uvDn).r;

    if (depthU < 0.0 || depthU > 1.0 || depthU > maxDepth) discard;
    if (depthD < 0.0 || depthD > 1.0 || depthD > maxDepth) discard;

    vec3 upEye = reconstructEyePos(uvUp, depthU);
    vec3 dnEye = reconstructEyePos(uvDn, depthD);

    vec3 ddy  = upEye - centerEye;
    vec3 ddy2 = centerEye - dnEye;
    if (abs(ddy2.z) < abs(ddy.z)) {
        ddy = ddy2;
    }

    vec3 N = normalize(cross(ddx, ddy));

    // Some invert if N.z > 0
    //if (N.z > 0.0) {
    //    N = -N;
    //}

    // Map [-1..1] -> [0..1] for display
    vec3 normalColor = 0.5 * (N + vec3(1.0));
    FragColor = vec4(normalColor, 1.0);
}
