#version 330 core

in vec2 TexCoords;
out vec4 FragColor;

// Samplers
uniform sampler2D uNormalsTex;   
uniform sampler2D uLightAttenTex;
uniform sampler2D uDepthTex;     

// Camera/eye parameters
uniform mat4 invProjectionMatrix; 
uniform float zNear;
uniform float zFar;
uniform vec3 viewPos;

// Light
uniform vec3 lightPos;    
uniform vec3 lightColor;

// Material
uniform vec3 fluidColor;
uniform float shininess;

// Fresnel
uniform float fresnelPower;  
uniform float fresnelScale;  
uniform float fresnelBias;   

vec3 reconstructEyePos(vec2 uv, float depthLin)
{
    float dEye = depthLin * (zFar - zNear) + zNear;
    vec2 ndc    = uv * 2.0 - 1.0; 
    vec4 clip   = vec4(ndc, -1.0, 1.0);
    vec4 eye4   = invProjectionMatrix * clip;
    vec3 dir    = normalize(eye4.xyz / eye4.w);

    return dEye * dir;
}

void main()
{
    // Depth => position
    float depthLin = texture(uDepthTex, TexCoords).r;
    if (depthLin >= 0.9999) {
        // background => skip
        FragColor = vec4(1.0);
        return;
    }
    vec3 position = reconstructEyePos(TexCoords, depthLin);

    // Normal
    vec3 packedNormal = texture(uNormalsTex, TexCoords).rgb;
    if (length(packedNormal) < 0.001) {
        FragColor = vec4(1.0);
        return;
    }
    vec3 normal = normalize(packedNormal * 2.0 - 1.0);

    // Blinn-Phong shading
    vec3 viewDir  = normalize(viewPos - position);
    vec3 lightDir = normalize(lightPos - position);
    vec3 halfDir  = normalize(lightDir + viewDir);

    vec3 ambient  = 0.1 * lightColor;
    float ndl     = max(dot(normal, lightDir), 0.0);
    vec3 diffuse  = ndl * lightColor;

    float ndh     = max(dot(normal, halfDir), 0.0);
    float spec    = pow(ndh, shininess);
    vec3 specular = spec * lightColor;

    // Combine
    vec3 lighting = ambient + diffuse + specular;

    //  Fresnel
    float cosTheta = max(dot(viewDir, normal), 0.0);
    float fresnelTerm = fresnelBias 
                      + fresnelScale * pow((1.0 - cosTheta), fresnelPower);
    fresnelTerm = clamp(fresnelTerm, 0.0, 1.0);

    //  Base shading
    vec3 transmissionColor = fluidColor * lighting;

    // Reflection color (approx)
    vec3 reflectionColor = lightColor * 2.0; // or pick a sky color, etc.

    vec3 colorFresnel = mix(transmissionColor, reflectionColor, fresnelTerm);

    // Volumetric attenuation from attennuation pass
    vec3 attenuationRGB = texture(uLightAttenTex, TexCoords).rgb;
    colorFresnel *= attenuationRGB;

    // Alpha from attenuation pass
    float alpha = texture(uLightAttenTex, TexCoords).a;

    FragColor = vec4(colorFresnel, alpha);
}
