#include "include/Camera.h"
#include <cstdio>
#include <cstdlib>

namespace FluidSimulation {

    static Camera* g_camera = nullptr;

    Camera::Camera(vec3 position, vec3 target, vec3 up, float fov_, float aspect_, float zNear_, float zFar_)
        : campos(position), forward(VectorSub(target, position)), upVector(up),
          fov(fov_), aspectRatio(aspect_), zNear(zNear_), zFar(zFar_), rotationMatrix(IdentityMatrix()) 
    {
        float norm = sqrtf(forward.x*forward.x + forward.y*forward.y + forward.z*forward.z);
        forward = ScalarMult(forward, 1.0f / norm);
        m_light_pos = vec3(15.0f, 15.0f, 15.0f);
    }

    void Camera::processKey(unsigned char key) {
        switch (key) {
            case 'w':
                campos += ScalarMult(forward, 0.5f);
                break;
            case 's':
                campos -= ScalarMult(forward, 0.5f);
                break;
            case 'a':
                campos -= ScalarMult(CrossProduct(forward, upVector), 0.5f);
                break;
            case 'd':
                campos += ScalarMult(CrossProduct(forward, upVector), 0.5f);
                break;
            case 'j':
                rotateSide(0.03f);
                break;
            case 'l':
                rotateSide(-0.03f);
                break;
            case 'i':
                rotateUp(0.05f);
                break;
            case 'k':
                rotateUp(-0.05f);
                break;
            case 32: // SPACEBAR
                campos += ScalarMult(upVector, 1.0f);
                break;
            case 'c':
                campos -= ScalarMult(upVector, 1.0f);
                break;
            case 'v':
                toggleDisplayBorder();
                break;
            case 0x1b: // ESC
                exit(0);
        }
    }

    void Camera::reshape(int width, int height) {
        glViewport(0, 0, width, height);
        aspectRatio = (float)width / (float)height;
    }

    mat4 Camera::getViewMatrix() const {
        return lookAt(campos, VectorAdd(campos, forward), upVector);
    }

    mat4 Camera::getProjectionMatrix() const {
        return perspective(fov, aspectRatio, zNear, zFar);
    }

    void Camera::rotateSide(float angle) {
        mat3 rot = mat4tomat3(Ry(angle));
        forward = MultMat3Vec3(rot, forward);
    }

    void Camera::rotateUp(float angle) {
        vec3 side = CrossProduct(forward, upVector);
        mat3 rot = mat4tomat3(ArbRotate(side, angle));
        forward = MultMat3Vec3(rot, forward);
    }

    void initCamera(Camera& camera) {
        g_camera = &camera;
    }

    float Camera::getFovyRadians() const {
        return fov * static_cast<float>(M_PI) / 180.0f;
    }

    vec3 Camera::getEyeSpacePosition() const {
        return vec3(0.0f, 0.0f, 0.0f);
    }    

} // namespace FluidSimulation