// FluidRenderer.cpp
#include "include/FluidRenderer.h"
#include "GL_utilities.h" // For loadShaders
#include <iostream>

#define MAIN
#include "VectorUtils4.h"
#undef MAIN

namespace FluidSimulation {

// Constructor
FluidRenderer::FluidRenderer(int width, int height, Camera* camera)
    : m_fbWidth(width), m_fbHeight(height), m_camera(camera),
      m_particleVBO(0), m_boundaryVBO(0) {
        init();
}

void FluidRenderer::deleteFramebufferResources(ShaderProgram& shaderProgram) {
    if (shaderProgram.fboTexture != 0) {
        glDeleteTextures(1, &shaderProgram.fboTexture);
        shaderProgram.fboTexture = 0;
    }
    if (shaderProgram.depthRBO != 0) {
        glDeleteRenderbuffers(1, &shaderProgram.depthRBO);
        shaderProgram.depthRBO = 0;
    }
    if (shaderProgram.fbo != 0) {
        glDeleteFramebuffers(1, &shaderProgram.fbo);
        shaderProgram.fbo = 0;
    }
}

// Destructor
FluidRenderer::~FluidRenderer() {
    // Cleanup Normal Render Program
    deleteFramebufferResources(m_normalRenderProgram);

    // Cleanup Depth Render Program
    deleteFramebufferResources(m_depthRenderProgram);
    deleteFramebufferResources(m_depthFilterProgram);
    // Cleanup Visualization Render Program
    deleteFramebufferResources(m_visualizationProgram);

    // Cleanup Shader Programs
    if (m_visualizationProgram.programID != 0) {
        glDeleteProgram(m_visualizationProgram.programID);
        m_visualizationProgram.programID = 0;
    }
    if (m_depthRenderProgram.programID != 0) {
        glDeleteProgram(m_depthRenderProgram.programID);
        m_depthRenderProgram.programID = 0;
    }
    if (m_normalRenderProgram.programID != 0) {
        glDeleteProgram(m_normalRenderProgram.programID);
        m_normalRenderProgram.programID = 0;
    }

    std::cout << "FluidRenderer destroyed and all resources cleaned up." << std::endl;
}

// Initialization
void FluidRenderer::init() {
    initGLEW();
    initOpenGLSettings();
    initShaders();
    initShaderUniforms();
    initFBOs();
    initFullscreenQuad();
}

// Initialize GLEW
void FluidRenderer::initGLEW() {
    glewExperimental = GL_TRUE;
    GLenum err = glewInit();
    if (err != GLEW_OK) {
        fprintf(stderr, "Error initializing GLEW: %s\n", glewGetErrorString(err));
        exit(1);
    }
    std::cout << "GLEW initialized successfully." << std::endl;
}

// Initialize OpenGL Settings
void FluidRenderer::initOpenGLSettings() {
    glEnable(GL_PROGRAM_POINT_SIZE);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);

    // Enable OpenGL Debug Output
    //glEnable(GL_DEBUG_OUTPUT);
    //glDebugMessageCallback(MessageCallback, 0);

    // Create and bind a Vertex Array Object (VAO) - mandatory for OpenGL 3.3+
    glGenVertexArrays(1, &particles_vao);
    glBindVertexArray(particles_vao);

    std::cout << "OpenGL settings initialized." << std::endl;
}

// Debug callback function
void GLAPIENTRY MessageCallback(GLenum source,
                                GLenum type,
                                GLuint id,
                                GLenum severity,
                                GLsizei length,
                                const GLchar* message,
                                const void* userParam)
{
    std::cerr << "GL CALLBACK: " << (type == GL_DEBUG_TYPE_ERROR ? "** GL ERROR **" : "")
              << " type = " << type
              << ", severity = " << severity
              << ", message = " << message << std::endl;
}

// Initialize Shaders
void FluidRenderer::initShaders() {
    // Normal Render Shader
    m_normalRenderProgram.programID = loadShaders("../shaders/particle.vert", "../shaders/particle.frag");
    if (m_normalRenderProgram.programID == 0) {
        std::cerr << "Failed to load Normal Render Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_normalRenderProgram.programID);
    std::cout << "Normal Render Shaders initialized successfully." << std::endl;

    // =================== Depth shader stages =========================
    // Depth Render Shader
    m_depthRenderProgram.programID = loadShaders("../shaders/depth.vert", "../shaders/depth.frag");
    if (m_depthRenderProgram.programID == 0) {
        std::cerr << "Failed to load Depth Render Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_depthRenderProgram.programID);
    std::cout << "Depth Render Shaders initialized successfully." << std::endl;

    // Depth filter program
    m_depthFilterProgram.programID = loadShaders(
        "../shaders/depth_filter.vert",
        "../shaders/depth_filter.frag"
    );
    if (m_depthFilterProgram.programID == 0) {
        std::cerr << "Failed to load Depth Filter Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_depthFilterProgram.programID);
    std::cout << "Depth Filter Shaders initialized successfully." << std::endl;

    // Depth Normals Program
    m_depthNormalsProgram.programID = loadShaders(
        "../shaders/depth_normals.vert",
        "../shaders/depth_normals.frag"
    );
    if (m_depthNormalsProgram.programID == 0) {
        std::cerr << "Failed to load Depth Normals Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_depthNormalsProgram.programID);
    std::cout << "Depth Normals Shaders initialized successfully." << std::endl;

    // =================== Thickness shader stages =========================
    // Thickness Render Shader
    m_thicknessRenderProgram.programID = loadShaders(
        "../shaders/thickness.vert",
        "../shaders/thickness.frag"
    );
    if (m_thicknessRenderProgram.programID == 0) {
        std::cerr << "Failed to load Thickness Render Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_thicknessRenderProgram.programID);
    std::cout << "Thickness Render Shaders initialized successfully." << std::endl;

        // Thickness Filter Program
    m_thicknessFilterProgram.programID = loadShaders(
        "../shaders/thickness_filter.vert",
        "../shaders/thickness_filter.frag"
    );
    if (m_thicknessFilterProgram.programID == 0) {
        std::cerr << "Failed to load Thickness Filter Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_thicknessFilterProgram.programID);
    std::cout << "Thickness Filter Shaders initialized successfully." << std::endl;

    // Thickness Attenuation Program
    m_thicknessAttenProgram.programID = loadShaders(
        "../shaders/thickness_attenuation.vert",
        "../shaders/thickness_attenuation.frag"
    );
    if (m_thicknessAttenProgram.programID == 0) {
        std::cerr << "Failed to load Thickness Attenuation Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_thicknessAttenProgram.programID);
    std::cout << "Thickness Attenuation Shaders initialized successfully." << std::endl;


    // =================== Visulisation shader =========================
    // Visualization Shader
    m_visualizationProgram.programID = loadShaders("../shaders/visualize_grayscale.vert", "../shaders/visualize_grayscale.frag");
    if (m_visualizationProgram.programID == 0) {
        std::cerr << "Failed to load Visualization Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_visualizationProgram.programID);
    std::cout << "Visualization Shaders initialized successfully." << std::endl;

    // =================== Final Combine shader =========================
    m_finalCombineProgram.programID = loadShaders(
        "../shaders/final_combine.vert",
        "../shaders/final_combine.frag"
    );
    if (m_finalCombineProgram.programID == 0) {
        std::cerr << "Failed to load Final Combine Shaders." << std::endl;
        exit(1);
    }
    glUseProgram(m_finalCombineProgram.programID);
    std::cout << "Final Combine Shaders initialized successfully." << std::endl;
}

void FluidRenderer::setupShaderProgram(ShaderProgram& shaderProgram)
{
    glUseProgram(shaderProgram.programID);

    // 1. Try fetching the attribute for position
    //    Note: Some shaders use "aPosition", some use "aPos".
    shaderProgram.positionAttribLoc = glGetAttribLocation(shaderProgram.programID, "aPosition");
    if (shaderProgram.positionAttribLoc == -1) {
        // If "aPosition" is not found, try "aPos" (used by visualization)
        shaderProgram.positionAttribLoc = glGetAttribLocation(shaderProgram.programID, "aPos");
        if (shaderProgram.positionAttribLoc == -1) {
            fprintf(stderr, "Shader (ID %d): Could not find attribute 'aPosition' or 'aPos'\n",
                    shaderProgram.programID);
        } else {
            printf("Shader (ID %d): Found attribute 'aPos' at location %d\n",
                   shaderProgram.programID, shaderProgram.positionAttribLoc);
        }
    } else {
        printf("Shader (ID %d): Found attribute 'aPosition' at location %d\n",
               shaderProgram.programID, shaderProgram.positionAttribLoc);
    }

    //  Fetch uniforms that may or may not exist in this shader.
    //  If the location is -1, that uniform is absent; just skip it.
    shaderProgram.modelMatrixLoc      = glGetUniformLocation(shaderProgram.programID, "modelMatrix");
    shaderProgram.viewMatrixLoc       = glGetUniformLocation(shaderProgram.programID, "viewMatrix");
    shaderProgram.projectionMatrixLoc = glGetUniformLocation(shaderProgram.programID, "projectionMatrix");
    shaderProgram.pointRadiusLoc      = glGetUniformLocation(shaderProgram.programID, "pointRadius");
    shaderProgram.pointScaleLoc       = glGetUniformLocation(shaderProgram.programID, "pointScale");
    shaderProgram.uColorLoc           = glGetUniformLocation(shaderProgram.programID, "uColor");
    shaderProgram.sphereRadiusLoc     = glGetUniformLocation(shaderProgram.programID, "sphereRadius");
    shaderProgram.nearPlaneLoc        = glGetUniformLocation(shaderProgram.programID, "zNear");
    shaderProgram.farPlaneLoc         = glGetUniformLocation(shaderProgram.programID, "zFar");

    
    // Set default values if the uniform is present
    
    // For point rendering (as in particle shaders):
    if (shaderProgram.pointRadiusLoc != -1) {
        glUniform1f(shaderProgram.pointRadiusLoc, m_pointRadius);
    }
    if (shaderProgram.pointScaleLoc != -1) {
        glUniform1f(shaderProgram.pointScaleLoc, m_pointScale);
    }
    // For model matrix, set to Identity by default
    if (shaderProgram.modelMatrixLoc != -1) {
        mat4 modelMatrix = IdentityMatrix();
        glUniformMatrix4fv(shaderProgram.modelMatrixLoc, 1, GL_TRUE, modelMatrix.m);
    }
    // For near/far planes, if present
    if (shaderProgram.nearPlaneLoc != -1 || shaderProgram.farPlaneLoc != -1) {
        float cameraNearPlane = m_camera->getNearPlane(); 
        float cameraFarPlane  = m_camera->getFarPlane();

        if (shaderProgram.nearPlaneLoc != -1) {
            glUniform1f(shaderProgram.nearPlaneLoc, cameraNearPlane);
        }
        if (shaderProgram.farPlaneLoc != -1) {
            glUniform1f(shaderProgram.farPlaneLoc, cameraFarPlane);
        }
    }

    std::cout << "Shader (ID " << shaderProgram.programID
              << ") uniform/attribute locations have been fetched and set where applicable.\n";
}

// Initialize Shader Uniforms
void FluidRenderer::initShaderUniforms()
{
    // Normal render program:
    setupShaderProgram(m_normalRenderProgram);

    // =================== Depth shader stages =========================
    // Depth render program:
    setupShaderProgram(m_depthRenderProgram);
    // Depth Filter program:
    setupShaderProgram(m_depthFilterProgram);

    setupShaderProgram(m_depthNormalsProgram);
    
    // =================== Thickness shader stages =========================
    // Thickness buffer program:
    setupShaderProgram(m_thicknessRenderProgram);

    setupShaderProgram(m_thicknessFilterProgram);

    setupShaderProgram(m_thicknessAttenProgram);

    // Visualization program:
    setupShaderProgram(m_visualizationProgram);

    setupShaderProgram(m_finalCombineProgram);

    std::cout << "All shader programs' uniforms initialized via setupShaderProgram()." << std::endl;
}

void FluidRenderer::initializeFramebuffer(ShaderProgram& shaderProgram, GLenum textureInternalFormat, GLenum textureFormat, GLenum textureType) {
    // Generate and bind the framebuffer
    glGenFramebuffers(1, &shaderProgram.fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, shaderProgram.fbo);

    // Generate and bind the texture
    glGenTextures(1, &shaderProgram.fboTexture);
    glBindTexture(GL_TEXTURE_2D, shaderProgram.fboTexture);
    glTexImage2D(GL_TEXTURE_2D, 0, textureInternalFormat, m_fbWidth, m_fbHeight, 0, textureFormat, textureType, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, shaderProgram.fboTexture, 0);

    // Generate and bind the renderbuffer for depth
    glGenRenderbuffers(1, &shaderProgram.depthRBO);
    glBindRenderbuffer(GL_RENDERBUFFER, shaderProgram.depthRBO);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, m_fbWidth, m_fbHeight);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, shaderProgram.depthRBO);

    // Specify the list of draw buffers
    GLenum drawBuffers[] = { GL_COLOR_ATTACHMENT0 };
    glDrawBuffers(1, drawBuffers);

    // Check framebuffer completeness
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
        fprintf(stderr, "Framebuffer not complete for shader program ID %d: 0x%X\n", shaderProgram.programID, glCheckFramebufferStatus(GL_FRAMEBUFFER));
    } else {
        std::cout << "Framebuffer initialized successfully for shader program ID " << shaderProgram.programID << "." << std::endl;
    }

    // Unbind the framebuffer to return to default
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Initialize Framebuffers
void FluidRenderer::initFBOs() {
    // Initialize Normal Render Shader FBO with GL_RGBA8
    initializeFramebuffer(m_normalRenderProgram, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE);

    // =================== Depth shader stages =========================
    // Initialize Depth Render Shader FBO with GL_R32F
    initializeFramebuffer(m_depthRenderProgram, GL_R32F, GL_RED, GL_FLOAT);

    initializeFramebuffer(m_depthFilterProgram,
                      GL_R32F, GL_RED, GL_FLOAT);

    glGenTextures(1, &m_depthFilterTempTex);
    glBindTexture(GL_TEXTURE_2D, m_depthFilterTempTex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, m_fbWidth, m_fbHeight, 0, GL_RED, GL_FLOAT, nullptr);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    // Create extra FBO for it:
    glGenFramebuffers(1, &m_depthFilterTempFBO);
    glBindFramebuffer(GL_FRAMEBUFFER, m_depthFilterTempFBO);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_depthFilterTempTex, 0);
    GLenum drawBufs[1] = {GL_COLOR_ATTACHMENT0};
    glDrawBuffers(1, drawBufs);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
        std::cerr << "Temp FBO not complete!\n";
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    // depth noirmals program
    initializeFramebuffer(m_depthNormalsProgram, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE);

    // =================== Thickness shader stages =========================
    // Initialize Thickness Render Shader FBO with GL_R32F
    initializeFramebuffer(m_thicknessRenderProgram,
                      GL_R32F,   // internal format
                      GL_RED,    // texture format
                      GL_FLOAT); // data type


    // Create thicknessFilterProgram FBO
    initializeFramebuffer(m_thicknessFilterProgram, GL_R32F, GL_RED, GL_FLOAT);

    // Create a temporary texture for pass 1
    glGenTextures(1, &m_thicknessFilterTempTex);
    glBindTexture(GL_TEXTURE_2D, m_thicknessFilterTempTex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, m_fbWidth, m_fbHeight, 0, GL_RED, GL_FLOAT, nullptr);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    // Create extra FBO for pass 1
    glGenFramebuffers(1, &m_thicknessFilterTempFBO);
    glBindFramebuffer(GL_FRAMEBUFFER, m_thicknessFilterTempFBO);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_thicknessFilterTempTex, 0);
    GLenum drawBufs2[1] = {GL_COLOR_ATTACHMENT0};
    glDrawBuffers(1, drawBufs2);
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
        std::cerr << "Temp FBO for thickness filter not complete!\n";
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    
    initializeFramebuffer(m_thicknessAttenProgram, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE);
    // Initialize Visualization Render Shader FBO with GL_RGBA8
    initializeFramebuffer(m_visualizationProgram, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE);

    initializeFramebuffer(m_finalCombineProgram, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE);
    std::cout << "All Framebuffers for Normal, Depth, and Visualization shaders initialized successfully." << std::endl;
}


void FluidRenderer::initFullscreenQuad() {
    float quadVertices[] = {
        // positions   // texCoords
        -1.0f,  1.0f, 0.0f,  0.0f, 1.0f,
        -1.0f, -1.0f, 0.0f,  0.0f, 0.0f,
         1.0f, -1.0f, 0.0f,  1.0f, 0.0f,

        -1.0f,  1.0f, 0.0f,  0.0f, 1.0f,
         1.0f, -1.0f, 0.0f,  1.0f, 0.0f,
         1.0f,  1.0f, 0.0f,  1.0f, 1.0f
    };

    glGenVertexArrays(1, &quadVAO);
    GLuint quadVBO;
    glGenBuffers(1, &quadVBO);
    glBindVertexArray(quadVAO);
    glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
    // Position attribute
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
    // Texture Coord attribute
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
    glBindVertexArray(0);

    std::cout << "Fullscreen quad initialized." << std::endl;
}

void FluidRenderer::resizeFramebuffer(
    ShaderProgram& shaderProgram,
    GLenum textureInternalFormat,
    GLenum textureFormat,
    GLenum textureType,
    int newWidth,
    int newHeight)
{
    // Bind the FBO
    glBindFramebuffer(GL_FRAMEBUFFER, shaderProgram.fbo);

    // Resize the texture
    glBindTexture(GL_TEXTURE_2D, shaderProgram.fboTexture);
    glTexImage2D(GL_TEXTURE_2D, 0, textureInternalFormat,
                 newWidth, newHeight, 0,
                 textureFormat, textureType, nullptr);

    // Resize the renderbuffer
    glBindRenderbuffer(GL_RENDERBUFFER, shaderProgram.depthRBO);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, newWidth, newHeight);

    // Check completeness
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
        fprintf(stderr, "FBO (ID %d) not complete after resize: 0x%X\n",
                shaderProgram.programID, glCheckFramebufferStatus(GL_FRAMEBUFFER));
    }

    // Unbind
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}


// Resize Framebuffers
void FluidRenderer::resizeFBO(int newWidth, int newHeight) {
    if (newWidth <= 0 || newHeight <= 0) return;

    m_fbWidth = newWidth;
    m_fbHeight = newHeight;

    // Resize Normal Render Shader FBO
    resizeFramebuffer(m_normalRenderProgram,
                      GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE,
                      newWidth, newHeight);

    // =================== Depth shader stages =========================
    // Resize Depth Render Shader FBO
    resizeFramebuffer(m_depthRenderProgram,
                      GL_R32F, GL_RED, GL_FLOAT,
                      newWidth, newHeight);

    resizeFramebuffer(m_depthFilterProgram,
                      GL_R32F, GL_RED, GL_FLOAT,
                      newWidth, newHeight);

    // =================== Thickness shader stages =========================
     resizeFramebuffer(m_thicknessRenderProgram,
                      GL_R32F, GL_RED, GL_FLOAT,
                      newWidth, newHeight);
    
    // Resize Visualization Render Shader FBO
    resizeFramebuffer(m_visualizationProgram,
                      GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE,
                      newWidth, newHeight);

    std::cout << "Framebuffers resized to " << newWidth << "x" << newHeight << "." << std::endl;
}

// Set VBOs
void FluidRenderer::setVBOs(GLuint particleVBO, GLuint boundaryVBO) {
    m_particleVBO = particleVBO;
    m_boundaryVBO = boundaryVBO;
}

void FluidRenderer::renderPoints(GLuint vbo,
                                 size_t count,
                                 const ShaderProgram& shaderProgram,
                                 const vec3& color)
{
    glUseProgram(shaderProgram.programID);
    glUniform3f(shaderProgram.uColorLoc, color.x, color.y, color.z);

    // Bind the VBO
    glBindBuffer(GL_ARRAY_BUFFER, vbo);

    // Enable position attribute
    glEnableVertexAttribArray(shaderProgram.positionAttribLoc);
    glVertexAttribPointer(
        shaderProgram.positionAttribLoc, // location
        3,                               // number of components
        GL_FLOAT,                        // type
        GL_FALSE,                        // normalized?
        sizeof(float3),                  // stride
        (void*)0                         // offset
    );

    // Draw
    glDrawArrays(GL_POINTS, 0, static_cast<GLsizei>(count));

    // Disable
    glDisableVertexAttribArray(shaderProgram.positionAttribLoc);
}


// Render Particles
void FluidRenderer::renderParticles(size_t particleCount,
                                    const ShaderProgram& shaderProgram)
{
    renderPoints(m_particleVBO, particleCount, shaderProgram, vec3(0.29f, 0.573f, 1.0f));
}

// Render Boundaries
void FluidRenderer::renderBoundaries(size_t boundaryCount,
                                     const ShaderProgram& shaderProgram)
{
    renderPoints(m_boundaryVBO, boundaryCount, shaderProgram, vec3(0.29f, 1.0f, 0.57f));
}

// Render Normal Frame
void FluidRenderer::renderNormalFrame(size_t particleCount, size_t boundaryCount) {

    // Bind the Normal Render Shader framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, m_normalRenderProgram.fbo);
    
    // Set the viewport to match the framebuffer size
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    
    glClearColor(0.9f, 0.9f, 0.9f, 1.0f);
    
    // Clear color and depth buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    glDepthMask(GL_TRUE);

    glUseProgram(m_normalRenderProgram.programID);
    glBindVertexArray(particles_vao);


    mat4 projectionMatrix = m_camera->getProjectionMatrix();
    glUniformMatrix4fv(m_normalRenderProgram.projectionMatrixLoc, 1, GL_TRUE, projectionMatrix.m);

    mat4 viewMatrix = m_camera->getViewMatrix();
    glUniformMatrix4fv(m_normalRenderProgram.viewMatrixLoc, 1, GL_TRUE, viewMatrix.m);
    
    if (m_normalRenderProgram.pointRadiusLoc != -1) {
        glUniform1f(m_normalRenderProgram.pointRadiusLoc, m_pointRadius);
    }
    if (m_normalRenderProgram.pointScaleLoc != -1) {
        glUniform1f(m_normalRenderProgram.pointScaleLoc, m_pointScale);
    }

    renderParticles(particleCount, m_normalRenderProgram);

    // Render boundaries if enabled
    if (m_camera->shouldDisplayBorder()) {
        renderBoundaries(boundaryCount, m_normalRenderProgram);
    }

    // Unbind the framebuffer to return to default framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Render Depth Frame
void FluidRenderer::renderDepthFrame(size_t particleCount, size_t boundaryCount) {

    // Bind the Depth Render Shader framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, m_depthRenderProgram.fbo);
    
    // Set the viewport to match the framebuffer size
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    // Clear color and depth buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    glDepthMask(GL_TRUE);

    glUseProgram(m_depthRenderProgram.programID);
    glBindVertexArray(particles_vao);

    mat4 projectionMatrix = m_camera->getProjectionMatrix();
    glUniformMatrix4fv(m_depthRenderProgram.projectionMatrixLoc, 1, GL_TRUE, projectionMatrix.m);

    mat4 viewMatrix = m_camera->getViewMatrix();
    glUniformMatrix4fv(m_depthRenderProgram.viewMatrixLoc, 1, GL_TRUE, viewMatrix.m);

    if (m_depthRenderProgram.sphereRadiusLoc != -1) {
        glUniform1f(m_depthRenderProgram.sphereRadiusLoc, m_sphereRadius);
    }
    if (m_depthRenderProgram.pointRadiusLoc != -1) {
        glUniform1f(m_depthRenderProgram.pointRadiusLoc, m_pointRadius);
    }
    if (m_depthRenderProgram.pointScaleLoc != -1) {
        glUniform1f(m_depthRenderProgram.pointScaleLoc, m_pointScale);
    }

    float cameraNearPlane = m_camera->getNearPlane();
    float cameraFarPlane = m_camera->getFarPlane();

    if (m_depthRenderProgram.nearPlaneLoc != -1) {
        glUniform1f(m_depthRenderProgram.nearPlaneLoc, cameraNearPlane);
    }
    if (m_depthRenderProgram.farPlaneLoc != -1) {
        glUniform1f(m_depthRenderProgram.farPlaneLoc, cameraFarPlane);
    }

    renderParticles(particleCount, m_depthRenderProgram);

    // Unbind the framebuffer to return to default framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Currenntly onuseed - probably remove later, since we do adptive in shader
void FluidRenderer::computeGaussianWeights(int radius, std::vector<float>& outWeights)
{
    outWeights.resize(radius+1); // we store only half + center
    float sigma = float(radius) / 3.0f; // or pick your approach
    float sum = 0.0f;

    // Compute weights
    for (int i = 0; i <= radius; i++) {
        float w = exp(-(i*i) / (2.0f * sigma * sigma));
        outWeights[i] = w;
        // we only store half, so account for symmetry in sum
        sum += (i == 0) ? w : 2.0f * w;
    }

    // Normalize
    for (int i = 0; i <= radius; i++) {
        outWeights[i] /= sum;
    }
}

void FluidRenderer::renderDepthFilterFrame()
{
    // Defining the parameters for a world-space radius blur
    //    (and other relevant uniforms).
    float worldRadiusMeters  = 0.2f;   // e.g. 2 cm smoothing in real space
    float maxPixelRadius     = 60.0f;   // clamp for performance
    float sigmaSpatialFactor = 10.0f;   // controls the spatial falloff in the blur loop
    float sigmaR             = 4.0f;    // range weight exponent
    float depthThresh        = 0.2f;    // bilateral threshold for ignoring big depth jumps

    // Retrieve camera values for FOV, near/far from your Camera object
    float cameraFovY = m_camera->getFovyRadians();
    float zNear      = m_camera->getNearPlane();
    float zFar       = m_camera->getFarPlane();

    // =============== PASS 1: HORIZONTAL ===============
    {
        // Bind the temp FBO (for horizontal pass)
        glBindFramebuffer(GL_FRAMEBUFFER, m_depthFilterTempFBO);
        glViewport(0, 0, m_fbWidth, m_fbHeight);
        glClearColor(0, 0, 0, 1);
        glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_DEPTH_TEST);

        // Use the depthFilterProgram
        glUseProgram(m_depthFilterProgram.programID);

        // Input texture = the original depth buffer
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, m_depthRenderProgram.fboTexture);
        glUniform1i(glGetUniformLocation(m_depthFilterProgram.programID, "uDepthTex"), 0);

        // passDirection = horizontal
        float passDirX = 1.0f / m_fbWidth;
        float passDirY = 0.0f;
        glUniform2f(glGetUniformLocation(m_depthFilterProgram.programID, "passDirection"),
                    passDirX, passDirY);

        //  Set "per-pixel" uniforms:
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "cameraFovY"), cameraFovY);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "zNear"),      zNear);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "zFar"),       zFar);
        glUniform1i(glGetUniformLocation(m_depthFilterProgram.programID, "screenHeight"),
                    m_fbHeight);

        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "worldRadiusMeters"),
                    m_worldSmoothingRadius);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "maxPixelRadius"),
                    m_maxPixelRadius);

        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "sigmaSpatialFactor"),
                    m_spatialFactor);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "sigma_r"),
                    m_blurrStrength);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "depthThreshold"),
                    m_deptThreshold);

        // Render the fullscreen quad
        glBindVertexArray(quadVAO);
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);

        // Restore states
        glEnable(GL_DEPTH_TEST);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }

    // =============== PASS 2: VERTICAL ===============
    {
        // Bind final FBO (the main depthFilterProgram's FBO)
        glBindFramebuffer(GL_FRAMEBUFFER, m_depthFilterProgram.fbo);
        glViewport(0, 0, m_fbWidth, m_fbHeight);
        glClearColor(0, 0, 0, 1);
        glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_DEPTH_TEST);

        // Same program
        glUseProgram(m_depthFilterProgram.programID);

        // Input = pass 1's texture
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, m_depthFilterTempTex);
        glUniform1i(glGetUniformLocation(m_depthFilterProgram.programID, "uDepthTex"), 0);

        // passDirection = vertical
        float passDirX2 = 0.0f;
        float passDirY2 = 1.0f / m_fbHeight;
        glUniform2f(glGetUniformLocation(m_depthFilterProgram.programID, "passDirection"),
                    passDirX2, passDirY2);

        // Set the same per-pixel uniforms again
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "cameraFovY"), cameraFovY);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "zNear"),      zNear);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "zFar"),       zFar);
        glUniform1i(glGetUniformLocation(m_depthFilterProgram.programID, "screenHeight"),
                    m_fbHeight);

        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "worldRadiusMeters"),
                    worldRadiusMeters);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "maxPixelRadius"),
                    maxPixelRadius);

        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "sigmaSpatialFactor"),
                    sigmaSpatialFactor);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "sigma_r"),   sigmaR);
        glUniform1f(glGetUniformLocation(m_depthFilterProgram.programID, "depthThreshold"),
                    depthThresh);

        // Draw the fullscreen quad
        glBindVertexArray(quadVAO);
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);

        // Restore states
        glEnable(GL_DEPTH_TEST);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
}

void FluidRenderer::renderDepthNormalsFrame()
{
    // 1) Bind the Depth Normals FBO
    glBindFramebuffer(GL_FRAMEBUFFER, m_depthNormalsProgram.fbo);
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

    // 2) Use the depthNormals program
    glUseProgram(m_depthNormalsProgram.programID);

    // 3) Bind the *filtered depth* texture as input
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_depthFilterProgram.fboTexture); // temporary to se normals for spheres
    glUniform1i(glGetUniformLocation(m_depthNormalsProgram.programID, "uDepthTex"), 0);

    // 4) Set uniforms: zNear, zFar, uTexelSize, maxDepth, etc.
    //float zNear = m_camera->getNearPlane();
    //float zFar  = m_camera->getFarPlane();
    //glUniform1f(glGetUniformLocation(m_depthNormalsProgram.programID, "zNear"), zNear);
    //glUniform1f(glGetUniformLocation(m_depthNormalsProgram.programID, "zFar"),  zFar);
    mat4 proj = m_camera->getProjectionMatrix(); // Do we have to transpose our matrix???
    mat4 invProj = inverse(proj); // or use glm::inverse(...) if using GLM

    GLuint invProjLoc = glGetUniformLocation(m_depthNormalsProgram.programID, "invProjectionMatrix");
    glUniformMatrix4fv(invProjLoc, 1, GL_TRUE, invProj.m); 

    // For sampling partial derivatives:
    float texelSizeX = 1.0f / float(m_fbWidth);
    float texelSizeY = 1.0f / float(m_fbHeight);
    glUniform2f(glGetUniformLocation(m_depthNormalsProgram.programID, "uTexelSize"),
                texelSizeX, texelSizeY);

    // optional maxDepth
    glUniform1f(glGetUniformLocation(m_depthNormalsProgram.programID, "maxDepth"), 1.0f);

    // 5) Render Fullscreen Quad
    glBindVertexArray(quadVAO);
    glDrawArrays(GL_TRIANGLES, 0, 6);
    glBindVertexArray(0);

    // 6) Restore
    glEnable(GL_DEPTH_TEST);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}


void FluidRenderer::renderThicknessFrame(size_t particleCount, size_t boundaryCount)
{
    glBindFramebuffer(GL_FRAMEBUFFER, m_thicknessRenderProgram.fbo);
    glViewport(0, 0, m_fbWidth, m_fbHeight);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Disable depth test so *all* fragments draw
    glDisable(GL_DEPTH_TEST);

    // Enable additive blending
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);

    glUseProgram(m_thicknessRenderProgram.programID);
    glBindVertexArray(particles_vao);

    // Update standard uniforms
    if (m_thicknessRenderProgram.projectionMatrixLoc != -1) {
        mat4 projectionMatrix = m_camera->getProjectionMatrix();
        glUniformMatrix4fv(m_thicknessRenderProgram.projectionMatrixLoc, 1, GL_TRUE, projectionMatrix.m);
    }
    if (m_thicknessRenderProgram.viewMatrixLoc != -1) {
        mat4 viewMatrix = m_camera->getViewMatrix();
        glUniformMatrix4fv(m_thicknessRenderProgram.viewMatrixLoc, 1, GL_TRUE, viewMatrix.m);
    }
    if (m_thicknessRenderProgram.pointRadiusLoc != -1) {
        glUniform1f(m_thicknessRenderProgram.pointRadiusLoc, m_pointRadius);
    }
    if (m_thicknessRenderProgram.pointRadiusLoc != -1) {
        glUniform1f(m_thicknessRenderProgram.pointRadiusLoc, m_pointRadius);
    }
    if (m_thicknessRenderProgram.pointScaleLoc != -1) {
        glUniform1f(m_thicknessRenderProgram.pointScaleLoc, m_pointScale);
    }
    
    glUniform1f(glGetUniformLocation(m_thicknessRenderProgram.programID, "thicknessMass"), m_thicknessMass);

    renderParticles(particleCount, m_thicknessRenderProgram);

    // Restore states
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void FluidRenderer::renderThicknessFilterFrame()
{
    int radius = 24; // or something
    std::vector<float> weights;
    computeGaussianWeights(radius, weights);

    // 1) Horizontal pass -> m_thicknessFilterTempFBO
    {
        glBindFramebuffer(GL_FRAMEBUFFER, m_thicknessFilterTempFBO);
        glViewport(0, 0, m_fbWidth, m_fbHeight);
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_DEPTH_TEST);

        glUseProgram(m_thicknessFilterProgram.programID);

        // Bind the original thickness texture
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, m_thicknessRenderProgram.fboTexture);
        glUniform1i(glGetUniformLocation(m_thicknessFilterProgram.programID, "uThicknessTex"), 0);

        // set passDirection horizontally
        float passDirX = 1.0f / m_fbWidth;
        float passDirY = 0.0f;
        glUniform2f(glGetUniformLocation(m_thicknessFilterProgram.programID, "passDirection"), passDirX, passDirY);

        // kernelRadius
        glUniform1i(glGetUniformLocation(m_thicknessFilterProgram.programID, "kernelRadius"), radius);

        // upload weights
        glUniform1fv(glGetUniformLocation(m_thicknessFilterProgram.programID, "gaussWeights"), radius+1, weights.data());

        // Render quad
        glBindVertexArray(quadVAO);
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);

        glEnable(GL_DEPTH_TEST);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }

    // 2) Vertical pass -> m_thicknessFilterProgram.fbo
    {
        glBindFramebuffer(GL_FRAMEBUFFER, m_thicknessFilterProgram.fbo);
        glViewport(0, 0, m_fbWidth, m_fbHeight);
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_DEPTH_TEST);

        glUseProgram(m_thicknessFilterProgram.programID);

        // Now the input is m_thicknessFilterTempTex
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, m_thicknessFilterTempTex);
        glUniform1i(glGetUniformLocation(m_thicknessFilterProgram.programID, "uThicknessTex"), 0);

        // passDirection vertically
        float passDirX2 = 0.0f;
        float passDirY2 = 1.0f / m_fbHeight;
        glUniform2f(glGetUniformLocation(m_thicknessFilterProgram.programID, "passDirection"), passDirX2, passDirY2);

        glUniform1i(glGetUniformLocation(m_thicknessFilterProgram.programID, "kernelRadius"), radius);
        glUniform1fv(glGetUniformLocation(m_thicknessFilterProgram.programID, "gaussWeights"), radius+1, weights.data());

        // Draw
        glBindVertexArray(quadVAO);
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);

        glEnable(GL_DEPTH_TEST);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
}


void FluidRenderer::renderThicknessAttenFrame()
{
    // 1) Bind some FBO to store the attenuation result
    //    If you want a new FBO, e.g. m_thicknessAttenProgram
    glBindFramebuffer(GL_FRAMEBUFFER, m_thicknessAttenProgram.fbo);
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

    glUseProgram(m_thicknessAttenProgram.programID);

    // 2) Bind blurred thickness
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_thicknessFilterProgram.fboTexture);
    glUniform1i(glGetUniformLocation(m_thicknessAttenProgram.programID, "uThicknessTex"), 0);

    // 3) Set uniform: absorption, fluidColor, etc.
    float absorption = 0.2f; // tweak to taste
    GLuint absorpLoc = glGetUniformLocation(m_thicknessAttenProgram.programID, "alphaScale");
    glUniform1f(absorpLoc, m_alphaScale);

    // color
    float fluidR = 0.3f, fluidG = 0.3f, fluidB = 0.4f;
    glUniform3f(glGetUniformLocation(m_thicknessAttenProgram.programID, "absorptionRGB"),
                m_absorbtionR, m_absorbtionG, m_absorbtionB);

    // 4) Render a fullscreen quad
    glBindVertexArray(quadVAO);
    glDrawArrays(GL_TRIANGLES, 0, 6);
    glBindVertexArray(0);

    glEnable(GL_DEPTH_TEST);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Add this method inside the FluidRenderer class in FluidRenderer.cpp

void FluidRenderer::renderFinalCombineFrame()
{
    // Bind the Final Combine Shader framebuffer or default framebuffer
    // To render to screen, bind framebuffer 0
    glBindFramebuffer(GL_FRAMEBUFFER, m_finalCombineProgram.fbo); // Or 0 for screen
    
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f); // Clear to black
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glDisable(GL_DEPTH_TEST); // Disable depth testing for fullscreen quad
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);  
    
    glUseProgram(m_finalCombineProgram.programID);
    
    // Bind the required textures
    // Assuming:
    // - Normals buffer: m_depthNormalsProgram.fboTexture
    // - Light attenuation buffer: m_thicknessAttenProgram.fboTexture
    // - Depth buffer: m_depthFilterProgram.fboTexture
    
    // Set textures to texture units
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_depthNormalsProgram.fboTexture);
    glUniform1i(glGetUniformLocation(m_finalCombineProgram.programID, "uNormalsTex"), 0);
    
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, m_thicknessAttenProgram.fboTexture);
    glUniform1i(glGetUniformLocation(m_finalCombineProgram.programID, "uLightAttenTex"), 1);
    
    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, m_depthFilterProgram.fboTexture);
    glUniform1i(glGetUniformLocation(m_finalCombineProgram.programID, "uDepthTex"), 2);
    
    // Set additional uniforms
    // Assuming you have camera position in eye space
    // If not, you need to compute it or pass it from your Camera class
    vec3 cameraPos = m_camera->getEyeSpacePosition(); // You need to implement this method
    glUniform3f(glGetUniformLocation(m_finalCombineProgram.programID, "viewPos"),
                cameraPos.x, cameraPos.y, cameraPos.z);

    mat4 projMatrix  = m_camera->getProjectionMatrix();
    mat4 invProj     = inverse(projMatrix); 
    GLint invProjLoc = glGetUniformLocation(m_finalCombineProgram.programID, "invProjectionMatrix");
    glUniformMatrix4fv(invProjLoc, 1, GL_TRUE, invProj.m);

    // Set light position and color
    vec3 lightPos = m_camera->getLightPosition(); // Implement this as per your setup
    glUniform3f(glGetUniformLocation(m_finalCombineProgram.programID, "lightPos"),
                lightPos.x, lightPos.y, lightPos.z);
    
    vec3 lightColor = vec3(1.0f, 1.0f, 1.0f); // White light
    glUniform3f(glGetUniformLocation(m_finalCombineProgram.programID, "lightColor"),
                lightColor.x, lightColor.y, lightColor.z);
    
    // Set fluid color and shininess
    vec3 fluidColor = vec3(0.05f, 0.5f, 0.9f); // Example color
    glUniform3f(glGetUniformLocation(m_finalCombineProgram.programID, "fluidColor"),
                m_fluidColor.x, m_fluidColor.y, m_fluidColor.z);
    
    float shininess = 82.0f; // Example shininess
    glUniform1f(glGetUniformLocation(m_finalCombineProgram.programID, "shininess"), m_shininess);
    
    // Set Fresnel parameters
    float fresnelPower = 2.0f;
    float fresnelScale = 0.5f;
    float fresnelBias  = 0.1f;
    glUniform1f(glGetUniformLocation(m_finalCombineProgram.programID, "fresnelPower"), m_fresnelPower);
    glUniform1f(glGetUniformLocation(m_finalCombineProgram.programID, "fresnelScale"), m_fresnelScale);
    glUniform1f(glGetUniformLocation(m_finalCombineProgram.programID, "fresnelBias"), m_fresnelBias);
    
    // Render the fullscreen quad
    glBindVertexArray(quadVAO);
    glDrawArrays(GL_TRIANGLES, 0, 6);
    glBindVertexArray(0);
    
    // Re-enable depth testing
    glEnable(GL_DEPTH_TEST);
    
    // Unbind the framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}


// Visualize Buffer
void FluidRenderer::visualizeBuffer(RenderStage stage)
{
    // Bind the Visualization FBO
    glBindFramebuffer(GL_FRAMEBUFFER, m_visualizationProgram.fbo);
    glViewport(0, 0, m_fbWidth, m_fbHeight);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f); // White background for visualization
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDisable(GL_DEPTH_TEST);

    glUseProgram(m_visualizationProgram.programID);

    // Retrieve the appropriate texture based on RenderStage
    GLuint textureToBind = getTexture(stage);

    if (textureToBind != 0) {
        // Bind the texture to texture unit 0
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, textureToBind);
        glUniform1i(glGetUniformLocation(m_visualizationProgram.programID, "activeTexture"), 0);

        // Render the fullscreen quad
        glBindVertexArray(quadVAO);
        glDrawArrays(GL_TRIANGLES, 0, 6);
        glBindVertexArray(0);
    }
    else {
        // Handle the case where no valid texture is bound
        std::cerr << "Invalid RenderStage selected for visualization." << std::endl;
    }

    // Re-enable depth testing
    glEnable(GL_DEPTH_TEST);

    // Unbind the framebuffer
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

GLuint FluidRenderer::getTexture(RenderStage stage) const {
    switch (stage) {
        case RenderStage::ScreenSpacedRender:
            return m_finalCombineProgram.fboTexture;
        case RenderStage::DepthBuffer:
            return m_depthRenderProgram.fboTexture;
        case RenderStage::ThicknessBuffer:
            return m_thicknessRenderProgram.fboTexture;
        case RenderStage::FilteredDepth:
            return m_depthFilterProgram.fboTexture;
        case RenderStage::DepthNormals:
            return m_depthNormalsProgram.fboTexture;
        case RenderStage::FilteredThickness:
            return m_thicknessFilterProgram.fboTexture;
        case RenderStage::LightAttenuation:
            return m_thicknessAttenProgram.fboTexture;
        case RenderStage::ParticleView:
            return m_normalRenderProgram.fboTexture;
        default:
            return 0; // Return 0 if no valid stage was selected
    }
}

GLuint FluidRenderer::getVisualizationTexture() const {
    return m_visualizationProgram.fboTexture;
}

} // namespace FluidSimulation