#include "include/FluidSimulation.h"
#include "include/spawn_particles.h" // For createBoundaryParticles, initParticles
//#include "include/PCISPH_sim.h"      // For CUDA simulation functions
#include "settings/constants.h"
#include "cuda/include/PCISPH_CUDA_interface.h"
#include <iostream>

namespace FluidSimulation {

    FluidSimulation::FluidSimulation(int numParticles)
        : m_numParticles(numParticles), m_numBoundaryParticles(0),
          m_vbo_particles(0), m_vbo_boundary(0), m_timeStep(0) 
    {
        m_smoothingRadius = h_H;
        m_particleMass = h_PARTICLE_MASS;
        m_restDensity = h_REST_DENSITY;
        m_viscosity = h_VISCOSITY;

        m_simulationStepsPerFrame = 2;
        m_mode = SimulationMode::OneLargeCube;
    }

    FluidSimulation::~FluidSimulation() {
        FluidSimulationGPU::unregisterVBOFromCUDA();
        FluidSimulationGPU::unregisterBoundaryVBOFromCUDA();
        glDeleteBuffers(1, &m_vbo_particles);
        glDeleteBuffers(1, &m_vbo_boundary);
    }

    void FluidSimulation::Reset()
    {
        // Creaet new fluiid particles
        std::vector<float3> h_positions(m_numParticles);
        std::vector<float3> h_velocities(m_numParticles);
        float gridSpacing = 0.5 * m_smoothingRadius;
        initParticles(h_positions, h_velocities, m_numParticles, gridSpacing, m_mode);
    
        FluidSimulationGPU::resetSimulation(
            h_positions.data(), h_velocities.data(), m_numParticles);
    }

    void FluidSimulation::ResetFluidParameters()
    {
        m_smoothingRadius = h_H;
        m_particleMass = h_PARTICLE_MASS;
        m_restDensity = h_REST_DENSITY;
        m_viscosity = h_VISCOSITY;
        updateConstants();
    }

    void FluidSimulation::setSimulationMode(SimulationMode newMode)
    {
        m_mode = newMode;
        Reset();
    }

    void FluidSimulation::updateConstants() {
    // Send updated parameters to CUDA
        FluidSimulationGPU::updateConstantsOnDevice(
            m_particleMass, m_restDensity, m_viscosity, m_smoothingRadius);
    }

    void FluidSimulation::initSimulation() {
        // Upload constants to the GPU
        FluidSimulationGPU::initializeConstants();

        // Initialize fluid particles
        std::vector<float3> h_positions(m_numParticles);
        std::vector<float3> h_velocities(m_numParticles);
        float gridSpacing = 0.5 * m_smoothingRadius;
        initParticles(h_positions, h_velocities, m_numParticles, gridSpacing, m_mode);

        // Initialize particle device arrays
        FluidSimulationGPU::initializeFluidDeviceArrays(h_positions.data(), h_velocities.data(), m_numParticles);

        // Create particle VBO and register with CUDA
        glGenBuffers(1, &m_vbo_particles);
        glBindBuffer(GL_ARRAY_BUFFER, m_vbo_particles);
        glBufferData(GL_ARRAY_BUFFER, h_positions.size() * sizeof(float3), h_positions.data(), GL_DYNAMIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        FluidSimulationGPU::registerVBOWithCUDA(m_vbo_particles);

        // Initialize boundary
        std::vector<float3> boundary_positions;
        std::vector<float3> boundary_normals;
        int boundaryLayers = 1;
        float spacing = 0.5 * m_smoothingRadius; // Using 0.01 as per your code

        createBoundaryParticles(boundary_positions, boundary_normals, boundaryLayers, spacing);
        m_numBoundaryParticles = boundary_positions.size();

        // Initialize boundary device arrays
        FluidSimulationGPU::initializeBoundryDeviceArrays(boundary_positions.data(), boundary_normals.data(), m_numBoundaryParticles);

        // Create boundary VBO and register with CUDA
        glGenBuffers(1, &m_vbo_boundary);
        glBindBuffer(GL_ARRAY_BUFFER, m_vbo_boundary);
        glBufferData(GL_ARRAY_BUFFER, boundary_positions.size() * sizeof(float3), boundary_positions.data(), GL_DYNAMIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        FluidSimulationGPU::registerBoundaryVBOWithCUDA(m_vbo_boundary);

        // Now that we have boundary particles set up and registered, run the boundary neighborhood search
        FluidSimulationGPU::boundaryNBSearch(m_numBoundaryParticles);
    }

    void FluidSimulation::updateSimulation() {
        for (int i = 0; i < m_simulationStepsPerFrame; ++i) {
            FluidSimulationGPU::updateParticles(0.001);
        }
    }

}
