
#include "include/FluidSimulationApp.h"
#include <iostream>
#include "imgui.h"
#include "settings/constants.h"
#include "include/SimulationMode.h"

namespace FluidSimulation {

    // Constructor
    FluidSimulatorApp::FluidSimulatorApp(appSettings* appSettings)
        : m_appsettings(appSettings) {
        // Initialize FluidRenderer with the settings
        m_camera = new Camera(
            vec3(5.0f, 15.0f, 20.0f),
            vec3(0.0f, 4.0f, 0.0f),
            vec3(0.0f, 1.0f, 0.0f),
            45.0f,
            appSettings->width / (float)appSettings->height,
            0.1f,
            30.0f
        );

        m_fluidRenderer = new FluidRenderer(appSettings->width, appSettings->height, m_camera);
        
        m_fluidSimulation = new FluidSimulation(NUM_PARTICLES);
        m_fluidSimulation->initSimulation();

        m_fluidRenderer->setVBOs(m_fluidSimulation->getParticleVBO(), m_fluidSimulation->getBoundaryVBO());

        m_updatePhysics = true;

        //m_fluidRenderer->init();
        std::cout << "FluidSimulatorApp created with settings: "
                  << appSettings->width << "x" << appSettings->height << std::endl;
    }

    // Destructor
    FluidSimulatorApp::~FluidSimulatorApp() {
        delete m_fluidRenderer;
        std::cout << "FluidSimulatorApp destroyed!" << std::endl;
    }

    void FluidSimulatorApp::FluidParameterControl()
    {
        float buttonWidth = ImGui::GetContentRegionAvail().x;
        
        ImGui::Spacing();

        // Collapsible Panel for Fluid Simulation Parameters
        if (ImGui::CollapsingHeader("Fluid Parameters", ImGuiTreeNodeFlags_DefaultOpen))
        {
            static bool autoUpdate = false; // Auto-update toggle
            static float particleMass = m_fluidSimulation->getParticleMass();
            static float restDensity = m_fluidSimulation->getRestDensity();
            static float viscosity = m_fluidSimulation->getViscosity();
            static float smoothingRadius = m_fluidSimulation->getSmoothingRadius();

            // Full-width sliders
            ImGui::PushItemWidth(-1); // Make sliders take up the full available width

            // Particle Mass
            ImGui::Text("Particle Mass");
            ImGui::SliderFloat("##ParticleMass", &particleMass, 0.01f, 1.4f, "%.3f");

            // Rest Density
            static int restDensityInt = static_cast<int>(restDensity);
            if (ImGui::SliderInt("##RestDensity", &restDensityInt, 500, 1500, "%d", ImGuiSliderFlags_AlwaysClamp))
            {
                restDensityInt = (restDensityInt / 20) * 20;
                restDensity = static_cast<float>(restDensityInt);
            }
            // Viscosity
            ImGui::Text("Viscosity");
            ImGui::SliderFloat("##Viscosity", &viscosity, 0.01f, 0.1f, "%.3f");

            // Smoothing Radius
            ImGui::Text("Smoothing Radius");
            ImGui::SliderFloat("##SmoothingRadius", &smoothingRadius, 0.1f, 1.0f, "%.2f");

            
            ImGui::PopItemWidth(); // Restore default width

            ImGui::Spacing();

            ImGui::Checkbox("Auto Update", &autoUpdate);

            // Only show "Apply" button if Auto Update is disabled
            if (!autoUpdate)
            {
                if (ImGui::Button("Apply", ImVec2(buttonWidth, 0)))
                {
                    m_fluidSimulation->setParticleMass(particleMass);
                    m_fluidSimulation->setRestDensity(restDensity);
                    m_fluidSimulation->setViscosity(viscosity);
                    m_fluidSimulation->setSmoothingRadius(smoothingRadius);
                    
                    // Update constants on the CUDA device
                    m_fluidSimulation->updateConstants();
                }
            }
            else
            {
                // Auto-update enabled; immediately apply changes
                m_fluidSimulation->setParticleMass(particleMass);
                m_fluidSimulation->setRestDensity(restDensity);
                m_fluidSimulation->setViscosity(viscosity);
                m_fluidSimulation->setSmoothingRadius(smoothingRadius);

                // Update constants on the CUDA device
                m_fluidSimulation->updateConstants();
            }

            if (ImGui::Button("Reset Parmeters", ImVec2(buttonWidth, 0)))
            {
                m_fluidSimulation->ResetFluidParameters();

                particleMass = m_fluidSimulation->getParticleMass();
                restDensity = m_fluidSimulation->getRestDensity();
                viscosity = m_fluidSimulation->getViscosity();
                smoothingRadius = m_fluidSimulation->getSmoothingRadius();
                static int restDensityInt = static_cast<int>(restDensity);
            }

        }

    }
    
    void FluidSimulatorApp::RenderParameterControl()
    {
        float buttonWidth = ImGui::GetContentRegionAvail().x;
        ImGui::Spacing();

        if (ImGui::CollapsingHeader("Screen Space Parameters", ImGuiTreeNodeFlags_DefaultOpen))
        {
            // Sliders for each parameter, storing locally
            static float pointScale   = m_fluidRenderer->getPointScale();
            static float pointRadius  = m_fluidRenderer->getPointRadius();
            static float sphereRadius = m_fluidRenderer->getSphereRadius();

            static float worldSmoothingRadius = m_fluidRenderer->getWorldSmoothingRadius();
            static float maxPixelRadius       = m_fluidRenderer->getMaxPixelRadius();
            static float spatialFactor        = m_fluidRenderer->getSpatialFactor();
            static float blurStrength         = m_fluidRenderer->getBlurStrength();
            static float depthThreshold       = m_fluidRenderer->getDepthThreshold();

            static float thicknessMass           = m_fluidRenderer->getThicknessMass();
            static int thicknessBlurRadius    = m_fluidRenderer->getThicknessBlurRadius();
            static float absorptionR           = m_fluidRenderer->getAbsorptionR();
            static float absorptionG           = m_fluidRenderer->getAbsorptionG();
            static float absorptionB           = m_fluidRenderer->getAbsorptionB();
            static float alphaScale           = m_fluidRenderer->getAlphaScale();

            static float shininess            = m_fluidRenderer->getShininess();
            static float fresnelPower         = m_fluidRenderer->getFresnelPower();
            static float fresnelScale         = m_fluidRenderer->getFresnelScale();
            static float fresnelBias          = m_fluidRenderer->getFresnelBias();

            static float fluidColorArray[3] = {1.0f, 1.0f, 1.0f}; 
            static bool once = false;
            if (!once)
            {
                float3 initialColor = m_fluidRenderer->getFluidColor();
                fluidColorArray[0] = initialColor.x;
                fluidColorArray[1] = initialColor.y;
                fluidColorArray[2] = initialColor.z;
                once = true;
            }

            ImGui::Text("Fluid Color"); ImGui::SameLine();
            ImGuiColorEditFlags colorFlags = ImGuiColorEditFlags_NoInputs
                                        | ImGuiColorEditFlags_NoLabel
                                        | ImGuiColorEditFlags_PickerHueWheel;
            if (ImGui::ColorEdit3("##FluidColor", fluidColorArray, colorFlags))
            {
                m_fluidRenderer->setFluidColor(
                    make_float3(fluidColorArray[0], fluidColorArray[1], fluidColorArray[2])
                );
            }

            // Begin Sliders
            ImGui::PushItemWidth(-1); // let sliders use full width

            ImGui::Text("Point Radius");
            if (ImGui::SliderFloat("##PointRadius", &pointRadius, 1.0f, 50.0f, "%.1f"))
            {
                m_fluidRenderer->setPointRadius(pointRadius);
            }

            ImGui::Text("Point Scale");
            if (ImGui::SliderFloat("##PointScale", &pointScale, 1.0f, 30.0f, "%.1f"))
            {
                m_fluidRenderer->setPointScale(pointScale);
            }

            ImGui::Text("Sphere Radius");
            if (ImGui::SliderFloat("##SphereRadius", &sphereRadius, 0.0f, 1.0f, "%.3f"))
            {
                m_fluidRenderer->setSphereRadius(sphereRadius);
            }

            // -- World Smoothing Radius
            ImGui::Text("World Smoothing Radius");
            if (ImGui::SliderFloat("##WorldSmoothRadius", &worldSmoothingRadius, 0.0f, 1.0f, "%.3f"))
            {
                m_fluidRenderer->setWorldSmoothingRadius(worldSmoothingRadius);
            }

            // -- Max Pixel Radius
            ImGui::Text("Max Pixel Radius");
            if (ImGui::SliderFloat("##MaxPixelRadius", &maxPixelRadius, 1.0f, 200.0f, "%.1f"))
            {
                m_fluidRenderer->setMaxPixelRadius(maxPixelRadius);
            }

            // -- Spatial Factor
            ImGui::Text("Spatial Factor");
            if (ImGui::SliderFloat("##SpatialFactor", &spatialFactor, 1.0f, 30.0f, "%.1f"))
            {
                m_fluidRenderer->setSpatialFactor(spatialFactor);
            }

            // -- Blur Strength
            ImGui::Text("Blur Strength");
            if (ImGui::SliderFloat("##BlurStrength", &blurStrength, 0.0f, 10.0f, "%.1f"))
            {
                m_fluidRenderer->setBlurStrength(blurStrength);
            }

            // -- Depth Threshold
            ImGui::Text("Depth Threshold");
            if (ImGui::SliderFloat("##DepthThreshold", &depthThreshold, 0.0f, 1.0f, "%.3f"))
            {
                m_fluidRenderer->setDepthThreshold(depthThreshold);
            }
            
            ImGui::Text("Thickness Mass");
            if (ImGui::SliderFloat("##ThicknessMass", &thicknessMass, 0.0f, 0.4f, "%.4f"))
            {
                m_fluidRenderer->setThicknessMass(thicknessMass);
            }

            // -- Thickness Blur Radius (int)
            ImGui::Text("Thickness Blur Radius");
            if (ImGui::SliderInt("##ThickBlurRadius", &thicknessBlurRadius, 0, 100))
            {
                m_fluidRenderer->setThicknessBlurRadius(thicknessBlurRadius);
            }

            // -- Absorption
            ImGui::Text("Absorption");
            if (ImGui::SliderFloat("##AbsorptionR", &absorptionR, 0.0f, 1.0f, "%.2f"))
            {
                m_fluidRenderer->setAbsorptionR(absorptionR);
            }
            if (ImGui::SliderFloat("##AbsorptionG", &absorptionG, 0.0f, 1.0f, "%.2f"))
            {
                m_fluidRenderer->setAbsorptionG(absorptionG);
            }
            if (ImGui::SliderFloat("##AbsorptionB", &absorptionB, 0.0f, 1.0f, "%.2f"))
            {
                m_fluidRenderer->setAbsorptionB(absorptionB);
            }
            
            // -- Alpha scale
            ImGui::Text("Alpha Scale");
            if (ImGui::SliderFloat("##AlphaScale", &alphaScale, 0.0f, 8.0f, "%.1f"))
            {
                m_fluidRenderer->setAlphaScale(alphaScale);
            }

            // -- Shininess
            ImGui::Text("Shininess");
            if (ImGui::SliderFloat("##Shininess", &shininess, 0.0f, 200.0f, "%.1f"))
            {
                m_fluidRenderer->setShininess(shininess);
            }

            // -- Fresnel Power
            ImGui::Text("Fresnel Power");
            if (ImGui::SliderFloat("##FresnelPower", &fresnelPower, 0.0f, 10.0f, "%.2f"))
            {
                m_fluidRenderer->setFresnelPower(fresnelPower);
            }

            // -- Fresnel Scale
            ImGui::Text("Fresnel Scale");
            if (ImGui::SliderFloat("##FresnelScale", &fresnelScale, 0.0f, 1.0f, "%.2f"))
            {
                m_fluidRenderer->setFresnelScale(fresnelScale);
            }

            // -- Fresnel Bias
            ImGui::Text("Fresnel Bias");
            if (ImGui::SliderFloat("##FresnelBias", &fresnelBias, 0.0f, 1.0f, "%.2f"))
            {
                m_fluidRenderer->setFresnelBias(fresnelBias);
            }

            ImGui::PopItemWidth();

            // Reset
            if (ImGui::Button("Reset Parameters", ImVec2(buttonWidth, 0)))
            {
                // Revert to your constant defaults
                m_fluidRenderer->setPointScale(h_POINT_SCALE);
                m_fluidRenderer->setPointRadius(h_POINT_RADIUS);
                pointScale   = m_fluidRenderer->getPointScale();
                pointRadius  = m_fluidRenderer->getPointRadius();

                m_fluidRenderer->setSphereRadius(h_SPHERE_RADIUS);
                sphereRadius = m_fluidRenderer->getSphereRadius();

                m_fluidRenderer->setWorldSmoothingRadius(h_worldRadiusMeters);
                worldSmoothingRadius = m_fluidRenderer->getWorldSmoothingRadius();

                m_fluidRenderer->setMaxPixelRadius(h_maxPixelRadius);
                maxPixelRadius = m_fluidRenderer->getMaxPixelRadius();

                m_fluidRenderer->setSpatialFactor(h_sigmaSpatialFactor);
                spatialFactor = m_fluidRenderer->getSpatialFactor();

                m_fluidRenderer->setBlurStrength(h_sigmaR);
                blurStrength = m_fluidRenderer->getBlurStrength();

                m_fluidRenderer->setDepthThreshold(h_depthThresh);
                depthThreshold = m_fluidRenderer->getDepthThreshold();
                
                m_fluidRenderer->setThicknessMass(h_thicknessMass);
                thicknessMass = m_fluidRenderer->getThicknessMass();

                m_fluidRenderer->setThicknessBlurRadius(h_thickBlurrRadius);
                thicknessBlurRadius = m_fluidRenderer->getThicknessBlurRadius();

                m_fluidRenderer->setAbsorptionR(h_absorbtionR);
                absorptionR = m_fluidRenderer->getAbsorptionR();
                m_fluidRenderer->setAbsorptionG(h_absorbtionG);
                absorptionG = m_fluidRenderer->getAbsorptionG();
                m_fluidRenderer->setAbsorptionB(h_absorbtionB);
                absorptionB = m_fluidRenderer->getAbsorptionB();
                
                m_fluidRenderer->setAlphaScale(h_alphaScale);
                absorptionB = m_fluidRenderer->getAlphaScale();

                m_fluidRenderer->setShininess(h_shininess);
                shininess = m_fluidRenderer->getShininess();

                m_fluidRenderer->setFresnelPower(h_fresnelPower);
                fresnelPower = m_fluidRenderer->getFresnelPower();

                m_fluidRenderer->setFresnelScale(h_fresnelScale);
                fresnelScale = m_fluidRenderer->getFresnelScale();

                m_fluidRenderer->setFresnelBias(h_fresnelBias);
                fresnelBias = m_fluidRenderer->getFresnelBias();

                // Reset fluid color
                float3 defaultColor = h_fluidColor;
                m_fluidRenderer->setFluidColor(defaultColor);
                fluidColorArray[0] = defaultColor.x;
                fluidColorArray[1] = defaultColor.y;
                fluidColorArray[2] = defaultColor.z;
            }
        }
    }

    
    void FluidSimulatorApp::RenderControlPannel()
    {
        
        // Move control to app laters
        ImGui::Begin("Controls");

        float buttonWidth = ImGui::GetContentRegionAvail().x;
        if (ImGui::Button(m_updatePhysics ? "Pause" : "Play", ImVec2(buttonWidth, 0)))
        {
            m_updatePhysics = !m_updatePhysics;
        }

        if (ImGui::Button("Reset", ImVec2(buttonWidth, 0))) {
            //m_updatePhysics = false;
            m_fluidSimulation->Reset();
        }

        ImGui::PushItemWidth(-1); // Make sliders take up the full available width
        
        ImGui::Separator();

        static int selectedOption = static_cast<int>(m_fluidSimulation->getSimulationMode()); // Sync with current mode

        const char* options[] = { "One Large Cube", "Two Small Cubes" };

        if (ImGui::Combo("##DropDownMenu", &selectedOption, options, IM_ARRAYSIZE(options)))
        {
            m_fluidSimulation->setSimulationMode(static_cast<SimulationMode>(selectedOption));
        }
        
        const char* stageOptions[] = {
            "Particle View",
            "Screeenn Spced",
            "Depth Buffer",
            "Filtered Depth",
            "Depth Normals",
            "Thickness Buffer",
            "Filtered Thickness",
            "Light Attenuation"
        };

        if (ImGui::Combo("Rendering Stage", reinterpret_cast<int*>(&selectedStage), stageOptions, IM_ARRAYSIZE(stageOptions)))
        {
            std::cout << "Selected Rendering Stage: " << static_cast<int>(selectedStage) << std::endl;
        }

        ImGui::Separator();

        static int stepsPerFrame = m_fluidSimulation->getSimulationStepsPerFrame();

        // Full-width sliders

        // Particle Mass
        ImGui::Text("Simulation steps/frame");
        if (ImGui::SliderInt("##SimpulationStepsPereFramee", &stepsPerFrame, 1, 8, "%d"))
        {
            m_fluidSimulation->setSimulationStepsPerFrame(stepsPerFrame);
        }

        ImGui::PopItemWidth();

        FluidParameterControl();

        RenderParameterControl();

        ImGui::End();

    }

    void FluidSimulatorApp::RenderFluidView()
    {
        if (m_updatePhysics)
        {
            m_fluidSimulation->updateSimulation();
        }
        
        size_t particleCount = m_fluidSimulation->getParticleCount();
        size_t boundaryCount = m_fluidSimulation->getBoundaryParticleCount();

        switch (selectedStage)
        {
            case RenderStage::ParticleView:
                m_fluidRenderer->renderNormalFrame(particleCount, boundaryCount);
                break;

            case RenderStage::DepthBuffer:
                m_fluidRenderer->renderDepthFrame(particleCount, boundaryCount);
                m_fluidRenderer->visualizeBuffer(RenderStage::DepthBuffer);
                break;

            case RenderStage::FilteredDepth:
                m_fluidRenderer->renderDepthFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderDepthFilterFrame();
                m_fluidRenderer->visualizeBuffer(RenderStage::FilteredDepth);
                break;

            case RenderStage::DepthNormals:
                m_fluidRenderer->renderDepthFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderDepthFilterFrame();
                m_fluidRenderer->renderDepthNormalsFrame();
                break;

            case RenderStage::ThicknessBuffer:
                m_fluidRenderer->renderThicknessFrame(particleCount, boundaryCount);
                m_fluidRenderer->visualizeBuffer(RenderStage::ThicknessBuffer);
                break;
            
            case RenderStage::FilteredThickness:
                m_fluidRenderer->renderThicknessFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderThicknessFilterFrame();
                m_fluidRenderer->visualizeBuffer(RenderStage::FilteredThickness);
                break;

            case RenderStage::LightAttenuation:
                m_fluidRenderer->renderThicknessFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderThicknessFilterFrame();
                m_fluidRenderer->renderThicknessAttenFrame();
                break;
            
            case RenderStage::ScreenSpacedRender:
                m_fluidRenderer->renderDepthFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderDepthFilterFrame();
                m_fluidRenderer->renderDepthNormalsFrame();
                
                m_fluidRenderer->renderThicknessFrame(particleCount, boundaryCount);
                m_fluidRenderer->renderThicknessFilterFrame();
                m_fluidRenderer->renderThicknessAttenFrame();
                
                m_fluidRenderer->renderFinalCombineFrame();
                break;
            
            default:
                std::cerr << "Unknown RenderStage selected." << std::endl;
                break;
        }

        // Display FBO texture in ImGui window
        ImGui::Begin("Simulation");

        ImVec2 availableSize = ImGui::GetContentRegionAvail();
        GLuint textureID = 0;

        if (selectedStage == RenderStage::ParticleView
            || selectedStage == RenderStage::DepthNormals
            || selectedStage == RenderStage::LightAttenuation
            || selectedStage == RenderStage::ScreenSpacedRender)
        {
            textureID = m_fluidRenderer->getTexture(selectedStage); // RGBA8
        }
        else if (selectedStage == RenderStage::DepthBuffer  
                || selectedStage == RenderStage::ThicknessBuffer
                || selectedStage == RenderStage::FilteredDepth
                || selectedStage == RenderStage::FilteredThickness)
        {
            // Use the visualization texture
            textureID = m_fluidRenderer->getVisualizationTexture();
        }
        else
        {
            textureID = 0; // Handle other stages if needed
        }

        if (textureID != 0)
        {
            ImGui::Image((ImTextureID)(uintptr_t)textureID, availableSize, ImVec2(0, 1), ImVec2(1, 0));
        }
        else
        {
            ImGui::Text("No texture available for the selected stage.");
        }

        ImGui::End();
    }

    // Render function
    void FluidSimulatorApp::Render() {
        
        RenderControlPannel();
        RenderFluidView();
    }

    // Factory function
    FluidSimulatorApp* creatFluidSimulatorApp(appSettings* settings) {
        return new FluidSimulatorApp(settings);
    }

} // namespace FluidSimulation
