#include <GL/glew.h>
#include <GLFW/glfw3.h>   // Use GLFW instead of GLUT

#include "imgui.h"
#include "imgui_impl_glfw.h"     // Use the GLFW backend
#include "imgui_impl_opengl3.h"  // Use the OpenGL3 backend

#include <cstdlib> // For exit()
#include <cstdio>

#include "include/FluidSimulationApp.h"

void SetCustomImGuiTheme()
{
    ImGuiStyle& style = ImGui::GetStyle();

    // General spacing and rounding for a clean look
    style.TabRounding = 2.0f; // Subtle rounding
    style.WindowRounding = 3.0f; // Slight rounding for dockable windows
    style.ChildRounding = 2.0f; // Consistent with tabs
    style.PopupRounding = 2.0f;
    style.FrameRounding = 3.0f; // For sliders and buttons
    style.GrabRounding = 2.0f; // For slider knobs

    style.WindowPadding = ImVec2(4, 4);
    style.FramePadding = ImVec2(4, 4);
    style.ItemSpacing = ImVec2(12, 8);
    style.ScrollbarSize = 15.0f;

    // Blender/Unity-like Colors (Darker gray with contrasting highlights)
    ImVec4* colors = style.Colors;

    // Tab colors
    colors[ImGuiCol_Tab]                  = ImVec4(0.18f, 0.18f, 0.18f, 1.00f); // Neutral dark gray
    colors[ImGuiCol_TabHovered]           = ImVec4(0.28f, 0.28f, 0.28f, 1.00f); // Slightly lighter when hovered
    colors[ImGuiCol_TabActive]            = ImVec4(0.25f, 0.25f, 0.25f, 1.00f); // Active tab
    colors[ImGuiCol_TabUnfocused]         = ImVec4(0.15f, 0.15f, 0.15f, 1.00f); // Background tabs
    colors[ImGuiCol_TabUnfocusedActive]   = ImVec4(0.20f, 0.20f, 0.20f, 1.00f); // Active but unfocused tab

    // Docking preview highlight (less intrusive)
    colors[ImGuiCol_DockingPreview]       = ImVec4(0.10f, 0.10f, 0.10f, 0.70f);

    // Window borders and docking separators
    colors[ImGuiCol_Border]               = ImVec4(0.20f, 0.20f, 0.20f, 1.00f); // Subtle border
    colors[ImGuiCol_BorderShadow]         = ImVec4(0.00f, 0.00f, 0.00f, 0.00f); // No shadow

    // Remove overly bright lines on docking highlights
    colors[ImGuiCol_DockingEmptyBg]       = ImVec4(0.14f, 0.14f, 0.14f, 1.00f);

    colors[ImGuiCol_Text]                   = ImVec4(0.90f, 0.90f, 0.90f, 1.00f); // White text
    colors[ImGuiCol_WindowBg]               = ImVec4(0.14f, 0.14f, 0.14f, 1.00f); // Dark gray background
    colors[ImGuiCol_Border]                 = ImVec4(0.29f, 0.29f, 0.29f, 0.50f); // Soft borders
    colors[ImGuiCol_FrameBg]                = ImVec4(0.21f, 0.21f, 0.21f, 1.00f); // Input background
    colors[ImGuiCol_FrameBgHovered]         = ImVec4(0.33f, 0.33f, 0.33f, 1.00f); // Hover effect
    colors[ImGuiCol_FrameBgActive]          = ImVec4(0.26f, 0.59f, 0.98f, 0.67f); // Active field
    colors[ImGuiCol_TitleBg]                = ImVec4(0.10f, 0.10f, 0.10f, 1.00f); // Title bar
    colors[ImGuiCol_TitleBgActive]          = ImVec4(0.15f, 0.15f, 0.15f, 1.00f); // Active title bar
    colors[ImGuiCol_Button]                 = ImVec4(0.20f, 0.20f, 0.20f, 1.00f); // Buttons
    colors[ImGuiCol_ButtonHovered]          = ImVec4(0.30f, 0.30f, 0.30f, 1.00f); // Hovered buttons
    colors[ImGuiCol_ButtonActive]           = ImVec4(0.26f, 0.59f, 0.98f, 1.00f); // Active buttons
    colors[ImGuiCol_Header]                 = ImVec4(0.20f, 0.20f, 0.20f, 1.00f); // Collapsible headers
    colors[ImGuiCol_HeaderHovered]          = ImVec4(0.26f, 0.26f, 0.26f, 1.00f);
    colors[ImGuiCol_HeaderActive]           = ImVec4(0.26f, 0.59f, 0.98f, 0.67f);
    colors[ImGuiCol_ScrollbarBg]            = ImVec4(0.10f, 0.10f, 0.10f, 1.00f);
    colors[ImGuiCol_ScrollbarGrab]          = ImVec4(0.30f, 0.30f, 0.30f, 1.00f);
    colors[ImGuiCol_ScrollbarGrabHovered]   = ImVec4(0.40f, 0.40f, 0.40f, 1.00f);
    colors[ImGuiCol_ScrollbarGrabActive]    = ImVec4(0.50f, 0.50f, 0.50f, 1.00f);
    colors[ImGuiCol_CheckMark]              = ImVec4(0.26f, 0.59f, 0.98f, 1.00f); // Checkbox mark
}

void GLFWKeyCallback(GLFWwindow* window, int key, int scancode, int action, int mods) {
    // Retrieve FluidSimulatorApp from window user pointer
    FluidSimulation::FluidSimulatorApp* app = static_cast<FluidSimulation::FluidSimulatorApp*>(glfwGetWindowUserPointer(window));
    if (!app) return;

    // Retrieve the camera from the app
    FluidSimulation::Camera* cam = app->getCamera();
    if (!cam) return;

    // Map GLFW keys to ASCII keys and process them
    if (action == GLFW_PRESS || action == GLFW_REPEAT) {
        unsigned char ascii_key;
        switch (key) {
            case GLFW_KEY_W: ascii_key = 'w'; break;
            case GLFW_KEY_S: ascii_key = 's'; break;
            case GLFW_KEY_A: ascii_key = 'a'; break;
            case GLFW_KEY_D: ascii_key = 'd'; break;
            case GLFW_KEY_J: ascii_key = 'j'; break;
            case GLFW_KEY_L: ascii_key = 'l'; break;
            case GLFW_KEY_I: ascii_key = 'i'; break;
            case GLFW_KEY_K: ascii_key = 'k'; break;
            case GLFW_KEY_SPACE: ascii_key = ' '; break;
            case GLFW_KEY_C: ascii_key = 'c'; break;
            case GLFW_KEY_V: ascii_key = 'v'; break;
            case GLFW_KEY_ESCAPE: ascii_key = 0x1b; break;
            default: return; // Key not handled
        }

        cam->processKey(ascii_key);
    }
}

void GLFWFramebufferSizeCallback(GLFWwindow* window, int width, int height) {
    // Retrieve FluidSimulatorApp from window user pointer
    FluidSimulation::FluidSimulatorApp* app = static_cast<FluidSimulation::FluidSimulatorApp*>(glfwGetWindowUserPointer(window));
    if (!app) return;

    // Update camera aspect ratio
    FluidSimulation::Camera* cam = app->getCamera();
    if (cam) cam->reshape(width, height);

    // Resize FBO
    FluidSimulation::FluidRenderer* renderer = app->getRenderer(); // Add this getter
    if (renderer) renderer->resizeFBO(width, height);
}


int main(int argc, char** argv) {
    // Initialize GLFW
    if (!glfwInit()) {
        fprintf(stderr, "Failed to initialize GLFW\n");
        return -1;
    }

    // Set OpenGL version (3.3 Core)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // Create GLFW window
    GLFWwindow* window = glfwCreateWindow(1280, 720, "Simulation with ImGui Docking & FBO", NULL, NULL);
    if (!window) {
        fprintf(stderr, "Failed to create GLFW window\n");
        glfwTerminate();
        return -1;
    }

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1); // Enable V-sync

    // Set initial viewport
    int initialWidth = 1280, initialHeight = 720;
    glViewport(0, 0, initialWidth, initialHeight);


    // Initialize FBO with initial window size (1280x720)

    // Instantiate ImGuiManager
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;

    // Enable Docking
    io.ConfigFlags |= ImGuiConfigFlags_DockingEnable;

    // Optional: Enable Multi-Viewport / Platform Windows
    // io.ConfigFlags |= ImGuiConfigFlags_ViewportsEnable;

    io.Fonts->AddFontDefault();
    ImFont* mainFont = io.Fonts->AddFontFromFileTTF("../external/imgui/Fonts/Roboto-Regular.ttf", 16.0f);
    if (mainFont != nullptr) {
        io.FontDefault = mainFont;
    } else {
        fprintf(stderr, "Failed To Load Font \n");
    }
    // Setup Dear ImGui style
    //ImGui::StyleColorsDark();
    // ImGui::StyleColorsClassic();
    SetCustomImGuiTheme();

    // Setup Platform/Renderer backends
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init("#version 330");

    // Set GLFW callbacks
    //glfwSetWindowUserPointer(window, cameraPtr);
    //glfwSetKeyCallback(window, GLFWKeyCallback);
    //glfwSetFramebufferSizeCallback(window, GLFWFramebufferSizeCallback);
    FluidSimulation::appSettings appSettings = {1280, 720};
    FluidSimulation::FluidSimulatorApp* fluidSimulator = FluidSimulation::creatFluidSimulatorApp(&appSettings);

    // Set GLFW callbacks
    glfwSetWindowUserPointer(window, fluidSimulator);
    glfwSetKeyCallback(window, GLFWKeyCallback);
    glfwSetFramebufferSizeCallback(window, GLFWFramebufferSizeCallback);

    // Main loop
    while (!glfwWindowShouldClose(window)) {
        // Poll events
        glfwPollEvents();

        // Start ImGui frame
        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        // Obtain ImGuiIO reference
        //ImGuiIO& io = imguiManager.getIO();

        // Create a full-screen dockspace
        ImGui::SetNextWindowPos(ImVec2(0, 0));
        ImGui::SetNextWindowSize(io.DisplaySize);
        ImGuiWindowFlags dockspace_flags = ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoCollapse |
                                           ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove |
                                           ImGuiWindowFlags_NoBringToFrontOnFocus | ImGuiWindowFlags_NoNavFocus;
        
        ImGui::Begin("MainDockSpaceWindow", NULL, dockspace_flags);
        ImGuiID dockspace_id = ImGui::GetID("MainDockSpace");
        ImGui::DockSpace(dockspace_id, ImVec2(0.0f, 0.0f), ImGuiDockNodeFlags_None);
        ImGui::End();

        // Render Applicatiion Here
        fluidSimulator->Render();

        // Render ImGui
        //imguiManager.render();
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap buffers
        glfwSwapBuffers(window);
    }

    // Cleanup ImGui and OpenGL resources is handled by ImGuiManager destructor

    // Cleanup your own resources
    //cleanup();
    //glDeleteFramebuffers(1, &fbo);
    //glDeleteTextures(1, &fboTexture);
    //glDeleteRenderbuffers(1, &depthRBO);
    delete fluidSimulator;
    // Terminate GLFW
    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}