// pcisph_kernels.h

#ifndef PCISPH_KERNELS_H
#define PCISPH_KERNELS_H

#include <cuda_runtime.h> // Required for float3 and CUDA types

namespace FluidSimulationGPU {

__global__ void initialDensity(
    float3* positions,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
);

// Function to compute density errors
__global__ void computeDensityError(
    float3* positions,
    float* densities,
    float* density_errors,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    // Boundary data
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
);

// Function to update pressures
__global__ void updatePressures(
    float* pressure_increments,
    float* density_errors,
    int numParticles
);

// Function to compute pressure forces
__global__ void computePressureForces(
    float3* positions,
    float* pressure_increments,
    float3* pressure_forces,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    // Boundary data
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
);

__global__ void correctBoundaryPenetrations(
    float3* positions_pred,            // Predicted fluid positions
    float3* velocities_pred,           // Predicted fluid velocities
    const float3* boundary_positions,   // Boundary particle positions
    const float3* boundary_normals,     // Boundary particle normals
    unsigned int* cellIndices,          // Cell indices for fluid particles
    unsigned int* boundaryCellStart,    // Start indices for boundary particles per cell
    unsigned int* boundaryCellEnd,      // End indices for boundary particles per cell
    int numParticles,                   // Number of fluid particles
    int gridSizeX, int gridSizeY, int gridSizeZ, // Grid dimensions
    float r0,                           // Penetration threshold (e.g., 0.5 * h)
    float normalDamping,                // Damping factor for velocity
    float epsilon,                      // Friction coefficient (0 <= epsilon <= 1)
    float delta                         // Elasticity coefficient (0 <= delta <= 1)
);

// Function to compute viscosity and external forces
__global__ void computeInitialForces(
    float3* positions,
    float3* velocities,
    float3* accelerations,
    float3* normals,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
);

__global__ void calculateNormals(
    float3* positions,
    float3* velocities,
    float3* normals,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
);

// Function to enforce boundary conditions
__global__ void enforceBoundary(
    float3* positions,
    float3* velocities,
    int numParticles,
    float damping
);

// Function to predict velocities and positions
__global__ void predictVelocitiesPositions(
    float3* velocities,
    float3* velocities_pred,
    float3* positions,
    float3* positions_pred,
    float3* accelerations,
    float timeStep,
    int numParticles
);

// Function to update velocities and positions based on pressure forces
__global__ void updateVelocitiesPositions(
    float3* velocities_pred,
    float3* positions_pred,
    float3* pressure_forces,
    float timeStep,
    int numParticles
);

// Host function for neighbor search (uses device pointers)
void neigbourHoodSearch(
    int nParticles,
    float3* d_pos,
    float3* d_vel,
    float3* d_posSorted,
    float3* d_velSorted,
    unsigned int* d_cellIndc,
    unsigned int* d_parIndc,
    unsigned int* d_cEnd,
    unsigned int* d_cStart,
    const int bpg,
    const int tpb,
    int& gridSizeX,
    int& gridSizeY,
    int& gridSizeZ
);

} // namespace FluidSimulationGPU

#endif // PCISPH_KERNELS_H
