// init_cu.cu

#include "include/cuda_device_vars.h"
#include "../settings/constants.h"

#include "include/pcisph_kernels.h"
#include "include/float3_helper_math.h"

#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <device_launch_parameters.h>
#include <cstdio>

namespace FluidSimulationGPU {

// Global Variables
cudaGraphicsResource_t cuda_vbo_resource;
cudaGraphicsResource_t cuda_vbo_boundary_resource;

// Constants
__constant__ float PI;
__constant__ float H;
__constant__ float H2;
__constant__ float K_PCI;
__constant__ float POLY6_CONST;
__constant__ float SPIKY_GRAD_CONST;
__constant__ float VISC_LAP_CONST;

__constant__ float POLY6_D1_CONST;
__constant__ float SURFACE_TENS_CONST;
__constant__ float SURFACE_TENS_TERM;
__constant__ float SURFACE_TENS_COEFFICIENT;

__constant__ float PARTICLE_MASS;
__constant__ float REST_DENSITY;
__constant__ float VISCOSITY;

// Boundary constants
__constant__ float BOUND_X_MIN;
__constant__ float BOUND_X_MAX;
__constant__ float BOUND_Y_MIN;
__constant__ float BOUND_Y_MAX;
__constant__ float BOUND_Z_MIN;
__constant__ float BOUND_Z_MAX;

// Device Arrays
float3* d_positions = nullptr;
float3* d_velocities = nullptr;

unsigned int* d_cellIndices = nullptr;
unsigned int* d_particleIndices = nullptr;

float3* d_positions_sorted = nullptr;
float3* d_velocities_sorted = nullptr;

unsigned int* d_cellStart = nullptr;
unsigned int* d_cellEnd = nullptr;

// Arrays for densities, pressures, accelerations
float* d_densities = nullptr;
float* d_density_errors = nullptr;
float* d_pressure_increments = nullptr;
float3* d_pressure_forces = nullptr;
float3* d_accelerations = nullptr;
float3* d_normals = nullptr;

// Additional arrays for PCISPH
float3* d_velocities_pred = nullptr;
float3* d_positions_pred = nullptr;
float* d_sum_gradW2 = nullptr;

// Boundary particles
float3* d_boundary_positions = nullptr;
float3* d_boundary_normals = nullptr;

unsigned int* d_boundaryCellStart = nullptr;
unsigned int* d_boundaryCellEnd = nullptr;

unsigned int* d_boundaryCellIndices = nullptr;
unsigned int* d_boundaryParticleIndices = nullptr;    

void ComputePCISPHStiffness(
    float timeStep,
    float particleMass,
    float restDensity,
    float kernelRadius,
    float restSpacing,
    float spiky_d1_normalization)
{
    float mass = restDensity * powf(kernelRadius/2.0f, 3.f);

    printf("Particle mass: %f \n", mass);
    printf("Rest Density: %f \n", restDensity);

    //cudaMemcpyToSymbol(PARTICLE_MASS, &mass, sizeof(float));
    //cudaMemcpyToSymbol(REST_DENSITY, &restDensity, sizeof(float));

    float dtm_rho = (timeStep * mass) / restDensity;
    float beta = 2.0f * dtm_rho * dtm_rho;


    float3 grad_sum     = make_float3(0.f, 0.f, 0.f);
    float  sum_dot_grad = 0.f;

    float kernelRadius2 = kernelRadius * kernelRadius;

    // Sample points in [-kernelRadius, +kernelRadius] in increments of restSpacing
    for (float z = -kernelRadius; z <= kernelRadius + 1e-5f; z += restSpacing)
    {
        for (float y = -kernelRadius; y <= kernelRadius + 1e-5f; y += restSpacing)
        {
            for (float x = -kernelRadius; x <= kernelRadius + 1e-5f; x += restSpacing)
            {
                float3 r = make_float3(x, y, z);
                float r2 = dot(r, r);

                if (r2 < kernelRadius2 && r2 > 0.0f)  // avoid r=0 to prevent division by 0
                {
                    float r_norm = sqrtf(r2);
                    
                    float3 grad = (r / r_norm) * spiky_d1_normalization * powf(kernelRadius - r_norm, 2.f);

                    // Accumulate sums
                    grad_sum += grad;
                    sum_dot_grad += dot(grad, grad);
                }
            }
        }
    }

    // Compute dot of the sum of gradients
    float dot_sum_grad = dot(grad_sum, grad_sum);

    // Combine to get grad_term:
    //  grad_term ~ - [ (Σ gradW)^2 + Σ (gradW·gradW) ]
    float grad_term = - (dot_sum_grad + sum_dot_grad);

    // Finally, PCISPH stiffness factor k_PCI ~ -1 / (beta * grad_term)
    float k_PCI = -1.f / (beta * grad_term);

    printf("K_PCi precalc: %f \n", k_PCI);
    cudaMemcpyToSymbol(K_PCI, &k_PCI, sizeof(float));
}

void initializeConstants()
{
    // Host constants
    const float h_H2 = h_H * h_H;
    const float h_POLY6_CONST = 315.0f / (64.0f * M_PI * powf(h_H, 9));
    const float h_SPIKY_GRAD_CONST = -45.0f / (M_PI * powf(h_H, 6));
    const float h_VISC_LAP_CONST = 45.0f / (M_PI * powf(h_H, 6));

    const float poly6_d1 = -945.f / (32.f * M_PI * powf(h_H, 9.f));

    const float surface_tens_normalization = 32.f / (M_PI * powf(h_H, 9.f));
    const float surface_tens_term = powf(h_H, 6.f) / 64.f;
    

    // Copy constants to device
    cudaMemcpyToSymbol(PI, &h_PI, sizeof(float));
    cudaMemcpyToSymbol(H, &h_H, sizeof(float));
    cudaMemcpyToSymbol(H2, &h_H2, sizeof(float));
    cudaMemcpyToSymbol(POLY6_CONST, &h_POLY6_CONST, sizeof(float));
    cudaMemcpyToSymbol(SPIKY_GRAD_CONST, &h_SPIKY_GRAD_CONST, sizeof(float));
    cudaMemcpyToSymbol(VISC_LAP_CONST, &h_VISC_LAP_CONST, sizeof(float));


    cudaMemcpyToSymbol(POLY6_D1_CONST, &poly6_d1, sizeof(float));
    cudaMemcpyToSymbol(SURFACE_TENS_CONST, &surface_tens_normalization, sizeof(float));
    cudaMemcpyToSymbol(SURFACE_TENS_TERM, &surface_tens_term, sizeof(float));
    cudaMemcpyToSymbol(SURFACE_TENS_COEFFICIENT, &h_surface_tenns_coef, sizeof(float));


    cudaMemcpyToSymbol(PARTICLE_MASS, &h_PARTICLE_MASS, sizeof(float));
    cudaMemcpyToSymbol(REST_DENSITY, &h_REST_DENSITY, sizeof(float));
    cudaMemcpyToSymbol(VISCOSITY, &h_VISCOSITY, sizeof(float));

    // Copy boundary constants to device
    cudaMemcpyToSymbol(BOUND_X_MIN, &h_BOUND_X_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_X_MAX, &h_BOUND_X_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MIN, &h_BOUND_Y_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Y_MAX, &h_BOUND_Y_MAX, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MIN, &h_BOUND_Z_MIN, sizeof(float));
    cudaMemcpyToSymbol(BOUND_Z_MAX, &h_BOUND_Z_MAX, sizeof(float));

    ComputePCISPHStiffness(
        0.006,
        h_PARTICLE_MASS,
        h_REST_DENSITY,
        h_H,
        h_H/2.0f,
        h_SPIKY_GRAD_CONST
    );
}

// Register VBO with CUDA
void registerVBOWithCUDA(unsigned int vbo)
{
    cudaGraphicsGLRegisterBuffer(&cuda_vbo_resource, vbo, cudaGraphicsMapFlagsNone);
}

void registerBoundaryVBOWithCUDA(unsigned int vbo)
{
    cudaGraphicsGLRegisterBuffer(&cuda_vbo_boundary_resource, vbo, cudaGraphicsMapFlagsNone);
}

// Unregister VBO from CUDA
void unregisterVBOFromCUDA()
{
    cudaGraphicsUnregisterResource(cuda_vbo_resource);
}

void unregisterBoundaryVBOFromCUDA()
{
    cudaGraphicsUnregisterResource(cuda_vbo_boundary_resource);
}

// Initialize device arrays
void initializeFluidDeviceArrays(
    const float3* h_positions,
    const float3* h_velocities,
    int numParticles
)
{
    // Allocate device memory for velocities
    cudaMalloc(&d_velocities, numParticles * sizeof(float3));
    cudaMemcpy(d_velocities, h_velocities, numParticles * sizeof(float3), cudaMemcpyHostToDevice);

    // Allocate device memory for predicted velocities and positions
    cudaMalloc(&d_velocities_pred, numParticles * sizeof(float3));
    cudaMalloc(&d_positions_pred, numParticles * sizeof(float3));

    // Allocate device memory for cell indices and particle indices
    cudaMalloc(&d_cellIndices, numParticles * sizeof(unsigned int));
    cudaMalloc(&d_particleIndices, numParticles * sizeof(unsigned int));

    // Allocate device memory for sorted positions and velocities
    cudaMalloc(&d_positions_sorted, numParticles * sizeof(float3));
    cudaMalloc(&d_velocities_sorted, numParticles * sizeof(float3));

    // Allocate cell start and end arrays
    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    cudaMalloc(&d_cellStart, maxCells * sizeof(unsigned int));
    cudaMalloc(&d_cellEnd, maxCells * sizeof(unsigned int));

    // Allocate device memory for densities, pressure increments, and accelerations
    cudaMalloc(&d_densities, numParticles * sizeof(float));
    cudaMalloc(&d_density_errors, numParticles * sizeof(float));
    cudaMalloc(&d_pressure_increments, numParticles * sizeof(float));
    cudaMalloc(&d_pressure_forces, numParticles * sizeof(float3));
    cudaMalloc(&d_accelerations, numParticles * sizeof(float3));
    cudaMalloc(&d_normals, numParticles * sizeof(float3));

    // Allocate device memory for sum_gradW2
    cudaMalloc(&d_sum_gradW2, numParticles * sizeof(float));

    // Initialize positions_pred to initial positions
    cudaMemcpy(d_positions_pred, h_positions, numParticles * sizeof(float3), cudaMemcpyHostToDevice);
}

void initializeBoundryDeviceArrays(
    const float3* h_boundaryPositions,
    const float3* h_boundaryNormals, // Renamed parameter
    int numBoundaryParticles
)
{
    // Allocate cell start and end arrays
    int gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    int gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    int gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;

    // Allocate memory for boundary positions
    //cudaMalloc(&d_boundary_positions, numBoundaryParticles * sizeof(float3));
    //cudaMemcpy(d_boundary_positions, h_boundaryPositions, numBoundaryParticles * sizeof(float3), cudaMemcpyHostToDevice);

    // Allocate memory for boundary normals
    cudaMalloc(&d_boundary_normals, numBoundaryParticles * sizeof(float3));
    cudaMemcpy(d_boundary_normals, h_boundaryNormals, numBoundaryParticles * sizeof(float3), cudaMemcpyHostToDevice);

    cudaMalloc(&d_boundaryCellIndices, numBoundaryParticles * sizeof(unsigned int));
    cudaMalloc(&d_boundaryParticleIndices, numBoundaryParticles * sizeof(unsigned int));
    
    cudaMalloc(&d_boundaryCellStart, maxCells * sizeof(unsigned int));
    cudaMalloc(&d_boundaryCellEnd, maxCells * sizeof(unsigned int));
}


// NeighbourHoodSearch for boundary particles 1 time at start since they dont move
void boundaryNBSearch(int numBoundaryParticles) {
    
    // Map the VBO resource to get a device pointer
    size_t num_bytes;
    cudaGraphicsMapResources(1, &cuda_vbo_boundary_resource, 0);
    cudaGraphicsResourceGetMappedPointer((void**)&d_boundary_positions, &num_bytes, cuda_vbo_boundary_resource);
    
    int gsX, gsY, gsZ;

    int threadsPerBlock = 256;
    int blocksPerGrid = (numBoundaryParticles + threadsPerBlock - 1) / threadsPerBlock;
    
    float3* d_boundary_positions_sorted;
    float3* d_boundary_normals_sorted;
    

    cudaMalloc(&d_boundary_positions_sorted, numBoundaryParticles * sizeof(float3));
    cudaMalloc(&d_boundary_normals_sorted, numBoundaryParticles * sizeof(float3));
    
    neigbourHoodSearch(
            numBoundaryParticles, 
            d_boundary_positions,
            d_boundary_normals,
            d_boundary_positions_sorted,
            d_boundary_normals_sorted,
            d_boundaryCellIndices,
            d_boundaryParticleIndices,
            d_boundaryCellEnd,
            d_boundaryCellStart,
            blocksPerGrid,
            threadsPerBlock,
            gsX,
            gsY,
            gsZ); 
    
    // Unmap the VBO resource
    cudaGraphicsUnmapResources(1, &cuda_vbo_boundary_resource, 0);
    
    cudaFree(d_boundary_positions_sorted);
    cudaFree(d_boundary_normals_sorted);
    
    // Move some of above init code for boundary to seeparate function
}

void resetSimulation(
    const float3* h_positions,
    const float3* h_velocities,
    int numParticles
)
{
    cudaMemcpy(d_velocities, h_velocities, numParticles * sizeof(float3), cudaMemcpyHostToDevice);
    
    // Initialize positions_pred to initial positions
    cudaMemcpy(d_positions_pred, h_positions, numParticles * sizeof(float3), cudaMemcpyHostToDevice);

    // Map the VBO resource
    cudaError_t err = cudaGraphicsMapResources(1, &cuda_vbo_resource, 0);
    if (err != cudaSuccess) {
        fprintf(stderr, "Failed to map VBO resource: %s\n", cudaGetErrorString(err));
        return;
    }

    // Check if d_positions is already assigned
    if (d_positions == nullptr) {
        size_t num_bytes = 0;

        // Get the mapped pointer for positions
        err = cudaGraphicsResourceGetMappedPointer((void**)&d_positions, &num_bytes, cuda_vbo_resource);
        if (err != cudaSuccess) {
            fprintf(stderr, "Failed to get mapped pointer: %s\n", cudaGetErrorString(err));
            cudaGraphicsUnmapResources(1, &cuda_vbo_resource, 0);
            return;
        }
    }

    // Copy new positions to the mapped pointer
    err = cudaMemcpy(d_positions, h_positions, numParticles * sizeof(float3), cudaMemcpyHostToDevice);
    if (err != cudaSuccess) {
        fprintf(stderr, "Failed to copy new positions: %s\n", cudaGetErrorString(err));
    }

    // Unmap the VBO resource
    err = cudaGraphicsUnmapResources(1, &cuda_vbo_resource, 0);
    if (err != cudaSuccess) {
        fprintf(stderr, "Failed to unmap VBO resource: %s\n", cudaGetErrorString(err));
    }
}

void updateConstantsOnDevice(float particleMass, float restDensity, float viscosity, float smoothingRadius) {
    // Recalculate derived constants
    float h_H2 = smoothingRadius * smoothingRadius;
    float h_POLY6_CONST = 315.0f / (64.0f * M_PI * powf(smoothingRadius, 9));
    float h_SPIKY_GRAD_CONST = -45.0f / (M_PI * powf(smoothingRadius, 6));
    float h_VISC_LAP_CONST = 45.0f / (M_PI * powf(smoothingRadius, 6));

    const float poly6_d1 = -945.f / (32.f * M_PI * powf(smoothingRadius, 9.f));
    const float surface_tens_normalization = 32.f / (M_PI * powf(smoothingRadius, 9.f));
    const float surface_tens_term = powf(smoothingRadius, 6.f) / 64.f;
    
    // Send constants to CUDA device
    cudaMemcpyToSymbol(PARTICLE_MASS, &particleMass, sizeof(float));
    cudaMemcpyToSymbol(REST_DENSITY, &restDensity, sizeof(float));
    cudaMemcpyToSymbol(VISCOSITY, &viscosity, sizeof(float));
    cudaMemcpyToSymbol(H, &smoothingRadius, sizeof(float));
    cudaMemcpyToSymbol(H2, &h_H2, sizeof(float));
    cudaMemcpyToSymbol(POLY6_CONST, &h_POLY6_CONST, sizeof(float));
    cudaMemcpyToSymbol(SPIKY_GRAD_CONST, &h_SPIKY_GRAD_CONST, sizeof(float));
    cudaMemcpyToSymbol(VISC_LAP_CONST, &h_VISC_LAP_CONST, sizeof(float));

    cudaMemcpyToSymbol(POLY6_D1_CONST, &poly6_d1, sizeof(float));
    cudaMemcpyToSymbol(SURFACE_TENS_CONST, &surface_tens_normalization, sizeof(float));
    cudaMemcpyToSymbol(SURFACE_TENS_TERM, &surface_tens_term, sizeof(float));
    
    // Add surfacetension coefficiant when implmented
    //cudaMemcpyToSymbol(SURFACE_TENS_COEFFICIENT, &h_surface_tenns_coef, sizeof(float));
    
    ComputePCISPHStiffness(
        0.006,
        particleMass,
        restDensity,
        smoothingRadius,
        smoothingRadius/2.0f,
        h_SPIKY_GRAD_CONST
    );
}

// Cleanup device arrays
void cleanupDeviceArrays()
{
    cudaFree(d_velocities);
    cudaFree(d_velocities_pred);
    cudaFree(d_positions_pred);

    cudaFree(d_cellIndices);
    cudaFree(d_particleIndices);

    cudaFree(d_positions_sorted);
    cudaFree(d_velocities_sorted);

    cudaFree(d_cellStart);
    cudaFree(d_cellEnd);

    cudaFree(d_densities);
    cudaFree(d_density_errors);
    cudaFree(d_pressure_increments);
    cudaFree(d_pressure_forces);
    cudaFree(d_accelerations);

    cudaFree(d_boundary_positions);
    cudaFree(d_boundary_normals);

    cudaFree(d_boundaryCellIndices);
    cudaFree(d_boundaryParticleIndices);
    cudaFree(d_boundaryCellEnd);
    cudaFree(d_boundaryCellStart);

    cudaFree(d_sum_gradW2);

}


} // namespace FluidSimulationGPU
