// pcisph_kernels.cu
#include "include/pcisph_kernels.h"
#include "include/smoothing_kernels.h"

#include "helper_macros.h" // Make sure this is included
#include "include/float3_helper_math.h"

#include "include/cuda_device_vars.h"
#include "../settings/constants.h"


#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <device_launch_parameters.h>

// Include Thrust for sorting and reductions
#include <thrust/device_ptr.h>
#include <thrust/sort.h>
#include <thrust/sequence.h>
#include <thrust/reduce.h>


namespace FluidSimulationGPU {



// Compute density errors (used in PCISPH)

__global__ void initialDensity(
    float3* positions,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;

    float3 pos_i = positions[idx];
    float density = 0.0f;

    unsigned int cellIndex = cellIndices[idx];
    unsigned int neighborCellIdx;

    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        unsigned int startIdx = cellStart[neighborCellIdx];
        unsigned int endIdx   = cellEnd[neighborCellIdx];

        if (startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = startIdx; j < endIdx; ++j) 
            {
                float3 pos_j = positions[j];
                float3 r = pos_i - pos_j;
                float r2 = dot(r, r);

                // Check if in range
                if (r2 < H2 && r2 > 0.0001f) {
                    float W = poly6Kernel(r2);
                    density += PARTICLE_MASS * W;
                }
            }
        }

        // Boundary neighbors
        unsigned int b_startIdx = boundaryCellStart[neighborCellIdx];
        unsigned int b_endIdx   = boundaryCellEnd[neighborCellIdx];

        if (b_startIdx != 0xFFFFFFFF)
        {
            for (unsigned int j = b_startIdx; j < b_endIdx; ++j) 
            {
                float3 pos_j = boundaryPositions[j];
                float3 r = pos_i - pos_j;
                float r2 = dot(r, r);
                if (r2 < H2 && r2 > 0.0001f) {
                    float W = poly6Kernel(r2);
                    density += PARTICLE_MASS * W; 
                }
            }
        }

    }
    FOREACH_NEIGHBOR_CELL_END();

    densities[idx] = density;
}

__global__ void computeDensityError(
    float3* positions,
    float* densities,
    float* density_errors,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    // Boundary data
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;

    float3 pos_i = positions[idx];
    float density = 0.0f;

    unsigned int cellIndex = cellIndices[idx];
    unsigned int neighborCellIdx;

    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        unsigned int startIdx = cellStart[neighborCellIdx];
        unsigned int endIdx   = cellEnd[neighborCellIdx];

        if (startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = startIdx; j < endIdx; ++j) 
            {
                float3 pos_j = positions[j];
                float3 r = pos_i - pos_j;
                float r2 = dot(r, r);

                // Check if in range
                if (r2 < H2 && r2 > 0.0001f) {
                    float W = poly6Kernel(r2);
                    density += PARTICLE_MASS * W;
                }
            }
        }

        // Boundary neighbors
        unsigned int b_startIdx = boundaryCellStart[neighborCellIdx];
        unsigned int b_endIdx   = boundaryCellEnd[neighborCellIdx];

        if (b_startIdx != 0xFFFFFFFF)
        {
            for (unsigned int j = b_startIdx; j < b_endIdx; ++j) 
            {
                float3 pos_j = boundaryPositions[j];
                float3 r = pos_i - pos_j;
                float r2 = dot(r, r);
                if (r2 < H2 && r2 > 0.0001f) {
                    float W = poly6Kernel(r2);
                    density += PARTICLE_MASS * W; 
                }
            }
        }

    }
    FOREACH_NEIGHBOR_CELL_END();

    densities[idx]      = density;
    density_errors[idx] = density - REST_DENSITY;
}

__global__ void calculateNormals(
    float3* positions,
    float3* velocities,
    float3* normals,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;
    
    float3 pos_i = positions[idx];
    float density = 0.0f;

    float3 normal = make_float3(0.f, 0.f, 0.f);

    unsigned int cellIndex = cellIndices[idx];
    unsigned int neighborCellIdx;

    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        unsigned int startIdx = cellStart[neighborCellIdx];
        unsigned int endIdx   = cellEnd[neighborCellIdx];

        if (startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = startIdx; j < endIdx; ++j) 
            {
                float3 pos_j = positions[j];
                float dens_j = densities[j];
                float3 r = pos_i - pos_j;
                float r2 = dot(r, r);
                
                normal += kernel_poly6_d1(r2) * r / dens_j;
                // poly6GradKernel
                // normal += poly6grad(r) / other_density; 
            }
        }

    }
    FOREACH_NEIGHBOR_CELL_END();

    normal *= H * PARTICLE_MASS * POLY6_D1_CONST;

    normals[idx] = normal;
}


// Update pressures based on density errors
__global__ void updatePressures(
    float* pressure_increments,
    float* density_errors,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float deltaP = -density_errors[idx] * K_PCI;
        pressure_increments[idx] = deltaP;
    }
}

__global__ void computePressureForces(
    float3* positions,
    float* pressure_increments,
    float3* pressure_forces,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    float3* boundaryPositions,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;

    float3 pos_i = positions[idx];
    float deltaP_i = pressure_increments[idx];

    float3 fPressure = make_float3(0.0f, 0.0f, 0.0f);

    unsigned int cellIndex = cellIndices[idx];
    unsigned int neighborCellIdx;

    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        // Fluid neighbors
        unsigned int startIdx = cellStart[neighborCellIdx];
        unsigned int endIdx   = cellEnd[neighborCellIdx];

        if (startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = startIdx; j < endIdx; ++j) 
            {
                // Avoid self-interaction
                if (j == idx) continue;

                float3 pos_j = positions[j];
                // Using mirrored pressure for fluid
                float deltaP_j = deltaP_i;  // Or pressure_increments[j], but mirrored works for PCISPH ass approximation

                float3 r = pos_i - pos_j;
                float distSq = dot(r, r);
                if (distSq < H2 && distSq > 0.0001f) 
                {
                    float dist  = sqrtf(distSq);
                    float gradW = spikyKernelGrad(dist);
                    float3 gradWij = gradW * (r / dist);
                    
                    float factor = - (deltaP_i + deltaP_j) / (REST_DENSITY * REST_DENSITY);
                    fPressure += factor * gradWij * (PARTICLE_MASS * PARTICLE_MASS);
                }
            }
        }

        // Boundary neighbors
        unsigned int b_startIdx = boundaryCellStart[neighborCellIdx];
        unsigned int b_endIdx   = boundaryCellEnd[neighborCellIdx];

        if (b_startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = b_startIdx; j < b_endIdx; ++j) 
            {
                float3 pos_j = boundaryPositions[j];
                // Mirror pressure: p_b = p_i
                float deltaP_j = deltaP_i;

                float3 r = pos_i - pos_j;
                float distSq = dot(r, r);
                if (distSq < H2 && distSq > 0.0001f) 
                {
                    float dist  = sqrtf(distSq);
                    float gradW = spikyKernelGrad(dist);
                    float3 gradWij = gradW * (r / dist);
                    
                    float factor = - (deltaP_i + deltaP_j) / (REST_DENSITY * REST_DENSITY);
                    fPressure += factor * gradWij * (PARTICLE_MASS * PARTICLE_MASS);
                }
            }
        }
    }
    FOREACH_NEIGHBOR_CELL_END();

    pressure_forces[idx] = fPressure;
}

__global__ void correctBoundaryPenetrations(
    float3* positions_pred,
    float3* velocities_pred,
    const float3* boundary_positions,
    const float3* boundary_normals,
    unsigned int* cellIndices,
    unsigned int* boundaryCellStart,
    unsigned int* boundaryCellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ,
    float r0,
    float normalDamping,
    float epsilon,
    float delta)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;

    float3 pos_i = positions_pred[idx];
    float3 vel_i = velocities_pred[idx];

    unsigned int cellIndex = cellIndices[idx];
    unsigned int neighborCellIdx;

    float w_sum = 0.0f;
    float depth_sum = 0.0f;
    float3 normal_sum = make_float3(0.0f, 0.0f, 0.0f);

    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        unsigned int b_startIdx = boundaryCellStart[neighborCellIdx];
        unsigned int b_endIdx   = boundaryCellEnd[neighborCellIdx];

        if (b_startIdx == 0xFFFFFFFF) continue;

        for (unsigned int b = b_startIdx; b < b_endIdx; ++b) 
        {
            float3 pos_b = boundary_positions[b];
            float3 n_b   = boundary_normals[b];
            float3 r     = pos_i - pos_b;
            float distSq = dot(r, r);
            float dist   = sqrtf(distSq);

            if (dist < r0 && dist > 1e-6f) {
                // Compute penetration weight
                float w = (r0 - dist) / r0;
                w_sum += w;
                depth_sum  += w * (r0 - dist);

                // Accumulate normals weighted by penetration
                normal_sum += w * n_b;
            }
        }
    }
    FOREACH_NEIGHBOR_CELL_END();

    // apply corrections
    if (w_sum > 0.0f) 
    {
        float3 n_c = normal_sum / w_sum;
        float n_len = sqrtf(dot(n_c, n_c));
        if (n_len > 1e-6f) {
            n_c /= n_len; // Normalize

            float correction = depth_sum / w_sum;

            // Correct the fluid particles position
            pos_i += n_c * correction;

            // Velocity into normal and tangential
            float  v_n         = dot(vel_i, n_c);
            float3 v_normal    = v_n * n_c;
            float3 v_tangential = vel_i - v_normal;

            // Apply friction and elasticity
            float3 v_new = epsilon * v_tangential - delta * v_normal;

            vel_i = v_new;
        }
    }

    positions_pred[idx]  = pos_i;
    velocities_pred[idx] = vel_i;
}

// Compute viscosity and external forces (excluding pressure forces)
__global__ void computeInitialForces(
    float3* positions,
    float3* velocities,
    float3* accelerations,
    float3* normals,
    float* densities,
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numParticles) return;

    float3 pos_i = positions[idx];
    float3 vel_i = velocities[idx];
    float3 normal_i = normals[idx];
    float dens_i = densities[idx];


    float3 fViscosity = make_float3(0.0f, 0.0f, 0.0f);
    float3 fGravity   = make_float3(0.0f, -10.0f, 0.0f) * PARTICLE_MASS;
    float3 st_cohesion = make_float3(0.0f, 0.0f, 0.0f);
    float3 st_curvature = make_float3(0.0f, 0.0f, 0.0f);
    // Current particle's cell
    unsigned int cellIndex = cellIndices[idx];
    // We'll store neighbor cell index here
    unsigned int neighborCellIdx;

    // ------------------------------------------------
    // Replace triple nested loop with the macro
    // ------------------------------------------------
    FOREACH_NEIGHBOR_CELL_BEGIN(cellIndex, gridSizeX, gridSizeY, gridSizeZ, neighborCellIdx)
    {
        unsigned int startIdx = cellStart[neighborCellIdx];
        unsigned int endIdx   = cellEnd[neighborCellIdx];

        if (startIdx != 0xFFFFFFFF) 
        {
            for (unsigned int j = startIdx; j < endIdx; ++j)
            {
                if (j == idx) continue;

                float3 pos_j = positions[j];
                float3 vel_j = velocities[j];
                float3 normal_j = normals[j];
                float dens_j = densities[j];

                float3 r = pos_i - pos_j;

                float distSq = dot(r, r);
                float dist = sqrtf(distSq);
                
                if (dist < H && dist > 0.0001f) 
                {

                    // Viscosity force
                    float3 velDiff = vel_j - vel_i;
                    float viscLap  = viscosityKernelLap(dist);
                    fViscosity += VISCOSITY * velDiff * viscLap * PARTICLE_MASS / REST_DENSITY;
                
                    // Implement surface tension here

                    float st_correction_factor = 2.f * REST_DENSITY / (dens_i + dens_j);

                    // -> surface tension (cohesion)
                    float st_kernel = kernel_surface_tension(dist);
                    float3 direction = r/dist;
                    st_cohesion += st_correction_factor * st_kernel * direction;

                    // -> surface tension (curvature)
                    st_curvature += st_correction_factor * (normal_i - normal_j);
                }
            }
        }
    }
    FOREACH_NEIGHBOR_CELL_END();

    // finish surface tensiont calculations
    //st_cohesion *= -SURFACE_TENS_COEFFICIENT * PARTICLE_MASS * PARTICLE_MASS * SURFACE_TENS_CONST;
    //st_curvature *= -SURFACE_TENS_COEFFICIENT * PARTICLE_MASS;
    st_cohesion *= 0.f;
    st_curvature *= 0.f;
    float3 fSurfaceTension = (st_cohesion + st_curvature);
    
    // Total acceleration (excluding pressure)
    accelerations[idx] = (fViscosity + fGravity) / PARTICLE_MASS;
}

// Enforce boundary conditions, very simple should bee improved.
__global__ void enforceBoundary(
    float3* positions,
    float3* velocities,
    int numParticles,
    float damping
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // X-axis
        if (positions[idx].x < BOUND_X_MIN) {
            positions[idx].x = BOUND_X_MIN;
            velocities[idx].x *= -damping;
        }
        if (positions[idx].x > BOUND_X_MAX) {
            positions[idx].x = BOUND_X_MAX;
            velocities[idx].x *= -damping;
        }

        // Y-axis
        if (positions[idx].y < BOUND_Y_MIN) {
            positions[idx].y = BOUND_Y_MIN;
            velocities[idx].y *= -damping;
        }
        if (positions[idx].y > BOUND_Y_MAX) {
            positions[idx].y = BOUND_Y_MAX;
            velocities[idx].y *= -damping;
        }

        // Z-axis
        if (positions[idx].z < BOUND_Z_MIN) {
            positions[idx].z = BOUND_Z_MIN;
            velocities[idx].z *= -damping;
        }
        if (positions[idx].z > BOUND_Z_MAX) {
            positions[idx].z = BOUND_Z_MAX;
            velocities[idx].z *= -damping;
        }
    }
}

// Kernel to predict velocities and positions
__global__ void predictVelocitiesPositions(
    float3* velocities,
    float3* velocities_pred,
    float3* positions,
    float3* positions_pred,
    float3* accelerations,
    float timeStep,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < numParticles) {
        // Compute predicted velocity
        velocities_pred[idx] = velocities[idx] + accelerations[idx] * timeStep;
        // Compute predicted position
        positions_pred[idx] = positions[idx] + velocities_pred[idx] * timeStep;
    }
}

// Update velocities and positions based on pressure forces in 
// pressure correcting loop. 
__global__ void updateVelocitiesPositions(
    float3* velocities_pred,
    float3* positions_pred,
    float3* pressure_forces,
    float timeStep,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        // Update predicted velocity
        velocities_pred[idx] += (pressure_forces[idx] / PARTICLE_MASS) * timeStep;

        // Update predicted position
        positions_pred[idx] += (pressure_forces[idx] / PARTICLE_MASS) * (timeStep * timeStep);
    }
}


} // namespace FluidSimulationGPU