#include "include/cuda_device_vars.h"
#include "include/pcisph_kernels.h"
#include "include/spatialHash_NBSearch_kernels.h"
#include "../settings/constants.h"

#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <device_launch_parameters.h>
#include <thrust/device_ptr.h>
#include <thrust/reduce.h>

namespace FluidSimulationGPU {

void updateParticles(float deltaTime)
{
    // Map the VBO resource to get a device pointer
    size_t num_bytes;
    cudaGraphicsMapResources(1, &cuda_vbo_resource, 0);
    cudaGraphicsResourceGetMappedPointer((void**)&d_positions, &num_bytes, cuda_vbo_resource);

    int threadsPerBlock = 256;
    int blocksPerGrid = (NUM_PARTICLES + threadsPerBlock - 1) / threadsPerBlock;

    // Time step should be small but we can run updateParticles multiple times during display
    float timeStep = 0.006f;

    // -------------------- Neighbor Search --------------------
    int gridSizeX, gridSizeY, gridSizeZ;
    neigbourHoodSearch(
        NUM_PARTICLES,
        d_positions,
        d_velocities,
        d_positions_sorted,
        d_velocities_sorted,
        d_cellIndices,
        d_particleIndices,
        d_cellEnd,
        d_cellStart,
        blocksPerGrid,
        threadsPerBlock,
        gridSizeX,
        gridSizeY,
        gridSizeZ);

    // -------------------- Prediction Step --------------------
    initialDensity<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions_pred,
        d_densities,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        d_boundary_positions,
        d_boundaryCellStart,
        d_boundaryCellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    calculateNormals<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        d_normals,
        d_densities,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    computeInitialForces<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        d_accelerations,
        d_normals,
        d_densities,
        d_cellIndices,
        d_cellStart,
        d_cellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    predictVelocitiesPositions<<<blocksPerGrid, threadsPerBlock>>>(
        d_velocities,
        d_velocities_pred,
        d_positions,
        d_positions_pred,
        d_accelerations,
        timeStep,
        NUM_PARTICLES);
    cudaDeviceSynchronize();

    // Initialize pressure increments to zero
    cudaMemset(d_pressure_increments, 0, NUM_PARTICLES * sizeof(float));

    float r0 = 0.5f * h_H;          // Penetration threshold
    float epsilon = 1.0f; // Moderate to high friction
    float delta = 0.5f;   // Low to moderate elasticity
    float normalDamping = 0.5f; // Adjust as needed

    // ------------ Iterative Pressure Correction Loop --------------------

    const int maxIterations = 3;

    for (int iter = 0; iter < maxIterations; ++iter) {
    // Launch the boundary correction kernel
        computeDensityError<<<blocksPerGrid, threadsPerBlock>>>(
            d_positions_pred,
            d_densities,
            d_density_errors,
            d_cellIndices,
            d_cellStart,
            d_cellEnd,
            d_boundary_positions,
            d_boundaryCellStart,
            d_boundaryCellEnd,
            NUM_PARTICLES,
            gridSizeX, gridSizeY, gridSizeZ);
        cudaDeviceSynchronize();
        
        correctBoundaryPenetrations<<<blocksPerGrid, threadsPerBlock>>>(
            d_positions_pred,
            d_velocities_pred,
            d_boundary_positions,
            d_boundary_normals,
            d_cellIndices,
            d_boundaryCellStart,
            d_boundaryCellEnd,
            NUM_PARTICLES,
            gridSizeX, gridSizeY, gridSizeZ,
            r0,
            normalDamping,
            epsilon,
            delta
        );
        cudaDeviceSynchronize();


        updatePressures<<<blocksPerGrid, threadsPerBlock>>>(
            d_pressure_increments,
            d_density_errors,
            NUM_PARTICLES);
        cudaDeviceSynchronize();

        computePressureForces<<<blocksPerGrid, threadsPerBlock>>>(
            d_positions_pred,
            d_pressure_increments,
            d_pressure_forces,
            d_cellIndices,
            d_cellStart,
            d_cellEnd,
            d_boundary_positions,
            d_boundaryCellStart,
            d_boundaryCellEnd,
            NUM_PARTICLES,
            gridSizeX, gridSizeY, gridSizeZ);
        cudaDeviceSynchronize();

        updateVelocitiesPositions<<<blocksPerGrid, threadsPerBlock>>>(
            d_velocities_pred,
            d_positions_pred,
            d_pressure_forces,
            timeStep,
            NUM_PARTICLES);
        cudaDeviceSynchronize();
        
    }

    
    correctBoundaryPenetrations<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions_pred,
        d_velocities_pred,
        d_boundary_positions,
        d_boundary_normals,
        d_cellIndices,
        d_boundaryCellStart,
        d_boundaryCellEnd,
        NUM_PARTICLES,
        gridSizeX, gridSizeY, gridSizeZ,
        r0,
        normalDamping,
        epsilon,
        delta
    );
    
    cudaDeviceSynchronize();

    // -------------------- Finalization --------------------

    cudaMemcpy(d_velocities, d_velocities_pred, NUM_PARTICLES * sizeof(float3), cudaMemcpyDeviceToDevice);
    cudaMemcpy(d_positions, d_positions_pred, NUM_PARTICLES * sizeof(float3), cudaMemcpyDeviceToDevice);

    // Enforce boundary conditions
    /*
    float damping = 0.75f;
    enforceBoundary<<<blocksPerGrid, threadsPerBlock>>>(
        d_positions,
        d_velocities,
        NUM_PARTICLES,
        damping);
    cudaDeviceSynchronize();
    */
    

    // Unmap the VBO resource
    cudaGraphicsUnmapResources(1, &cuda_vbo_resource, 0);
}

} // namespace FluidSimulationGPU
