#include "include/spatialHash_NBSearch_kernels.h"
#include "include/cuda_device_vars.h"
#include "../settings/constants.h"

#include <thrust/device_ptr.h>
#include <thrust/sort.h>
#include <thrust/sequence.h>
#include <thrust/reduce.h>

namespace FluidSimulationGPU {

// Compute cell indices
__global__ void computeCellIndices(
    float3* positions,
    unsigned int* cellIndices,
    int numParticles,
    float3 gridMin,
    float3 gridCellSize,
    int gridSizeX, int gridSizeY, int gridSizeZ
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        float3 pos = positions[idx];

        int ix = (int)((pos.x - gridMin.x) / gridCellSize.x);
        int iy = (int)((pos.y - gridMin.y) / gridCellSize.y);
        int iz = (int)((pos.z - gridMin.z) / gridCellSize.z);

        // Clamp indices to grid bounds
        ix = max(0, min(ix, gridSizeX - 1));
        iy = max(0, min(iy, gridSizeY - 1));
        iz = max(0, min(iz, gridSizeZ - 1));

        unsigned int cellIndex = ix + iy * gridSizeX + iz * gridSizeX * gridSizeY;

        cellIndices[idx] = cellIndex;
    }
}

// Reordering kernel
__global__ void reorderData(
    unsigned int* particleIndices,
    float3* positions,
    float3* velocities,
    float3* positions_sorted,
    float3* velocities_sorted,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int sortedIdx = particleIndices[idx];

        positions_sorted[idx] = positions[sortedIdx];
        velocities_sorted[idx] = velocities[sortedIdx];
    }
}

// Find cell start and end indices
__global__ void findCellStartEnd(
    unsigned int* cellIndices,
    unsigned int* cellStart,
    unsigned int* cellEnd,
    int numParticles
)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < numParticles) {
        unsigned int cellIndex = cellIndices[idx];

        if (idx == 0) {
            cellStart[cellIndex] = 0;
        } else {
            unsigned int prevCellIndex = cellIndices[idx - 1];
            if (prevCellIndex != cellIndex) {
                cellStart[cellIndex] = idx;
                cellEnd[prevCellIndex] = idx;
            }
        }

        if (idx == numParticles - 1) {
            cellEnd[cellIndex] = numParticles;
        }
    }
}

void neigbourHoodSearch(
        int nParticles, 
        float3* d_pos,
        float3* d_vel,
        float3* d_posSorted,
        float3* d_velSorted,
        unsigned int* d_cellIndc,
        unsigned int* d_parIndc,
        unsigned int* d_cEnd,
        unsigned int* d_cStart,
        const int bpg,
        const int tpb,
        int &gridSizeX,
        int &gridSizeY,
        int &gridSizeZ) 
{

    // Compute grid size based on smoothing radius
    float3 gridMin = make_float3(h_BOUND_X_MIN, h_BOUND_Y_MIN, h_BOUND_Z_MIN);
    float3 gridMax = make_float3(h_BOUND_X_MAX, h_BOUND_Y_MAX, h_BOUND_Z_MAX);
    float3 gridCellSize = make_float3(h_H, h_H, h_H);

    gridSizeX = (int)ceil((h_BOUND_X_MAX - h_BOUND_X_MIN) / h_H);
    gridSizeY = (int)ceil((h_BOUND_Y_MAX - h_BOUND_Y_MIN) / h_H);
    gridSizeZ = (int)ceil((h_BOUND_Z_MAX - h_BOUND_Z_MIN) / h_H);
    int maxCells = gridSizeX * gridSizeY * gridSizeZ;
    
    // Compute cell indices for current positions
    computeCellIndices<<<bpg, tpb>>>(
        d_pos, d_cellIndc, nParticles, gridMin, gridCellSize, gridSizeX, gridSizeY, gridSizeZ);
    cudaDeviceSynchronize();

    // Wrap device pointers with thrust iterators
    thrust::device_ptr<unsigned int> dev_cellIndices_ptr(d_cellIndc);
    thrust::device_ptr<unsigned int> dev_particleIndices_ptr(d_parIndc);

    // Initialize particle indices
    thrust::sequence(dev_particleIndices_ptr, dev_particleIndices_ptr + nParticles);

    // Sort particle indices based on cell indices
    thrust::sort_by_key(dev_cellIndices_ptr, dev_cellIndices_ptr + nParticles, dev_particleIndices_ptr);

    // Reorder positions and velocities
    reorderData<<<bpg, tpb>>>(
        d_parIndc,
        d_pos,
        d_vel,
        d_posSorted,
        d_velSorted,
        nParticles);
    cudaDeviceSynchronize();

    // Copy sorted data back to original arrays
    cudaMemcpy(d_pos, d_posSorted, nParticles * sizeof(float3), cudaMemcpyDeviceToDevice);
    cudaMemcpy(d_vel, d_velSorted, nParticles * sizeof(float3), cudaMemcpyDeviceToDevice);

    // Build cell start and end indices
    cudaMemset(d_cStart, 0xFF, maxCells * sizeof(unsigned int));
    cudaMemset(d_cEnd, 0xFF, maxCells * sizeof(unsigned int));

    findCellStartEnd<<<bpg, tpb>>>(
        d_cellIndc,
        d_cStart,
        d_cEnd,
        nParticles);
    cudaDeviceSynchronize();
}

} // namespace FluidSimulationGPU
