// FluidRenderer.h
#pragma once

#include <GL/glew.h>
#include <vector>
#include <cuda_runtime.h>
#include "include/Camera.h"
#include "settings/constants.h"

namespace FluidSimulation {

// Enumeration for different rendering stages
enum class RenderStage {
    ParticleView,
    ScreenSpacedRender,
    DepthBuffer,
    FilteredDepth,
    DepthNormals,
    ThicknessBuffer,
    FilteredThickness,
    LightAttenuation
};

// Structure to encapsulate shader program details
struct ShaderProgram {
    // Shader program ID
    GLuint programID; 
    // Attribute locations
    GLuint positionAttribLoc;

    // Uniform locations
    GLuint modelMatrixLoc;
    GLuint viewMatrixLoc;
    GLuint projectionMatrixLoc;
    GLuint pointRadiusLoc;
    GLuint pointScaleLoc;
    GLuint uColorLoc;
    GLuint sphereRadiusLoc;  // For depth shader

    // New Uniform Locations for Depth Normalization
    GLuint nearPlaneLoc;
    GLuint farPlaneLoc;

    // Framebuffer Object and its attachments
    GLuint fbo;
    GLuint fboTexture;
    GLuint depthRBO;

    // Constructor to initialize members
    ShaderProgram()
        : programID(0),
          positionAttribLoc(0),
          modelMatrixLoc(0),
          viewMatrixLoc(0),
          projectionMatrixLoc(0),
          pointRadiusLoc(0),
          pointScaleLoc(0),
          uColorLoc(0),
          sphereRadiusLoc(0),
          nearPlaneLoc(0),
          farPlaneLoc(0),
          fbo(0),
          fboTexture(0),
          depthRBO(0) {}
};

class FluidRenderer {
private:
    // Shader programs
    ShaderProgram m_normalRenderProgram;     // For particle shaders
    ShaderProgram m_finalCombineProgram;

    ShaderProgram m_depthRenderProgram;      // For depth shaders
    ShaderProgram m_depthFilterProgram;
    ShaderProgram m_depthNormalsProgram;
    
    ShaderProgram m_thicknessRenderProgram; // For thickness shaders
    ShaderProgram m_thicknessFilterProgram; 
    ShaderProgram m_thicknessAttenProgram; 

    ShaderProgram m_visualizationProgram;    // For visualization shaders

    GLuint m_depthFilterTempTex;
    GLuint m_depthFilterTempFBO; // temporary fbo forfilter passes

    GLuint m_thicknessFilterTempTex;
    GLuint m_thicknessFilterTempFBO;
    
    // VBOs for particles and boundaries
    GLuint m_particleVBO;
    GLuint m_boundaryVBO;

    // Framebuffer dimensions
    int m_fbWidth;
    int m_fbHeight;

    // Pointer to the Camera object
    Camera* m_camera;

    GLuint particles_vao;

    // Fullscreen quad VAO
    GLuint quadVAO;

    float m_pointScale = h_POINT_SCALE;
    float m_pointRadius = h_POINT_RADIUS;
    
    float m_sphereRadius = h_SPHERE_RADIUS;

    float m_worldSmoothingRadius = h_worldRadiusMeters;
    float m_maxPixelRadius = h_maxPixelRadius;
    float m_spatialFactor = h_sigmaSpatialFactor;
    float m_blurrStrength = h_sigmaR;
    float m_deptThreshold = h_depthThresh;

    float m_thicknessMass = h_thicknessMass;
    int thicknesBlurrRadius = h_thickBlurrRadius;
    float m_absorbtionR = h_absorbtionR;
    float m_absorbtionG = h_absorbtionG;
    float m_absorbtionB = h_absorbtionB;
    float m_alphaScale = h_alphaScale;

    float3 m_fluidColor = h_fluidColor;   
    float m_shininess = h_shininess;

    float m_fresnelPower = h_fresnelPower;
    float m_fresnelScale = h_fresnelScale;
    float m_fresnelBias  = h_fresnelBias;

    // Initialization helper methods
    void initGLEW();                    // Initialize GLEW
    void initOpenGLSettings();          // Set up OpenGL state
    void initShaders();                 // Load and configure shaders
    void initShaderUniforms();          // Configure shader uniforms
    void initializeFramebuffer(ShaderProgram& shaderProgram, GLenum textureInternalFormat, GLenum textureFormat, GLenum textureType);
    void initFBOs();                    // Set up framebuffers for each shader program
    void initFullscreenQuad();          // Set up fullscreen quad for visualization
    void setupShaderProgram(ShaderProgram& shaderProgram);

    void computeGaussianWeights(int radius, std::vector<float>& outWeights);
    
    void deleteFramebufferResources(ShaderProgram& shaderProgram);
    void resizeFramebuffer(
        ShaderProgram& shaderProgram,
        GLenum textureInternalFormat,
        GLenum textureFormat,
        GLenum textureType,
        int newWidth,
        int newHeight);

    void renderPoints(
        GLuint vbo,
        size_t count,
        const ShaderProgram& shaderProgram,
        const vec3& color);
    // Rendering helper methods
    void renderParticles(size_t particleCount, const ShaderProgram& shader);
    void renderBoundaries(size_t boundaryCount, const ShaderProgram& shader);

public:
    // Constructor and Destructor
    FluidRenderer(int width, int height, Camera* camera);
    ~FluidRenderer();

    // Public methods
    void init();                                // Consolidated initialization function
    void resizeFBO(int newWidth, int newHeight); // Resize framebuffers
    void setVBOs(GLuint particleVBO, GLuint boundaryVBO); // Set VBOs

    // Separate rendering functions
    void renderNormalFrame(size_t particleCount, size_t boundaryCount); // Render using Normal Shaders
    
    void renderDepthFrame(size_t particleCount, size_t boundaryCount);  // Render using Depth Shaders
    void renderDepthFilterFrame();
    void renderDepthNormalsFrame();

    void renderThicknessFrame(size_t particleCount, size_t boundaryCount);
    void renderThicknessFilterFrame();
    void renderThicknessAttenFrame();

    void renderFinalCombineFrame();
    void visualizeBuffer(RenderStage stage);                           // Visualize selected render stage

    // Methods to retrieve textures based on RenderStage
    GLuint getTexture(RenderStage stage) const;            // Get original FBO texture
    GLuint getVisualizationTexture() const;                // Get visualization FBO texture

    float getPointScale() { return m_pointScale; };
    float getPointRadius() { return m_pointRadius; };
    void setPointScale(float pScale) { m_pointScale = pScale; };
    void setPointRadius(float pRadius) { m_pointRadius = pRadius; };

    float getSphereRadius() const { return m_sphereRadius; }
    void  setSphereRadius(float radius) { m_sphereRadius = radius; }

    float getWorldSmoothingRadius() const { return m_worldSmoothingRadius; }
    void  setWorldSmoothingRadius(float worldRadius) { m_worldSmoothingRadius = worldRadius; }

    float getMaxPixelRadius() const { return m_maxPixelRadius; }
    void  setMaxPixelRadius(float maxPixRad) { m_maxPixelRadius = maxPixRad; }

    float getSpatialFactor() const { return m_spatialFactor; }
    void  setSpatialFactor(float spatialFactor) { m_spatialFactor = spatialFactor; }

    float getThicknessMass() const { return m_thicknessMass; }
    void  setThicknessMass(float thicknessMass) { m_thicknessMass = thicknessMass; }
    
    float getBlurStrength() const { return m_blurrStrength; }
    void  setBlurStrength(float blurStrength) { m_blurrStrength = blurStrength; }

    float getDepthThreshold() const { return m_deptThreshold; }
    void  setDepthThreshold(float depthThresh) { m_deptThreshold = depthThresh; }

    int getThicknessBlurRadius() const { return thicknesBlurrRadius; }
    void setThicknessBlurRadius(int blurRadius) { thicknesBlurrRadius = blurRadius; }

    float getAbsorptionR() const { return m_absorbtionR; }
    void  setAbsorptionR(float absorptionR) { m_absorbtionR = absorptionR; }

    float getAbsorptionG() const { return m_absorbtionG; }
    void  setAbsorptionG(float absorptionG) { m_absorbtionG = absorptionG; }
    
    float getAbsorptionB() const { return m_absorbtionB; }
    void  setAbsorptionB(float absorptionB) { m_absorbtionB = absorptionB; }
    
    float getAlphaScale() const { return m_alphaScale; }
    void  setAlphaScale(float alphaScale) { m_alphaScale = alphaScale; }
    
    float getShininess() const { return m_shininess; }
    void  setShininess(float shininess) { m_shininess = shininess; }

    float getFresnelPower() const { return m_fresnelPower; }
    void  setFresnelPower(float power) { m_fresnelPower = power; }

    float getFresnelScale() const { return m_fresnelScale; }
    void  setFresnelScale(float scale) { m_fresnelScale = scale; }

    float getFresnelBias() const { return m_fresnelBias; }
    void  setFresnelBias(float bias) { m_fresnelBias = bias; }

    float3 getFluidColor() const { return m_fluidColor; };
    void setFluidColor(float3 newColor) { m_fluidColor = newColor; };
};

} // namespace FluidSimulation
