// spawn_particles.cpp
#include "include/spawn_particles.h"
#include "constants.h"

#include <cmath>          // For std::pow
#include <ctime>
#include <vector>
#include <cstdlib>

namespace FluidSimulation {
// Initialize fluid particles
void initParticles( 
    std::vector<float3> &h_positions,
    std::vector<float3> &h_velocities,
    const int NUM_PARTICLES,
    const float GRID_SPACING,
    SimulationMode mode
)
{
    // Seed the random number generator
    srand(static_cast<unsigned int>(time(0)));

    auto initializeCube = [&](int startIndex, int particlesPerAxis, float offsetX, float offsetY, float offsetZ) {
        int index = startIndex;
        for (int i = 0; i < particlesPerAxis && index < NUM_PARTICLES; ++i) {
            for (int j = 0; j < particlesPerAxis && index < NUM_PARTICLES; ++j) {
                for (int k = 0; k < particlesPerAxis && index < NUM_PARTICLES; ++k) {
                    float x = (i - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING + offsetX;
                    float y = (j - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING + offsetY;
                    float z = (k - (particlesPerAxis - 1) / 2.0f) * GRID_SPACING + offsetZ;

                    float perturbation = 0.1f * GRID_SPACING;
                    float rand_x = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                    float rand_y = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;
                    float rand_z = ((float)rand() / RAND_MAX - 0.5f) * 2.0f * perturbation;

                    h_positions[index] = make_float3(x + rand_x, y + rand_y, z + rand_z);
                    h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f);
                    ++index;
                }
            }
        }
        return index; // Return the next index
    };

    int particlesPerAxis = 0;
    int index = 0;

    switch (mode)
    {
        case SimulationMode::OneLargeCube:
            particlesPerAxis = static_cast<int>(ceil(pow(NUM_PARTICLES, 1.0f / 3.0f)));
            index = initializeCube(0, particlesPerAxis, 0.0f, 4.0f, 0.5f);
            break;

        case SimulationMode::TwoSmallCubes:
        {
            int halfParticles = NUM_PARTICLES / 2; // Split particles evenly between two cubes
            particlesPerAxis = static_cast<int>(ceil(pow(halfParticles, 1.0f / 3.0f)));

            index = initializeCube(0, particlesPerAxis, 0.0f, 4.0f, 1.0f);
            index = initializeCube(index, particlesPerAxis, 8.0f, 4.0f, -1.0f);
            break;
        }
    }

    // Initialize remaining particles, if any
    for (; index < NUM_PARTICLES; ++index) {
        h_positions[index] = make_float3(0.0f, 10.0f, 0.0f);
        h_velocities[index] = make_float3(0.0f, 0.0f, 0.0f);
    }
}


void createBoundaryParticles(
    std::vector<float3> &boundary_positions, 
    std::vector<float3> &boundary_normals, // Renamed from boundary_velocities
    int boundaryLayers, 
    float spacing) 
{
    // Define boundary normals based on boundary geometry
    // For axis-aligned boxes, normals are straightforward

    // Compute inner box dimensions based on the specified fractions
    float box_size_x = (h_BOUND_X_MAX - h_BOUND_X_MIN);
    float box_size_y = (h_BOUND_Y_MAX - h_BOUND_Y_MIN);
    float box_size_z = (h_BOUND_Z_MAX - h_BOUND_Z_MIN);

    // Compute inner box center positions for X and Z axes
    float x_center = (h_BOUND_X_MAX + h_BOUND_X_MIN) / 2.0f;
    float z_center = (h_BOUND_Z_MAX + h_BOUND_Z_MIN) / 2.0f;

    // Define inner box minimum and maximum coordinates
    float inner_min_x = x_center - box_size_x / 2.0f;
    float inner_max_x = x_center + box_size_x / 2.0f;

    float inner_min_y = h_BOUND_Y_MIN; // Aligned with Y-min
    float inner_max_y = h_BOUND_Y_MIN + box_size_y;

    float inner_min_z = z_center - box_size_z / 2.0f;
    float inner_max_z = z_center + box_size_z / 2.0f;

    // Compute the number of particles along each dimension for the inner box
    int nx_inner = static_cast<int>((inner_max_x - inner_min_x) / spacing) + 1;
    int ny_inner = static_cast<int>((inner_max_y - inner_min_y) / spacing) + 1;
    int nz_inner = static_cast<int>((inner_max_z - inner_min_z) / spacing) + 1;

    // Lambda function to add boundary particles for a specific face
    auto addBoundaryFace = [&](int axis, bool positive) {
        for (int layer = 0; layer < boundaryLayers; layer++) {
            float positionOffset = layer * spacing;

            int loop_i_count, loop_j_count;
            // Determine which loops correspond to each axis
            if (axis == 0) {
                // X-axis walls: vary Y and Z
                loop_i_count = ny_inner;
                loop_j_count = nz_inner;
            } else if (axis == 1) {
                // Y-axis walls: vary X and Z
                loop_i_count = nx_inner;
                loop_j_count = nz_inner;
            } else {
                // Z-axis walls: vary X and Y
                loop_i_count = nx_inner;
                loop_j_count = ny_inner;
            }

            for (int i = 0; i < loop_i_count; i++) {
                for (int j = 0; j < loop_j_count; j++) {
                    float x, y, z, nx, ny, nz;
                    switch (axis) {
                        case 0: // X-axis (Left/Right walls)
                            x = positive ? (inner_max_x + positionOffset) : (inner_min_x - positionOffset);
                            y = inner_min_y + i * spacing;
                            z = inner_min_z + j * spacing;
                            nx = positive ? -1.0f : 1.0f;
                            ny = 0.0f;
                            nz = 0.0f;
                            break;

                        case 1: // Y-axis (Floor/Ceiling)
                            y = positive ? (inner_max_y + positionOffset) : (inner_min_y - positionOffset);
                            x = inner_min_x + i * spacing;
                            z = inner_min_z + j * spacing;
                            nx = 0.0f;
                            ny = positive ? -1.0f : 1.0f;
                            nz = 0.0f;
                            break;

                        case 2: // Z-axis (Front/Back walls)
                            z = positive ? (inner_max_z + positionOffset) : (inner_min_z - positionOffset);
                            x = inner_min_x + i * spacing;
                            y = inner_min_y + j * spacing;
                            nx = 0.0f;
                            ny = 0.0f;
                            nz = positive ? -1.0f : 1.0f;
                            break;
                    }

                    boundary_positions.emplace_back(make_float3(x, y, z));
                    boundary_normals.emplace_back(make_float3(nx, ny, nz)); // Assign normals
                }
            }
        }
    };

    // Add boundary particles for all three axes
    // Axis 0: X-axis (Left and Right Walls)
    addBoundaryFace(0, false); // X-min (Left Wall)
    addBoundaryFace(0, true);  // X-max (Right Wall)

    // Axis 1: Y-axis (Floor and Ceiling)
    addBoundaryFace(1, false); // Y-min (Floor)
    addBoundaryFace(1, true);  // Y-max (Ceiling)

    // Axis 2: Z-axis (Front and Back Walls)
    addBoundaryFace(2, false); // Z-min (Front Wall)
    addBoundaryFace(2, true);  // Z-max (Back Wall)
}

}